unit Horzbaru;

interface

uses
	Messages, Windows, Classes, Graphics, Controls, Menus, ExtCtrls;

type
	{integer = integer;}
   TDrawTickEvent = procedure(Sender: TObject) of object;

   TCustomTrack = class(TCustomControl)
       private
           fOwnerDrawTicks: TDrawTickEvent;
       protected
           procedure DrawTicks;
   	public
           procedure DrawMajorTicks(pos: integer); virtual; abstract;
           procedure DrawMinorTicks(pos: integer); virtual; abstract;
           {callback to draw ticks}
           property OwnerDrawTicks: TDrawTickEvent write fOwnerDrawTicks;
   end;

	THorzTrackBar = class(TCustomTrack)
   	private
       	fCursor: TObject;
       	fMax, fMin, fPosition, fPageSize: integer;
           fScreenPos: integer;
           fSliding: boolean;
           fCursorRect: TRect;
           {in screen coordinates:
           fMajorTickInterval: interval between major ticks > 0
           fFirstMajorTickPos: position of the lowest first tick mark > 0
           cMinorTickCount: number of minor ticks between each tick mark > 0}

           {fMajorTickInterval, fFirstMajorTickPos, fMinorTickCount: integer;}

           fDitherBmp, fCursorBmp, fMaskBmp, fBackgroundBmp: TBitmap;
           fOnChange: TNotifyEvent;

           function CursorToScreen( Value: integer): integer;
           function ScreenToCursor( Value: integer): integer;
           function LimitPosition( Value: integer): integer; {clips value to the control}
			procedure SetMax( Value : integer );
			procedure SetMin( Value : integer );
			procedure SetPosition( Value : integer );
           function GetScreenPosition: integer;
           procedure SetScreenPosition( Value : integer );
			procedure LoadThumbBitmaps;
			procedure UpdateDitherBitmap;
           procedure DrawTrack;
           procedure DrawCursor;
			procedure WMGetDlgCode( var Msg : TWMGetDlgCode ); message wm_GetDlgCode;
			procedure CMEnabledChanged( var Msg : TMessage ); message cm_EnabledChanged;
		protected
			procedure Paint; override;
			procedure Change; dynamic;
			procedure DoEnter; override;
			procedure DoExit; override;
			procedure KeyDown( var Key : Word; Shift : TShiftState ); override;
           procedure MouseDown( Button : TMouseButton; Shift : TShiftState; X, Y : Integer ); override;
			procedure MouseMove( Shift : TShiftState; X, Y : Integer ); override;
			procedure MouseUp( Button : TMouseButton; Shift : TShiftState; X, Y : Integer ); override;
   	public
			constructor Create( AOwner : TComponent ); override;
           constructor CreateInCursor( AOwner: TComponent; theCursor: TObject);
			destructor Destroy; override;
           function CursorVisible: boolean;
           procedure SetParams(theMax, theMin, thePos: integer);
           procedure DrawMajorTicks(pos: integer); override;
           procedure DrawMinorTicks(pos: integer); override;
      published
       	property ScreenPos: integer read GetScreenPosition write SetScreenPosition;
			property Max : integer read fMax write SetMax default 999;
			property Min : integer read fMin write SetMin default 0;
           property PageSize : integer read fPageSize write fPageSize default 50;
			property Position : integer read fPosition write SetPosition;
			property OnChange : TNotifyEvent read fOnChange write fOnChange;

			{ Inherited Properties & Events }
			property Color;
			property DragCursor;
			property DragMode;
			property Enabled;
			property HelpContext;
			property Hint;
			property ParentShowHint;
			property PopupMenu;
			property ShowHint;
			property TabOrder;
			property TabStop default True;
			property Visible;

			property OnClick;
			property OnDragDrop;
			property OnDragOver;
			property OnEndDrag;
			property OnEnter;
			property OnExit;
			property OnKeyDown;
			property OnKeyPress;
			property OnKeyUp;
			property OnMouseDown;
			property OnMouseMove;
			property OnMouseUp;
   end;

  procedure Register;

implementation

uses SysUtils {, cursorsu, viewdata};

{$R CURSORS}

procedure TCustomTrack.DrawTicks;
begin
   {calls the owner to draw ticks}
   if Assigned(fOwnerDrawTicks) then fOwnerDrawTicks(Self);
end;


procedure THorzTrackBar.DrawMajorTicks(pos: integer);
begin
	with Canvas do
   begin
   	Pen.Color := clBlack;
       Pen.Width := 1;
		MoveTo(pos, 8);
   	LineTo(pos, 17);
	end;
end;

procedure THorzTrackBar.DrawMinorTicks(pos: integer);
begin
	with Canvas do
   begin
   	Pen.Color := clBlack;
       Pen.Width := 1;
		MoveTo(pos, 12);
   	LineTo(pos, 17);
	end;
end;

function THorzTrackBar.CursorToScreen( Value: integer): integer;
begin
	Result := Muldiv(Value - fMin,  ClientWidth - 1, fMax - fMin);
   if Result < 0 then Result := 0;
   if Result >= ClientWidth then Result := ClientWidth - 1;
end;

function THorzTrackBar.ScreenToCursor( Value: integer): integer;
begin
	Result := fMin + Muldiv(Value, fMax - fMin, ClientWidth - 1);
end;

function THorzTrackBar.LimitPosition( Value: integer): integer;
begin
	Result :=Value;
   if Result > fMax then Result := fMax;
   if Result < fMin then Result := fMin;
end;

procedure THorzTrackBar.SetMax( Value : integer );
begin
	if value <> fMax then
   begin
   	fMax := Value;
       if fPosition > fMax then screenPos := CursorToScreen(fMax);
       fPageSize := (fMax - fMin) div 20;
       Invalidate;
   end;
end;

procedure THorzTrackBar.SetMin( Value : integer );
begin
	if value <> fMin then
   begin
   	fMin := Value;
       if fPosition < fMin then screenPos := CursorToScreen(fMin);
       fPageSize := (fMax - fMin) div 20;
       Invalidate;
   end;
end;

procedure THorzTrackBar.SetPosition( Value : integer );
begin
	if value <> fPosition then
   begin
       fPosition := Value;

       if csDesigning in ComponentState then
			Invalidate
		else
		begin
       	{ Erase old thumb image by drawing background bitmap }
      		Canvas.Draw( fCursorRect.Left, fCursorRect.Top, FBackgroundBmp );
			DrawCursor; { Draw thumb at new location }
			Change;  { Trigger Change event }
       end;
   end;
end;

function THorzTrackBar.GetScreenPosition;
begin
	Result := CursorToScreen(fPosition);
end;

procedure THorzTrackBar.SetScreenPosition( Value : Integer );
begin
	if Value < 0 then Value := 0;
   if Value >= ClientWidth then Value := ClientWidth - 1;
	fPosition := ScreenToCursor(Value);
   Invalidate;
end;

procedure THorzTrackBar.LoadThumbBitmaps;
const
	hCursor: PChar = 'HCURSOR';
   hMask: PChar = 'HMASK';
begin
	fCursorBmp.Handle := LoadBitmap(hInstance, hCursor);
	fMaskBmp.Handle := LoadBitmap(hInstance, hMask);
end;


procedure THorzTrackBar.UpdateDitherBitmap;
var
  i, j : integer;
begin
	with fDitherBmp.Canvas do
	begin
		Brush.Color := clWhite;
   	FillRect( Rect( 0, 0, fDitherBmp.Width, fDitherBmp.Height ) );
   	for i := 0 to 7 do
   		for j := 0 to 7 do
				if ( i + j ) mod 2 <> 0 then
					Pixels[ i, j ] := clSilver;
	end;
end;

procedure THorzTrackBar.DrawTrack;
begin
	Canvas.Brush.Color := clWhite;
   if not Enabled then
	   Canvas.Brush.Bitmap := fDitherBmp;

   Canvas.FillRect(ClientRect);
end;


procedure THorzTrackBar.DrawCursor;
var
	workBmp : TBitmap;
	workRct : TRect;
begin
	fScreenPos := CursorToScreen(fPosition);
   fCursorRect := Rect(fScreenPos - 7, 1, fScreenPos + 8, 16);

   fBackgroundBmp.Width := 15;
   fBackgroundBmp.Height := 15;
	fBackgroundBmp.Canvas.CopyRect( Rect(0, 0, fCursorBmp.Width, fCursorBmp.Height),
                                  Canvas, fCursorRect );
	workBmp := TBitmap.Create;
	try
		workBmp.Height := fCursorBmp.Height;
		workBmp.Width := fCursorBmp.Width;

		workRct := Rect( 0, 0, fCursorBmp.Width, fCursorBmp.Height);

		workBmp.Canvas.CopyMode := cmSrcCopy;
		workBmp.Canvas.CopyRect( WorkRct, fBackgroundBmp.Canvas, workRct );

		workBmp.Canvas.CopyMode := cmSrcAnd;
		workBmp.Canvas.CopyRect( WorkRct, fMaskBmp.Canvas, WorkRct );

		workBmp.Canvas.CopyMode := cmSrcPaint;
       WorkBmp.Canvas.CopyRect( WorkRct, fCursorBmp.Canvas, WorkRct );

		if not Enabled then
		begin
			WorkBmp.Canvas.Brush.Bitmap := fDitherBmp;
			WorkBmp.Canvas.FloodFill( WorkRct.Right - 3, WorkRct.Bottom - 3,
                                clSilver, fsSurface );
		end;

		Canvas.CopyRect( fCursorRect, WorkBmp.Canvas, WorkRct );
   finally
			workBmp.Free;
	end;
end;


procedure THorzTrackBar.WMGetDlgCode( var Msg : TWMGetDlgCode );
begin
	inherited;
	Msg.Result := dlgc_WantArrows;
end;

procedure THorzTrackBar.CMEnabledChanged( var Msg : TMessage );
begin
  inherited;
  Invalidate;
end;

procedure THorzTrackBar.Paint;
begin
	with Canvas do
   begin
       DrawTrack;
       DrawTicks;
       DrawCursor;
   end;
end;

procedure THorzTrackBar.Change;
begin
	if Assigned( FOnChange ) then FOnChange( Self );
end;

procedure THorzTrackBar.DoEnter;
begin
	inherited DoEnter;
	Refresh;
end;

procedure THorzTrackBar.DoExit;
begin
	inherited DoExit;
	Refresh;
end;

procedure THorzTrackBar.KeyDown( var Key : Word; Shift : TShiftState );
begin
	inherited KeyDown( Key, Shift );
	case Key of
   	vk_Prior:
       	Position := LimitPosition(fPosition + fPageSize);
		vk_Next:
			Position := LimitPosition(fPosition - FPageSize);
		vk_End:
			Position := fMin;
		vk_Home:
			Position := fMax;
		vk_Left:
			if fPosition > fMin then Position := LimitPosition(fPosition - 1);
		vk_Up:
       	if fPosition < fMax then Position := LimitPosition(fPosition + 1);
		vk_Right:
			if fPosition < fMax then Position := LimitPosition(fPosition + 1);
		vk_Down:
			if fPosition > fMin then Position := LimitPosition(fPosition - 1);
	end;
end;

procedure THorzTrackBar.MouseDown( Button : TMouseButton; Shift : TShiftState; X, Y : Integer );
begin
	inherited MouseDown( Button, Shift, X, Y );
	SetFocus;

	if ( Button = mbLeft ) and PtInRect( fCursorRect, Point( X, Y ) ) then
		fSliding := True
	else
	begin
       if ScreenToCursor(X) < fPosition then
       	Position := LimitPosition(fPosition - fPageSize)
   	else
      		Position := LimitPosition(fPosition + fPageSize);
	end;
end;

procedure THorzTrackBar.MouseMove( Shift : TShiftState; X, Y : Integer );
var   {	h : Integer;}
		p: integer;
begin
	inherited MouseMove( Shift, X, Y );

	if PtInRect( FCursorRect, Point( X, Y ) ) then
   	Cursor := crSizeWE
	else
   	Cursor := crDefault;
	if fSliding then
	begin
    {	h := ClientWidth - 7;}
       p:=  fMin + Muldiv(X, fMax - fMin, ClientWidth - 1);
		if p > fMax then p := fMax;
       if p < fMin then p := fMin;
   	Position := p;
   end;
end;

procedure THorzTrackBar.MouseUp( Button : TMouseButton; Shift : TShiftState; X, Y : Integer );
begin
	inherited MouseUp( Button, Shift, X, Y );
	if ( Button = mbLeft ) then fSliding := False;
end;

constructor THorzTrackBar.Create( AOwner : TComponent );
begin
	inherited Create( AOwner );
	Width := 200;
	Height := 17;
	fMin := 0;
	fMax := 65536;
	fPosition := 0;
	fPageSize := 65536 div 20;
	fSliding := False;

	fCursorBmp := TBitmap.Create;
  	fCursorBmp.Width := 16;
	fCursorBmp.Height := 16;
	fMaskBmp := TBitmap.Create;
	fBackgroundBmp := TBitmap.Create;
   fDitherBmp := TBitmap.Create;
  	fDitherBmp.Width := 8;
	fDitherBmp.Height := 8;
	UpdateDitherBitmap;
	LoadThumbBitmaps;
end;

constructor THorzTrackBar.CreateInCursor( AOwner: TComponent; theCursor: TObject);
begin
	Create(AOwner);
   fCursor := theCursor;
end;

destructor THorzTrackBar.Destroy;
begin
  	FreeAndNil(fCursorBmp);
	FreeAndNil(fMaskBmp);
	FreeAndNil(fBackgroundBmp);
   FreeAndNil(fDitherBmp);
   inherited Destroy;
end;

function THorzTrackBar.CursorVisible: boolean;
begin
	if (fPosition <= fMax) and (fPosition >= fMin) then Result := True else Result := False;
end;

procedure THorzTrackBar.SetParams(theMax, theMin, thePos: integer);
begin
	fMax := theMax;
   fMin := theMin;
   fPosition := thePos;
   fPageSize := (fMax - fMin) div 20;
end;


procedure Register;
begin
	RegisterComponents( 'Samples', [THorzTrackBar] );
end;



end.
