unit ROIAxisDlgu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Spin, ComCtrls, Buttons, ROIFrmu;

type
  TROIAxisDlg = class(TForm)
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    fromFrameEdit: TSpinEdit;
    toFrameEdit: TSpinEdit;
    cXTicksEdit: TSpinEdit;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Bevel1: TBevel;
    Label4: TLabel;
    FromYEdit: TSpinEdit;
    Label5: TLabel;
    ToYEdit: TSpinEdit;
    Bevel2: TBevel;
    Label6: TLabel;
    cYTicksEdit: TSpinEdit;
    procedure FormShow(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    ROIfrm: TROIFrm;
  end;

var
  ROIAxisDlg: TROIAxisDlg;

implementation

{$R *.DFM}


procedure TROIAxisDlg.FormShow(Sender: TObject);
begin
   with ROIfrm do
   begin
       {frames are 0-based}
       fromFrameEdit.Value := xLeft;
       toFrameEdit.Value := xLeft + xWidth - 1;
       cXTicksEdit.Value := cXTickMarks;
       FromYEdit.Value := yTop - yHeight + 1;
       ToYEdit.Value := yTop;
       cYTicksEdit.Value := cYTickMarks;
   end;
end;

type
   TAxisFrmError = (AFE_NO_ERROR, AFE_FROM_FRAME, AFE_TO_FRAME, AFE_XTICKS,
       AFE_Y_FROM, AFE_Y_TO, AFE_YTICKS, AFE_X_ORDER, AFE_Y_ORDER);

resourcestring
   sInvalidFromFrame = 'Invalid "From Frame" value.';
   sInvalidToFrame = 'Invalid "To Frame" value.';
   sInvalidXTicks = 'Invalid "Number of ticks" value for the X- axis.';
   sInvalidYFrom = 'Invalid "From" value for the Y- axis.';
   sInvalidYTo = 'Invalid "To" value for the Y- axis.';
   sInvalidYTicks = 'Invalid "Number of ticks" value for the Y- axis.';
   sInvalidXOrder = 'X-Axis "From Frame" value must be inferior than the "To Frame" value.';
   sInvalidYOrder = 'Y-Axis "From" value must be inferior than the "To" value.';

   
procedure TROIAxisDlg.BitBtn1Click(Sender: TObject);
var    AxisFrmError: TAxisFrmError;
begin
   AxisFrmError := AFE_NO_ERROR;

   with ROIfrm do
   begin
       if (fromFrameEdit.Value < 0) or (fromFrameEdit.Value >= dataSize) then
           AxisFrmError := AFE_FROM_FRAME;
       if (toFrameEdit.Value < 0) or (toFrameEdit.Value >= dataSize) then
           AxisFrmError := AFE_TO_FRAME;
       if fromFrameEdit.Value >= toFrameEdit.Value then
           AxisFrmError := AFE_X_ORDER;

       if (cXTicksEdit.Value < 0) or (cXTicksEdit.Value > 10) then
           AxisFrmError := AFE_XTICKS;

       if (FromYEdit.Value > f_MAX_Y_VALUE) or (FromYEdit.Value < - f_MAX_Y_VALUE - 1) then
           AxisFrmError := AFE_Y_FROM;
       if (ToYEdit.Value > f_MAX_Y_VALUE) or (ToYEdit.Value < - f_MAX_Y_VALUE - 1) then
           AxisFrmError := AFE_Y_TO;
       if FromYEdit.Value >= ToYEdit.Value then
           AxisFrmError := AFE_Y_ORDER;

       if (cYTicksEdit.Value < 0) or (cYTicksEdit.Value > 10) then
           AxisFrmError := AFE_YTICKS;
   end;

   case AxisFrmError of
       AFE_FROM_FRAME: MessageDlg(sInvalidFromFrame, mtError, [mbOK], 0);
       AFE_TO_FRAME:   MessageDlg(sInvalidToFrame, mtError, [mbOK], 0);
       AFE_XTICKS:     MessageDlg(sInvalidXTicks, mtError, [mbOK], 0);
       AFE_Y_FROM:     MessageDlg(sInvalidYFrom, mtError, [mbOK], 0);
       AFE_Y_TO:       MessageDlg(sInvalidYTo, mtError, [mbOK], 0);
       AFE_YTICKS:     MessageDlg(sInvalidYTicks, mtError, [mbOK], 0);
       AFE_X_ORDER:    MessageDlg(sInvalidXOrder, mtError, [mbOK], 0);
       AFE_Y_ORDER:    MessageDlg(sInvalidYOrder, mtError, [mbOK], 0);
   end;

   if AxisFrmError = AFE_NO_ERROR then
   with ROIfrm do
   begin
       SavePrevMag;
       xLeft := fromFrameEdit.Value;
       xWidth := toFrameEdit.Value - xLeft + 1;
       cXTickMarks := cXTicksEdit.Value;
       yTop := ToYEdit.Value;
       yHeight := yTop - FromYEdit.Value + 1;
       cYTickMarks := cYTicksEdit.Value;
   end
   else
       ModalResult := mrNone;
end;

end.
