unit ROIu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, Trackcur, Horzbaru, ComCtrls, MPUnit, ROIThreadu, ExtCtrls;

type
  TROIFrm = class(TForm)
    MainMenu1: TMainMenu;
    Axis1: TMenuItem;
    XAxis1: TMenuItem;
    YAxis1: TMenuItem;
    Data1: TMenuItem;
    Copydatatoclipboard1: TMenuItem;
    StatusBar1: TStatusBar;
    HorzTrackBar1: THorzTrackBar;
    VertTrackBar1: TVertTrackBar;
    N1: TMenuItem;
    ResetCursorRelativeOrigins1: TMenuItem;
    PopupMenu1: TPopupMenu;
    ResetRelative1: TMenuItem;
    N2: TMenuItem;
    XAxisFullRange1: TMenuItem;
    YAxisFullRange1: TMenuItem;
    PreviousMagnification1: TMenuItem;
    YAxisAutoscale1: TMenuItem;
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure XAxis1Click(Sender: TObject);
    procedure YAxis1Click(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure ResetCursorRelativeOrigins1Click(Sender: TObject);
    procedure HorzTrackBar1Change(Sender: TObject);
    procedure Copydatatoclipboard1Click(Sender: TObject);
    procedure XAxisFullRange1Click(Sender: TObject);
    procedure SetXAxissize1Click(Sender: TObject);
    procedure YAxisFullRange1Click(Sender: TObject);
    procedure YAxisAutoscale1Click(Sender: TObject);
    procedure PreviousMagnification1Click(Sender: TObject);
    procedure VertTrackBar1Change(Sender: TObject);
  private
    { Private declarations }
    bInitialized, {prevents FormPaint before first FormResize}
    bAdjustingCursorPos: boolean; {prevent activation of OnChange events}
    fDataSize: integer;
    fDisplayMode: TDisplayMode;

    prev_xRight, prev_xWidth,
    prev_yTop, prev_yHeight: integer;

    {rectangle areas}
    bitmapWidth, bitmapHeight: integer;
    viewRect,
    xAxisRect,
    yAxisRect,
    cornerRect: TRect;

    xCursorPos,
    yCursorPos,
    refXCursorPos,
    refYCursorPos: integer;

    roiThread: TROIThread;
    procedure AdjustCursorPos;
    procedure AdjustLogicalCursorPos;
    procedure DrawBitmap;
    procedure DrawData;
    procedure DrawXAxis;
    procedure DrawXCursorOnBitmap;
    procedure DrawXCursorOnScreen;
    procedure DrawYAxis;
    procedure DrawYCursorOnBitmap;
    procedure DrawYCursorOnScreen;
    procedure OnMagnificationChanged;
    function  RectBitmapToWindow(aRect: TRect): TRect;
    function  RectWindowToBitmap(aRect: TRect): TRect;
    procedure ResizeElements;
    procedure SetDataSize(newSize: integer);
    procedure SetPrecision(newPrecision: TPrecision);
    procedure UpdateCursorCaptions;
    function  XBitmapToData(xBitmap: integer): integer;
    function  XDataToBitmap(xData: integer): integer;
    function  YBitmapToData(yBitmap: integer): integer;
    function  YDataToBitmap(yData: integer): integer;
  public
    { Public declarations }
    f_MAX_Y_VALUE: integer;

    {axis}
    xRight, xWidth,
    yTop, yHeight: integer; {logical values}
    cXTickMarks, cYTickMarks: integer;

    channel: integer;
    ROIRect: TRect; {contains the absolute coordinates of the ROI rectangle}
    ROIRegion: TRect; {the ROI in the frame buffer}
    screenBitmapROIRect: TRect;

    trace: TTrace;

    ROIThreadMutex: THandle;
    framesDisplayed: integer;
    nextFrame: integer; {the index in dataBuffer for the next frame}
    circdataBuffer: array of integer; {a circular buffer updated by the thread}
    databuffer: array of integer; {used by TROIFrm to store data: non circular}

    offscreenBitmap: TBitmap;

    procedure CopyCircData; {copies circdataBuffer into dataBuffer - used by ROIThread}
    procedure CopyOffscreenToScreen; {forces update of the screen}
    procedure OnScanningStart; {resets the circular buffer}
    procedure OnScanningEnd;
    procedure RedrawOffscreen; {called by the thread}
    procedure SavePrevMag;

    property dataSize: integer read fDataSize write SetDataSize;
    property displayMode: TDisplayMode write fDisplayMode;
    property precision: TPrecision write SetPrecision;
  end;

var
  ROIFrm: TROIFrm;

implementation

{$R *.DFM}

uses ROIAxisDlgu, Mainfrm, Clipbrd;

const
   TRACK_BAR_WIDTH = 17;
   TRACK_BAR_HEIGHT = 17;
   Y_AXIS_WIDTH = 75;
   X_AXIS_HEIGHT = 35;

{**************************** PRIVATE METHODS *********************************}

procedure TROIFrm.AdjustCursorPos;
begin
   bAdjustingCursorPos := True;
   VertTrackBar1.Position := Muldiv(yCursorPos - (yTop - yHeight + 1),
       bitmapHeight - 1, yHeight - 1);
   HorzTrackBar1.Position := Muldiv(xCursorPos - (xRight + xWidth - 1), bitmapWidth - 1, - xWidth + 1);
   bAdjustingCursorPos := False;
end;


procedure TROIFrm.AdjustLogicalCursorPos;
begin
   xCursorPos := XBitmapToData(HorzTrackBar1.Position);
   yCursorPos := YBitmapToData(VertTrackBar1.Position);
end;


procedure TROIFrm.DrawBitmap;
begin
   with offscreenBitmap.Canvas do
   begin
       Brush.Color := clBlack;
       FillRect(Rect(0, 0, bitmapWidth - 1, bitmapHeight - 1));
   end;
end;


procedure TROIFrm.DrawData;
var    lastIndex, firstIndex, i, j, k, indexofnextpt,
       bufferlimit, dataVal, maxVal, minVal: integer;
       bitmapY, bitmaprY: integer;
       firstPPt: ^TPoint;
begin
   if framesDisplayed > 0 then
   begin
       lastIndex := -1; firstIndex := -1;
       if framesDisplayed < dataSize then
           bufferlimit := nextFrame
       else
           bufferlimit := dataSize;

       for i := bitmapWidth - 1 downto 0 do {index in trace}
       begin
           j := XBitmapToData(i); {index in data buffer}
           indexofnextpt := XBitmapToData(i - 1) - 1;

           if (j >= 0) and (j < bufferlimit) then
           begin
               if lastIndex = -1 then lastIndex := i;
               firstIndex := i;
               trace[i].MaxPt.y := dataBuffer[j];
               trace[i].MinPt.y := trace[i].MaxPt.y;

               if (indexofnextPt > j) and (indexofnextPt < bufferlimit) and
                   (j < bufferlimit - 1) then
                   for k := j + 1 to indexofnextPt do
                   begin
                       dataVal := dataBuffer[k];
                       if dataVal > trace[i].MaxPt.y then trace[i].MaxPt.y := dataVal;
                       if dataVal < trace[i].MinPt.y then trace[i].MinPt.y := dataVal;
                   end;
           end; {j}
       end; {i}

       if (lastIndex <> -1) and (firstIndex <> -1) and (firstIndex < lastIndex) then
       begin
           if YAxisAutoscale1.Checked then
           begin
               {Autoscale}
               maxVal := trace[firstIndex].MaxPt.y;
               minVal := trace[firstIndex].MinPt.y;

               {find the maximum and minimum values}
               for i := firstIndex + 1 to lastIndex do
               begin
                   if maxVal < trace[i].MaxPt.y then maxVal := trace[i].MaxPt.y;
                   if minVal < trace[i].MinPt.y then minVal := trace[i].MinPt.y;
               end;

               {at least 100 pixel difference for autoscale}
               if maxVal >= minVal + 100 then
               begin
                   bitmapY := YDataToBitmap(YCursorPos);
                   bitmaprY := YDataToBitmap(refYCursorPos);

                   yTop := (maxVal + minVal) div 2 + Muldiv(10, (maxVal - minVal + 1) div 2, 9);
                   yHeight := Muldiv(10, maxVal - minVal + 1, 9);
                   if yTop > f_MAX_Y_VALUE then yTop := f_MAX_Y_VALUE;
                   if yTop - yHeight < - f_MAX_Y_VALUE - 1 then
                       yHeight := f_MAX_Y_VALUE + 1 + yTop;

                   DrawYAxis;
                   YCursorPos := YBitmapToData(bitmapY);
                   refYCursorPos := YBitmapToData(bitmaprY);
                   UpdateCursorCaptions;
               end;
           end;

           {logical to bitmap Y}
           for i := firstIndex to lastIndex do
           begin
               dataVal := trace[i].MaxPt.y; trace[i].MaxPt.y := YDataToBitmap(dataVal);
               dataVal := trace[i].MinPt.y; trace[i].MinPt.y := YDataToBitmap(dataVal);
           end;

           with offscreenbitmap.Canvas do
           begin
               Pen.Color := clWhite;
               Pen.Mode := pmCopy;
               Pen.Style := psSolid;
               Pen.Width := 1;
           end;

           firstPPt := @trace[firstIndex].MaxPt;
           Windows.Polyline(offscreenBitmap.Canvas.Handle, firstPPt,
               2 * (lastIndex - firstIndex + 1));
       end
   end; {framesDisplayed > 0}
end;


procedure TROIFrm.DrawXAxis;
var    i, x: integer;
       s: string;
begin
   Canvas.Brush.Color := clBtnFace;
   Canvas.FillRect(xAxisRect);

   with Canvas.Pen do
   begin
       Color := clBlue;
       Width := 1;
       Style := psSolid;
       Mode := pmCopy;
   end;

   for i := 0 to cXTickMarks - 1 do
   begin
       x := viewRect.Left + MulDiv(i, bitmapWidth - 1, cXTickMarks - 1);
       Canvas.MoveTo(x, xAxisRect.Top);
       Canvas.LineTo(x, xAxisRect.Top + 10);
   end;

   s := Format('%.2g', [xWidth / (cXTickMarks - 1)]) + ' frames / division';
   Canvas.Textout(ClientWidth div 2 - Canvas.TextWidth(s) div 2, xAxisRect.Bottom - 20, s);
end;


procedure TROIFrm.DrawXCursorOnBitmap;
var    bitmapXCursor: integer;
begin
   bitmapXCursor := XDataToBitmap(xCursorPos);
   with offscreenbitmap.Canvas do
   begin
       Pen.Color := clYellow;
       Pen.Mode := pmXor;
       Pen.Style := psDot;
   end;
   Canvas.MoveTo(bitmapXCursor, 0);
   Canvas.LineTo(bitmapXCursor, bitmapHeight - 1);
end;


procedure TROIFrm.DrawXCursorOnScreen;
var    bitmapXCursor: integer;
       rc: TRect;
begin
   bitmapXCursor := XDataToBitmap(xCursorPos);
   rc := Rect(bitmapXCursor - 1, 0, bitmapXCursor + 1, bitmapHeight - 1);
   Canvas.CopyRect(rc, offscreenBitmap.Canvas, RectBitmapToWindow(rc));
end;


procedure TROIFrm.DrawYAxis;
var    i, y, tw, th: integer;
       s: string;
begin
   Canvas.Brush.Color := clBtnFace;
   Canvas.FillRect(yAxisRect);

   with Canvas.Pen do
   begin
       Color := clBlue;
       Width := 1;
       Style := psSolid;
       Mode := pmCopy;
   end;

   for i := 0 to cYTickMarks - 1 do
   begin
       y := viewRect.Top + MulDiv(i, bitmapHeight - 1, cYTickMarks - 1);
       Canvas.MoveTo(yAxisRect.Right - 10, y);
       Canvas.LineTo(yAxisRect.Right, y);
       s := IntToStr(yTop - Muldiv(i, yHeight - 1, cYTickMarks - 1));
       tw := Canvas.TextWidth(s);
       th := Canvas.TextHeight(s);
       Canvas.TextOut(yAxisRect.Right - tw - 15, y - th div 2, s);
   end;
end;


procedure TROIFrm.DrawYCursorOnBitmap;
var    bitmapYCursor: integer;
begin
   bitmapYCursor := YDataToBitmap(yCursorPos);
   with offscreenbitmap.Canvas do
   begin
       Pen.Color := clYellow;
       Pen.Mode := pmXor;
       Pen.Style := psDot;
   end;
   Canvas.MoveTo(0, bitmapYCursor);
   Canvas.LineTo(bitmapWidth - 1, bitmapYCursor);
end;


procedure TROIFrm.DrawYCursorOnScreen;
var    bitmapYCursor: integer;
       rc: TRect;
begin
   bitmapYCursor := YDataToBitmap(yCursorPos);
   rc := Rect(0, bitmapYCursor - 1,  bitmapWidth - 1, bitmapYCursor + 1);
   Canvas.CopyRect(rc, offscreenBitmap.Canvas, RectBitmapToWindow(rc));
end;


procedure TROIFrm.OnMagnificationChanged;
begin
   RedrawOffScreen;
   AdjustLogicalCursorPos;
   Invalidate;
   UpdateCursorCaptions;
end;


function  TROIFrm.RectBitmapToWindow(aRect: TRect): TRect;
begin
   OffsetRect(aRect, viewRect.Left, viewRect.Top);
   Result := aRect;
end;


function  TROIFrm.RectWindowToBitmap(aRect: TRect): TRect;
begin
   OffsetRect(aRect, -viewRect.Left, -viewRect.Top);
   Result := aRect;
end;


procedure TROIFrm.ResizeElements;
var    i: integer;
begin
   {rectangles}
   cornerRect := Rect(ClientWidth - TRACK_BAR_WIDTH, 0, ClientWidth - 1, TRACK_BAR_HEIGHT - 1);
   yAxisRect := Rect(0, 0, Y_AXIS_WIDTH - 1, ClientHeight - StatusBar1.Height - 1);
   xAxisRect := Rect(Y_AXIS_WIDTH, ClientHeight - StatusBar1.Height - X_AXIS_HEIGHT - 1, ClientWidth,
       ClientHeight - StatusBar1.Height - 1);
   viewRect := Rect(Y_AXIS_WIDTH, TRACK_BAR_HEIGHT, ClientWidth - TRACK_BAR_WIDTH - 1,
       ClientHeight - X_AXIS_HEIGHT - StatusBar1.Height - 1);

   {bitmap and trace}
   offscreenBitmap.Width := viewRect.Right - viewRect.Left + 1;
   offscreenBitmap.Height := viewRect.Bottom - viewRect.Top + 1;

   bitmapWidth := viewRect.Right - viewRect.Left + 1;
   bitmapHeight := viewRect.Bottom - viewRect.Top + 1;

   SetLength(trace, bitmapWidth);
   for i := 0 to bitmapWidth - 1 do
   begin
       trace[i].MaxPt.x := i;
       trace[i].MinPt.x := i;
   end;

   {track bars}
   HorzTrackBar1.SetBounds(0, Y_AXIS_WIDTH, ClientWidth - TRACK_BAR_WIDTH, TRACK_BAR_HEIGHT);
   VertTrackBar1.SetBounds(ClientWidth - TRACK_BAR_WIDTH, TRACK_BAR_HEIGHT,
       TRACK_BAR_WIDTH, ClientHeight - TRACK_BAR_HEIGHT - StatusBar1.Height);

   HorzTrackBar1.Min := 0;
   HorzTrackBar1.Max := bitmapWidth - 1;
   VertTrackBar1.Min := 0;
   VertTrackBar1.Max := bitmapHeight - 1;

   RedrawOffscreen;
end;


procedure TROIFrm.SetDataSize(newSize: integer);
begin
   fDataSize := newSize;
   SetLength(circDataBuffer, fDataSize);
   SetLength(dataBuffer, fDataSize);
   framesDisplayed := 0;
   nextFrame := 0;
end;


procedure TROIFrm.SetPrecision(newPrecision: TPrecision);
begin
   case newPrecision of
       PREC_8_BIT: f_MAX_Y_VALUE := 127;
       PREC_10_BIT: f_MAX_Y_VALUE := 511;
       PREC_12_BIT: f_MAX_Y_VALUE := 2047;
       PREC_14_BIT: f_MAX_Y_VALUE := 8191;
       else f_MAX_Y_VALUE := 32767;
   end;
end;


procedure TROIFrm.UpdateCursorCaptions;
begin
   StatusBar1.Panels[0].Text := 'X: ' + IntToStr(xCursorPos);
   StatusBar1.Panels[1].Text := 'Y: ' + IntToStr(yCursorPos);
   StatusBar1.Panels[2].Text := 'rX: ' + IntToStr(xCursorPos - refXCursorPos);
   StatusBar1.Panels[3].Text := 'rY: ' + IntToStr(YCursorPos - refYCursorPos);
end;


function  TROIFrm.XBitmapToData(xBitmap: integer): integer;
begin
   if framesDisplayed > 0 then
       Result := MulDiv(bitmapWidth - 1 - xBitmap, dataSize - 1, bitmapWidth - 1)
   else
       Result := -1;
end;


function  TROIFrm.XDataToBitmap(xData: integer): integer;
begin
   Result := bitmapWidth - 1 - Muldiv(xData, bitmapWidth - 1, dataSize - 1);
end;


function  TROIFrm.YBitmapToData(yBitmap: integer): integer;
begin
   Result := yTop - MulDiv(yBitmap, yHeight - 1, bitmapHeight - 1);
end;


function  TROIFrm.YDataToBitmap(yData: integer): integer;
begin
   Result := MulDiv(- yData + yTop, bitmapHeight - 1, yHeight - 1);
end;


{****************************** FORM EVENTS ***********************************}

procedure TROIFrm.FormCreate(Sender: TObject);
begin
   f_MAX_Y_VALUE := 2047;
   displayMode := DM_ANALYZING;
   offscreenBitmap := TBitmap.Create;
   offscreenBitmap.handleType := bmDDB;
   dataSize := 1000; {default buffer size}
   roiThreadMutex := CreateMutex(nil, False, nil);

   xRight := 0;
   xWidth := dataSize;
   yTop := f_MAX_Y_VALUE;
   yHeight := f_MAX_Y_VALUE * 2 + 1;

   cXTickMarks := 5;
   cYTickMarks := 5;
   prev_xRight := xRight;
   prev_xWidth := xWidth;
   prev_yTop := yTop;
   prev_yHeight := yHeight;
end;


procedure TROIFrm.FormResize(Sender: TObject);
begin
   ResizeElements;
   AdjustCursorPos;
   bInitialized := True;
   Invalidate;
end;


procedure TROIFrm.FormPaint(Sender: TObject);
var    rc, interRect: TRect;
begin
   if not bInitialized then Exit;
   rc := Canvas.ClipRect;
   Canvas.Brush.Color := clBtnFace;
   Canvas.Pen.Color := clBlack;
   Canvas.Pen.Mode := pmCopy;
   Canvas.Pen.Style := psSolid;

   if IntersectRect(interRect, rc, cornerRect) then
       Canvas.FillRect(cornerRect);

   if IntersectRect(interRect, rc, xAxisRect) then
       DrawXAxis;

   if IntersectRect(interRect, rc, yAxisRect) then
       DrawYAxis;

   if IntersectRect(interRect, rc, viewRect) then
       Canvas.CopyRect(interRect, offscreenBitmap.Canvas, RectWindowToBitmap(interRect));
end;


procedure TROIFrm.FormDestroy(Sender: TObject);
begin
   offscreenBitmap.Free;
   if not Mainform.bClosing then
       Mainform.ROIWndList.Remove(self);
   CloseHandle(roiThreadMutex);
end;


{*********************************** MENUS ************************************}

procedure TROIFrm.XAxis1Click(Sender: TObject);
begin
   ROIAxisDlg := TROIAxisDlg.Create(self);
   try
       ROIAxisDlg.ROIfrm := self;
       ROIAxisDlg.PageControl1.ActivePageIndex := 0;
       if ROIAxisDlg.ShowModal = mrOK then
           OnMagnificationChanged;
   finally
       ROIAxisDlg.Free;
   end;
end;


procedure TROIFrm.YAxis1Click(Sender: TObject);
begin
   ROIAxisDlg := TROIAxisDlg.Create(self);
   try
       ROIAxisDlg.ROIfrm := self;
       ROIAxisDlg.PageControl1.ActivePageIndex := 1;
       if ROIAxisDlg.ShowModal = mrOK then
           OnMagnificationChanged;
   finally
       ROIAxisDlg.Free;
   end;
end;


procedure TROIFrm.YAxisAutoscale1Click(Sender: TObject);
begin
   YAxisAutoscale1.Checked := not YAxisAutoscale1.Checked;
end;


procedure TROIFrm.XAxisFullRange1Click(Sender: TObject);
begin
   xRight := 0;
   xWidth := dataSize;
   OnMagnificationChanged;
end;


procedure TROIFrm.YAxisFullRange1Click(Sender: TObject);
begin
   yTop :=  f_MAX_Y_VALUE;
   yHeight := 2 * f_MAX_Y_VALUE + 1;
   OnMagnificationChanged;
end;


procedure TROIFrm.PreviousMagnification1Click(Sender: TObject);
begin
    xRight := prev_xRight;
    xWidth := prev_xWidth;
    yTop := prev_yTop;
    yHeight := prev_yHeight;
    OnMagnificationChanged;
end;


procedure TROIFrm.SetXAxissize1Click(Sender: TObject);
var    s: string;
       newSize: integer;
begin
   s := IntToStr(dataSize);
   if InputQuery('Set trace size', 'Enter trace size',s) then
   try
       newSize := StrToInt(s);
       if (newSize < 100) or (newSize > 10000) then
           MessageDlg('Trace size must be between 100 and 10000', mtError, [mbOK], 0)
       else
           dataSize := newSize;
   except
       MessageDlg('Invalid numeric value', mtError, [mbOK], 0);
   end;
end;


procedure TROIFrm.ResetCursorRelativeOrigins1Click(Sender: TObject);
begin
   refXCursorPos := xCursorPos;
   refYCursorPos := yCursorPos;
   UpdateCursorCaptions;
end;


procedure TROIFrm.Copydatatoclipboard1Click(Sender: TObject);
type   TCharArray = array[1..Maxint div 2] of Char;
var    i, cPts: integer;
       memhandle: THandle;
       pString: ^TCharArray;
       s1Length, stringindex: integer;
       s1: string;
begin
   if framesDisplayed > 0 then
   try
       Clipboard.Open;
       Clipboard.Clear;

       {fills string with data; 11 digits for each column}
       if framesDisplayed < dataSize then
           cPts := nextFrame
       else
           cPts := dataSize;

       memHandle := GlobalAlloc(GMEM_MOVEABLE or GMEM_DDESHARE or GMEM_ZEROINIT, 24 *
           cPts + 1);

       pString := GlobalLock(memHandle);
       stringindex := 0;

       for i := 0 to cPts - 1 do
       begin
           s1 := IntToStr(i) + TAB;
           s1Length := Length(s1);
           CopyMemory(@(pString^[stringIndex]), @s1[1], s1Length);
           stringIndex := stringIndex + s1Length; {points to the next character}

           s1 := IntToStr(dataBuffer[i]) + CR;
           s1Length := Length(s1);
           CopyMemory(@(pString^[stringIndex]), @s1[1], s1Length);
           stringIndex := stringIndex + s1Length; {points to the next character}
       end;

       Clipboard.SetAsHandle(CF_TEXT, memHandle);
   finally
       Clipboard.Close;
   end;
end;


{****************************** PUBLIC METHODS *********************************}

procedure TROIFrm.CopyCircData; {copies circdataBuffer into dataBuffer}
var i: integer;
begin
   if framesDisplayed > 0 then
   begin
       if (framesDisplayed < dataSize) then
       begin
           if nextFrame > 0 then
               for i := 0 to nextFrame - 1 do dataBuffer[i] := circdataBuffer[i];
       end
       else
       begin
           {uncircularize data}
           for i := nextFrame to dataSize - 1 do
               dataBuffer[i - nextFrame] := circdataBuffer[i];

           if nextFrame > 0 then
               for i := 0 to nextFrame - 1 do
                   dataBuffer[i + dataSize - nextFrame] := circdataBuffer[i];
       end;
   end;
end;


procedure TROIFrm.CopyOffscreenToScreen; {forces update of the screen}
begin
   Canvas.CopyRect(viewRect, offscreenBitmap.Canvas, Rect(0, 0, bitmapWidth - 1, bitmapHeight - 1));
end;


procedure TROIFrm.OnScanningStart; {resets the circular buffer}
begin
   SetXAxissize1.Enabled := False;
   Copydatatoclipboard1.Enabled := False;
   framesDisplayed := 0;
   nextFrame := 0;

   roiThread := TROIThread.Create(True); {thread does not start immediatly}
   roiThread.FreeOnTerminate := True;
   roiThread.roiFrm := self;
   roiThread.Resume;
end;


procedure TROIFrm.OnScanningEnd;
begin
   SetXAxissize1.Enabled := True;
   Copydatatoclipboard1.Enabled := True;
end;


procedure TROIFrm.RedrawOffscreen; {called by the thread}
begin
   {Drawing}
   DrawBitmap;

   {Data}
   DrawData;

   {Cursors}
   DrawXCursorOnBitmap;
   DrawYCursorOnBitmap;
end;


procedure TROIFrm.SavePrevMag;
begin
    prev_xRight := xRight;
    prev_xWidth := xWidth;
    prev_yTop := yTop;
    prev_yHeight := yHeight;
end;


{*********************************** EVENTS ***********************************}

procedure TROIFrm.HorzTrackBar1Change(Sender: TObject);
begin
   if not bAdjustingCursorPos then
   begin
       {draw}
       DrawYCursorOnBitmap; {erases prev cursor}
       DrawYCursorOnScreen; {gone!}
       yCursorPos := YBitmapToData(HorzTrackBar1.Position);
       DrawYCursorOnBitmap;
       DrawYCursorOnScreen;
       UpdateCursorCaptions;
   end;
end;


procedure TROIFrm.VertTrackBar1Change(Sender: TObject);
begin
   if not bAdjustingCursorPos then
   begin
       {draw}
       DrawXCursorOnBitmap; {erases prev cursor}
       DrawXCursorOnScreen; {gone!}
       xCursorPos := xBitmapToData(HorzTrackBar1.Position);
       DrawXCursorOnBitmap;
       DrawXCursorOnScreen;
       UpdateCursorCaptions;
   end;
end;


end.
