unit Trackcur;

interface

uses
   // Horzbaru declares ancestor of TVertTrackbar: declares methods to
   // draw ticks corresponding to grid in scope or viewer
	Windows, Messages, Classes, Graphics, Controls, Menus, ExtCtrls,
   Horzbaru;

type

   TVertTrackBar = class(TCustomTrack)
   	private
       	fMax, fMin, fPosition, fScreenPos: integer;
           fPageSize: integer;
           fSliding: boolean;
           fCursorRect: TRect;
           fCursor: TObject;

           {in screen coordinates:
           fMajorTickInterval: interval between major ticks > 0
           fFirstMajorTickPos: position of the lowest first tick mark > 0
           cMinorTickCount: number of minor ticks between each tick mark > 0}
           {fMajorTickInterval, fFirstMajorTickPos, fMinorTickCount: integer;}

           fDitherBmp, fCursorBmp, fMaskBmp, fBackgroundBmp: TBitmap;
           fOnChange: TNotifyEvent;

           function CursorToScreen( Value: integer): integer;
           function ScreenToCursor( Value: integer): integer;
           function LimitPosition( Value: integer): integer; {clips value to the control}
			procedure SetMax( Value : Integer );
			procedure SetMin( Value : Integer );
			procedure SetPosition( Value : Integer );
           function GetScreenPosition: integer;
           procedure SetScreenPosition( Value : Integer );
			procedure LoadThumbBitmaps;
			procedure UpdateDitherBitmap;
           procedure DrawTrack;
         {  procedure DrawTicks;}
           procedure DrawCursor;
			procedure WMGetDlgCode( var Msg : TWMGetDlgCode ); message wm_GetDlgCode;
			procedure CMEnabledChanged( var Msg : TMessage ); message cm_EnabledChanged;
		protected
			procedure Paint; override;
			procedure Change; dynamic;
			procedure DoEnter; override;
			procedure DoExit; override;
			procedure KeyDown( var Key : Word; Shift : TShiftState ); override;
           procedure MouseDown( Button : TMouseButton; Shift : TShiftState; X, Y : Integer ); override;
			procedure MouseMove( Shift : TShiftState; X, Y : Integer ); override;
			procedure MouseUp( Button : TMouseButton; Shift : TShiftState; X, Y : Integer ); override;
   	public
			constructor Create( AOwner : TComponent ); override;
           constructor CreateInCursor( AOwner: TComponent; theCursor: TObject);
			destructor Destroy; override;
           {procedure SetTickPos(majorInterval, majorPos, minorCount: integer);}
           function CursorVisible: boolean;
           procedure SetParams(theMax, theMin, thePos: integer);
           procedure DrawMajorTicks(pos: integer); override;
           procedure DrawMinorTicks(pos: integer); override;
   	published
       	property ScreenPos: integer read GetScreenPosition write SetScreenPosition;
			property Max : integer read fMax write SetMax default 2250;
			property Min : integer read fMin write SetMin default -2250;
           property PageSize : integer read fPageSize write fPageSize default 200;
			property Position : integer read fPosition write SetPosition;
			property OnChange : TNotifyEvent read fOnChange write fOnChange;

			{ Inherited Properties & Events }
			property Color;
			property DragCursor;
			property DragMode;
			property Enabled;
			property HelpContext;
			property Hint;
			property ParentShowHint;
			property PopupMenu;
			property ShowHint;
			property TabOrder;
			property TabStop default True;
			property Visible;

			property OnClick;
			property OnDragDrop;
			property OnDragOver;
			property OnEndDrag;
			property OnEnter;
			property OnExit;
			property OnKeyDown;
			property OnKeyPress;
			property OnKeyUp;
			property OnMouseDown;
			property OnMouseMove;
			property OnMouseUp;
  end;

  procedure Register;

implementation

uses SysUtils {, Cursorsu, ViewData};


procedure TVertTrackBar.DrawMajorTicks(pos: integer);
begin
	with Canvas do
   begin
   	Pen.Color := clBlack;
       Pen.Width := 1;
		MoveTo(1, pos);
   	LineTo(9, pos);
	end;
end;

procedure TVertTrackBar.DrawMinorTicks( pos: integer);
begin
	with Canvas do
   begin
   	Pen.Color := clBlack;
       Pen.Width := 1;
		MoveTo(1, pos);
   	LineTo(5, pos);
	end;
end;

function TVertTrackBar.CursorToScreen( Value: integer): integer;
begin
	Result := MulDiv(Value - fMax, ClientHeight - 1, fMin - fMax);
   if Result < 0 then Result := 0;
   if Result >= ClientHeight then Result := ClientHeight - 1;
end;

function TVertTrackBar.ScreenToCursor( Value: integer): integer;
begin
	Result := fMax - MulDiv(Value, fMax - fMin, ClientHeight - 1);
end;

function TVertTrackBar.LimitPosition( Value: integer): integer;
begin
	Result :=Value;
   if Result > fMax then Result := fMax;
   if Result < fMin then Result := fMin;
end;

procedure TVertTrackBar.SetMax( Value : Integer );
begin
	if value <> fMax then
   begin
   	fMax := Value;
       if fPosition > fMax then screenPos := CursorToScreen(fMax);
       fPageSize := (fMax - fMin) div 20;
       Invalidate;
   end;
end;

procedure TVertTrackBar.SetMin( Value : Integer );
begin
	if value <> fMin then
   begin
   	fMin := Value;
       if fPosition < fMin then screenPos := CursorToScreen(fMin);
       fPageSize := (fMax - fMin) div 20;
       Invalidate;
   end;
end;

procedure TVertTrackBar.SetPosition( Value : Integer );
begin
	if value <> fPosition then
   begin
       fPosition := Value;

       if csDesigning in ComponentState then
			Invalidate
		else
		begin
       	{ Erase old thumb image by drawing background bitmap }
      		Canvas.Draw( fCursorRect.Left, fCursorRect.Top, FBackgroundBmp );
			DrawCursor; { Draw thumb at new location }
			Change;  { Trigger Change event }
       end;
   end;
end;

function TVertTrackBar.GetScreenPosition;
begin
	Result := CursorToScreen(fPosition);
end;

procedure TVertTrackBar.SetScreenPosition( Value : Integer );
begin
	if Value < 0 then Value := 0;
   if Value >= ClientHeight then Value := ClientHeight - 1;
	fPosition := ScreenToCursor(Value);
   Invalidate;
end;

procedure TVertTrackBar.LoadThumbBitmaps;
const
	vCursor: PChar = 'VCURSOR';
   vMask: PChar = 'VMASK';
begin
	fCursorBmp.Handle := LoadBitmap(hInstance, vCursor);
	fMaskBmp.Handle := LoadBitmap(hInstance, vMask);
end;

procedure TVertTrackBar.UpdateDitherBitmap;
var
  i, j : integer;
begin
	with fDitherBmp.Canvas do
	begin
		Brush.Color := clWhite;
   	FillRect( Rect( 0, 0, fDitherBmp.Width, fDitherBmp.Height ) );
   	for i := 0 to 7 do
   		for j := 0 to 7 do
				if ( i + j ) mod 2 <> 0 then
					Pixels[ i, j ] := clBlack;
	end;
end;

procedure TVertTrackBar.DrawTrack;
begin
	Canvas.Brush.Color := clWhite;
   if not Enabled then
	   Canvas.Brush.Bitmap := fDitherBmp;

   Canvas.FillRect(ClientRect);
end;

{procedure TVertTrackBar.DrawTicks;
begin
	with Owner as TViewer do
   	if sweepList.Count > 0 then
				DrawYTicksOnCursorBars(self, DrawMajorTicks, DrawMinorTicks);
end;}

procedure TVertTrackBar.DrawCursor;
var
	offset : integer;
	workBmp : TBitmap;
	workRct : TRect;
begin
	fScreenPos := CursorToScreen(fPosition);
	offset := MulDiv(height - 1, fMax - ScreenToCursor(fScreenPos), fMax - fMin);
   fCursorRect := Rect(1, offset - 7, 16, offset + 8);

   fBackgroundBmp.Width := 15;
   fBackgroundBmp.Height := 15;
	fBackgroundBmp.Canvas.CopyRect( Rect(0, 0, fCursorBmp.Width, fCursorBmp.Height),
                                  Canvas, fCursorRect );
	workBmp := TBitmap.Create;
	try
		workBmp.Height := fCursorBmp.Height;
		workBmp.Width := fCursorBmp.Width;

		workRct := Rect( 0, 0, fCursorBmp.Width, fCursorBmp.Height);

		workBmp.Canvas.CopyMode := cmSrcCopy;
		workBmp.Canvas.CopyRect( WorkRct, fBackgroundBmp.Canvas, workRct );

		workBmp.Canvas.CopyMode := cmSrcAnd;
		workBmp.Canvas.CopyRect( WorkRct, fMaskBmp.Canvas, WorkRct );

		workBmp.Canvas.CopyMode := cmSrcPaint;
       WorkBmp.Canvas.CopyRect( WorkRct, fCursorBmp.Canvas, WorkRct );

		if not Enabled then
		begin
			WorkBmp.Canvas.Brush.Bitmap := fDitherBmp;
			WorkBmp.Canvas.FloodFill( WorkRct.Right - 3, WorkRct.Bottom - 3,
                                clSilver, fsSurface );
		end;

		Canvas.CopyRect( fCursorRect, WorkBmp.Canvas, WorkRct );
   finally
			workBmp.Free;
	end;
end;


procedure TVertTrackBar.WMGetDlgCode( var Msg : TWMGetDlgCode );
begin
	inherited;
	Msg.Result := dlgc_WantArrows;
end;

procedure TVertTrackBar.CMEnabledChanged( var Msg : TMessage );
begin
  inherited;
  Invalidate;
end;

procedure TVertTrackBar.Paint;
begin
	with Canvas do
   begin
       DrawTrack;
       DrawTicks;
       DrawCursor;
   end;
end;

procedure TVertTrackBar.Change;
begin
	if Assigned( FOnChange ) then FOnChange( Self );
end;

procedure TVertTrackBar.DoEnter;
begin
	inherited DoEnter;
	Refresh;
end;

procedure TVertTrackBar.DoExit;
begin
	inherited DoExit;
	Refresh;
end;

procedure TVertTrackBar.KeyDown( var Key : Word; Shift : TShiftState );
begin
	inherited KeyDown( Key, Shift );
	case Key of
   	vk_Prior:
       	Position := LimitPosition(fPosition + fPageSize);
		vk_Next:
			Position := LimitPosition(fPosition - FPageSize);
		vk_End:
			Position := fMin;
		vk_Home:
			Position := fMax;
		vk_Left:
			if fPosition > fMin then Position := LimitPosition(fPosition - 1);
		vk_Up:
       	if fPosition < fMax then Position := LimitPosition(fPosition + 1);
		vk_Right:
			if fPosition < fMax then Position := LimitPosition(fPosition + 1);
		vk_Down:
			if fPosition > fMin then Position := LimitPosition(fPosition - 1);
       $30 {VK_0}, $60 {vk_numpad0}:
       	Position := 0;
	end;
end;

procedure TVertTrackBar.MouseDown( Button : TMouseButton; Shift : TShiftState; X, Y : Integer );
var
  ptY : integer;
begin
	inherited MouseDown( Button, Shift, X, Y );
	SetFocus;

	if ( Button = mbLeft ) and PtInRect( fCursorRect, Point( X, Y ) ) then
		fSliding := True
	else
	begin
   	ptY := height - 1 - MulDiv(fPosition - fMin, height, fMax - fMin);
       if Y < PtY then
       	Position := LimitPosition(fPosition + fPageSize)
   	else
      		Position := LimitPosition(FPosition - FPageSize);
	end;
end;

procedure TVertTrackBar.MouseMove( Shift : TShiftState; X, Y : Integer );
var  p, h : Integer;
begin
	inherited MouseMove( Shift, X, Y );

	if PtInRect( FCursorRect, Point( X, Y ) ) then
   	Cursor := crSizeNS
	else
   	Cursor := crDefault;
	if fSliding then
	begin
   	h := Height - 7;
       p:= MulDiv(h - y , fMax - fMin, h) + fMin;
		if p > fMax then p := fMax;
       if p < fMin then p := fMin;
   	Position := p;
   end;
end;

procedure TVertTrackBar.MouseUp( Button : TMouseButton; Shift : TShiftState; X, Y : Integer );
begin
	inherited MouseUp( Button, Shift, X, Y );
	if ( Button = mbLeft ) then fSliding := False;
end;

constructor TVertTrackBar.CreateInCursor( AOwner: TComponent; theCursor: TObject);
begin
	Create(AOwner);
   fCursor := theCursor;
end;


constructor TVertTrackBar.Create( AOwner : TComponent );
begin
	inherited Create( AOwner );
	Width := 17;
	Height := 150;
	fMin := -2250;
	fMax := +2250;
	fPosition := 0;
	fPageSize := 200;
	fSliding := False;

	fCursorBmp := TBitmap.Create;
  	fCursorBmp.Width := 16;
	fCursorBmp.Height := 16;
	fMaskBmp := TBitmap.Create;
	fBackgroundBmp := TBitmap.Create;
   fDitherBmp := TBitmap.Create;
  	fDitherBmp.Width := 8;
	fDitherBmp.Height := 8;
	UpdateDitherBitmap;
	LoadThumbBitmaps;
end;

destructor TVertTrackBar.Destroy;
begin
 	FreeAndNil(fCursorBmp);
	FreeAndNil(fMaskBmp);
	FreeAndNil(fBackgroundBmp);
   FreeAndNil(fDitherBmp);
   inherited Destroy;
end;

function TVertTrackBar.CursorVisible: boolean;
begin
	if (fPosition <= fMax) and (fPosition >= fMin) then Result := True else Result := False;
end;

procedure TVertTrackBar.SetParams(theMax, theMin, thePos: integer);
begin
	fMax := theMax;
   fMin := theMin;
   fPosition := thePos;
   fPageSize := (fMax - fMin) div 20;
end;

procedure Register;
begin
	RegisterComponents( 'Samples', [TVertTrackBar] );
end;


end.
