unit analogu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, ToolWin, ComCtrls, ImgList, ExtCtrls, mpfileu, mpviewu, StdCtrls,
  Spin;

type
	TDataBuffer = array of int16;

  TAnalogFrm = class(TForm)
    StatusBar1: TStatusBar;
    MainMenu1: TMainMenu;
    Edit1: TMenuItem;
    CopytoBitmap1: TMenuItem;
    Panel1: TPanel;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton3: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    ToolButton6: TToolButton;
    ImageList1: TImageList;
    N1: TMenuItem;
    Axes1: TMenuItem;
    Frame1: TMenuItem;
    GotoFrame1: TMenuItem;
    LastFrame1: TMenuItem;
    FirstFrame1: TMenuItem;
    N3: TMenuItem;
    FastReverse1: TMenuItem;
    Stop1: TMenuItem;
    FastForward1: TMenuItem;
    N4: TMenuItem;
    PrevFrame1: TMenuItem;
    NextFrame1: TMenuItem;
    CopyDatatoClipboard1: TMenuItem;
    N2: TMenuItem;
    NewAnalogChannelsWindow1: TMenuItem;
    FramesinView1: TMenuItem;
    Label1: TLabel;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormResize(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
    procedure ToolButton2Click(Sender: TObject);
    procedure ToolButton3Click(Sender: TObject);
    procedure ToolButton4Click(Sender: TObject);
    procedure ToolButton5Click(Sender: TObject);
    procedure FirstFrame1Click(Sender: TObject);
    procedure LastFrame1Click(Sender: TObject);
    procedure GotoFrame1Click(Sender: TObject);
    procedure CopytoBitmap1Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure CopyDatatoClipboard1Click(Sender: TObject);
    procedure Axes1Click(Sender: TObject);
    procedure NewAnalogChannelsWindow1Click(Sender: TObject);
    procedure FramesinView1Click(Sender: TObject);
  private
    { Private declarations }
    bInitialized, bFormClosing: boolean; {prevents events from firing when form is created}
    cFramesInView,
    firstFrameInBuffer,
    fCurrentFrameIndex: integer;
    traces: array[2..3] of TTrace;
    f_MAX_Y_VALUE: integer;
    ch2Rect, ch3Rect: TRect;
    offscreenBitmap: TBitmap;
    cPtsInBuffers: array[2..3] of integer;
    dataBuffers: array[2..3] of TDataBuffer;
    function  AnalogXToTime(chIndex, sampleIndex: integer): string; {in ms or us}
    function  AnalogYToValue(chIndex, sampleValue: integer): string; {in units}
    procedure DrawBackground;
    procedure DrawTraces;
    procedure LoadAnalogDataFromFile;
    procedure ResizeElements;
    procedure SetBufferLength;
	 procedure SetCurrentFrameIndex(newIndex: integer);
    function  YToBitmap(chIndex, y: integer): integer;
  public
    { Public declarations }
    mpFile: TMPFile;
    yTops, yExtents: array[2..3] of integer;

    procedure DrawAnalogData;
    procedure OnNewFrame;
    procedure Initialize(thempFile: TMPFile);
    procedure SaveTrace(chIndex: integer; const fname: string);

    property  CurrentFrameIndex: integer read fCurrentFrameIndex write SetCurrentFrameIndex;
  end;

var
  AnalogFrm: TAnalogFrm;

implementation

{$R *.DFM}

uses mainfrm, anyaxisu, cpyanaldlgu, clipbrd;


procedure TAnalogFrm.Initialize(thempFile: TMPFile);
var	bCaptionOK: boolean;
		i, j: integer;
       s: string;
begin
   mpFile := thempFile;
	bCaptionOK := False; i := 0;

	cFramesInView := 1;
   SetBufferLength;

   f_MAX_Y_VALUE := mpFile.MaxPixelValue;
   yTops[2] := f_MAX_Y_VALUE; yTops[3] := yTops[2];
   yExtents[2] := 2 * (f_MAX_Y_VALUE + 1); yExtents[3] := yExtents[2];
   mpFile.ActiveFrameIndex := fCurrentFrameIndex;

   {Sets caption of the form}
   while not bCaptionOK do
   begin
   	i := i + 1;
       s := ExtractFileName(mpFile.filename) + '; Analog Channels - ' +  IntToStr(i);
       if mpFile.analogWndList.Count <> 0 then
       begin
       	bCaptionOK := True;
       	for j := 0 to mpFile.analogWndList.Count - 1 do
           	if TAnalogFrm(mpFile.analogWndList.Items[j]).Caption = s then
               	bCaptionOK := False;
       end
       else
       	bCaptionOK := True;
   end;
   Caption := s;
   bInitialized := True;
end;


procedure  TAnalogFrm.SetCurrentFrameIndex(newIndex: integer);
begin
   if (newIndex < 0) or (newIndex >= mpFile.FrameCount) then Exit;
   fCurrentFrameIndex := newIndex;
   mpFile.ActiveFrameIndex := fCurrentFrameIndex;
   OnNewFrame;
end;


function TAnalogFrm.AnalogXToTime(chIndex, sampleIndex: integer): string; {in ms or us}
var    dTime: double;
begin
   with mpFile do
   try
       dtime := (sampleIndex / ChDataPtsPerFrame[chIndex] + CurrentFrameIndex) / FrameRate;
       if dtime = 0 then
           Result := '0 ms'
       else if Abs(dtime) > 1E-3 then
           Result := Format('%8.3f', [dtime * 1000]) + ' ms'
       else
           Result := Format('%8.3f', [dtime * 1000000]) + ' ' + Chr(181) + 's';
   except
       Result := '';
   end;
end;


function TAnalogFrm.AnalogYToValue(chIndex, sampleValue: integer): string; {in units}
var    dValue: double;
begin
   with mpFile do
   try
       dValue := ChConvFactor[chIndex] * FullScaleToVal(ChInputRange[chIndex]) *
			sampleValue / (f_MAX_Y_VALUE + 1) + ChOffset[chIndex];

       Result := Format('%8.3g', [dValue]) + ' ' + PrefixToString(ChPrefix[chIndex]) +
           ChUnit[chIndex];
   except
       Result := '';
   end;
end;


procedure TAnalogFrm.DrawBackground;
begin
   with offscreenBitmap.Canvas, mpFile do
   begin
       Brush.Color := clBlack;
       FillRect(Rect(0, 0, ClientWidth - 1, ClientHeight - StatusBar1.Height - 1));
       if AnalogChEnabled[2] and AnalogChEnabled[3] then
       begin
           Pen.Color := clRed;
           MoveTo(ClientWidth div 2, 0);
           LineTo(ClientWidth div 2, offscreenBitmap.Height - 1);
       end;
   end;
end;


procedure TAnalogFrm.DrawTraces;
var chIndex, i, j, k, l, traceSize, traceLength: integer;
begin
   offscreenBitmap.Canvas.Pen.Color := clWhite;
   with mpFile do
   begin
   	LoadAnalogDataFromFile; {from CurrentFrameIndex .. + cFramesInView - 1}
       for chIndex := 2 to 3 do
       	if AnalogChEnabled[chIndex] then
       	begin
           	traceSize := Length(traces[chIndex]);
           	traceLength := 0;

           	for i := 0 to traceSize - 1 do
           	begin
               	j := MulDiv(i, cFramesInView * ChDataPtsPerFrame[chIndex] - 1, traceSize - 1);
               	if j < cPtsInBuffers[chIndex] then
               	begin
               		traces[chIndex][i].MaxPt.y := dataBuffers[chIndex][j];
               		traces[chIndex][i].MinPt.y := dataBuffers[chIndex][j];
                   	traceLength := traceLength + 1;

               		k := MulDiv(i + 1, cFramesInView * ChDataPtsPerFrame[chIndex] - 1, traceSize - 1) - 1;

               		if k >= cPtsInBuffers[chIndex] then k := cPtsInBuffers[chIndex] - 1;

               		if j < k then
                   		for l := j + 1 to k do
                   		begin
                       		if dataBuffers[chIndex][l] > traces[chIndex][i].MaxPt.y then traces[chIndex][i].MaxPt.y := dataBuffers[chIndex][l];
                       		if dataBuffers[chIndex][l] < traces[chIndex][i].MinPt.y then traces[chIndex][i].MinPt.y := dataBuffers[chIndex][l];
                   		end;

                       traces[chIndex][i].MaxPt.y := YToBitmap(chIndex, traces[chIndex][i].MaxPt.y);
                       traces[chIndex][i].MinPt.y := YToBitmap(chIndex, traces[chIndex][i].MinPt.y);
               	end;
           	end;

           	if traceLength > 0 then
           		Windows.PolyLine(offscreenBitmap.Canvas.Handle, traces[chIndex][0].MaxPt, 2 * (traceLength - 1));
       	end;
       end;
end;


procedure TAnalogFrm.LoadAnalogDataFromFile;
var	i, j, k, l, m,
		cdppf, {cached value}
		firstIndex, cFramesInBuffer, commonFrameStart, commonFrameEnd: integer;
		cFramesToLoad: array[2..3] of integer;
begin
   with mpFile do
   	for i := 2 to 3 do
   		if AnalogChEnabled[i] then
       	begin
           	cdppf := ChDataPtsPerFrame[i];
           	if cFramesInView + CurrentFrameIndex <= FrameCount then
           		cPtsInBuffers[i] := cFramesInView * cdppf
           	else
					cPtsInBuffers[i] := (FrameCount - CurrentFrameIndex)* cdppf;

				if firstFrameInBuffer = -1 then
   			begin
               	firstIndex := 0;
               	cFramesToLoad[i] := cPtsInBuffers[i] div cdppf;
   			end
   			else
   			begin
               	cFramesInBuffer := Length(dataBuffers[i]) div cdppf;
                   if FindCommonRegion(
                   	firstFrameInBuffer,
                       firstFrameInBuffer + cFramesInBuffer - 1,
                       CurrentFrameIndex,
                       CurrentFrameIndex + cPtsInBuffers[i] div cdppf - 1,
						commonFrameStart,
                       commonFrameEnd) then
                   begin
   					{push data}
                       if firstFrameInBuffer < CurrentFrameIndex then
                       begin
                           for l := commonFrameStart - firstFrameInBuffer to commonFrameEnd - firstFrameInBuffer do
                           		for m := 0 to cdppf - 1 do
										dataBuffers[i][(l - (commonFrameStart - firstFrameInBuffer))* cdppf + m] :=
                                       	dataBuffers[i][l * cdppf + m];
                           firstIndex := (commonFrameEnd - commonFrameStart + 1) * cdppf;
                           cFramesToLoad[i] := (cPtsInBuffers[i] div cdppf) - (commonFrameEnd - commonFrameStart + 1);
                       end
                       else if firstFrameInBuffer > CurrentFrameIndex then
                       begin
                           firstIndex := 0;
                           cFramesToLoad[i] := (cPtsInBuffers[i] div cdppf) - (commonFrameEnd - commonFrameStart + 1);
                           for l := commonFrameEnd - firstFrameInBuffer downto commonFrameStart - firstFrameInBuffer do
                           		for m := cdppf - 1 downto 0 do
										dataBuffers[i][(l + cFramesToLoad[i])* cdppf + m] :=
                                       	dataBuffers[i][l * cdppf + m];
                       end
                       else  {=}
                       begin
                       	firstIndex := 0;
                           cFramesToLoad[i] := 0;
                       end;
                   end
					else
                   begin
                   	{no region in common; we load the whole buffer}
                   	firstIndex := 0;
                       cFramesToLoad[i] := cPtsInBuffers[i] div cdppf;
                   end;
   			end;

				if cFramesToLoad[i] > 0 then
   				for j := 0 to cFramesToLoad[i] - 1 do
               		with mpFile.Frames[CurrentFrameIndex + j].channels[i] do
                   		for k := 0 to cdppf - 1 do
                       		dataBuffers[i][firstIndex + j * cdppf + k] := data[k];
   		end;
   firstFrameInBuffer := CurrentFrameIndex;
end;



procedure TAnalogFrm.ResizeElements;
var    i: integer;
begin
   offscreenBitmap.Width := ClientWidth;
   offscreenBitmap.Height := ClientHeight - StatusBar1.Height - ToolBar1.Height;
   with mpFile do
   begin
       if AnalogChEnabled[2] and not AnalogChEnabled[3] then
       begin
           ch2Rect := Rect(0, Panel1.Height, ClientWidth - 1,
           	ClientHeight - StatusBar1.Height {- Panel1.Height} - 1);
           SetLength(traces[2], ClientWidth);
           for i := 0 to ClientWidth - 1 do
           begin
               traces[2][i].MaxPt.x := i;
               traces[2][i].MinPt.x := i;
           end;
       end;
       if not AnalogChEnabled[2] and AnalogChEnabled[3] then
       begin
           ch3Rect := Rect(0, Panel1.Height, ClientWidth - 1,
           	ClientHeight - StatusBar1.Height {- Panel1.Height} - 1);
           SetLength(traces[3], ClientWidth);
           for i := 0 to ClientWidth - 1 do
           begin
               traces[3][i].MaxPt.x := i;
               traces[3][i].MinPt.x := i;
           end;
       end;
       if AnalogChEnabled[2] and AnalogChEnabled[3] then
       begin
           ch2Rect := Rect(0, Panel1.Height, ClientWidth div 2 - 1,
           	ClientHeight - StatusBar1.Height {- Panel1.Height} - 1);
           ch3Rect := Rect(ClientWidth div 2 + 1, Panel1.Height, ClientWidth - 1,
           	ClientHeight - StatusBar1.Height {- Panel1.Height} - 1);
           SetLength(traces[2], ClientWidth div 2);
           for i := 0 to (ClientWidth - 1) div 2 - 1 do
           begin
               traces[2][i].MaxPt.x := i;
               traces[2][i].MinPt.x := i;
           end;
           SetLength(traces[3], ClientWidth - ClientWidth div 2 - 1);
           for i := ClientWidth div 2 + 1 to ClientWidth - 1 do
           begin
               traces[3][i - (ClientWidth div 2 + 1)].MaxPt.x := i;
               traces[3][i - (ClientWidth div 2 + 1)].MinPt.x := i;
           end;
       end;
   end;
end;


procedure TAnalogFrm.SetBufferLength;
var	i: integer;
begin
	for i := 2 to 3 do
		if mpFile.AnalogChEnabled[i] then
   		SetLength(dataBuffers[i], cFramesInView * mpFile.ChDataPtsPerFrame[i])
   	else
   		SetLength(dataBuffers[i], 0);
   firstFrameInBuffer := -1;
end;


function  TAnalogFrm.YToBitmap(chIndex, y: integer): integer;
begin
   Result := Muldiv(y - yTops[chIndex], offscreenBitmap.Height - 1, - yExtents[chIndex] + 1);
   if Result < 0 then Result := 0;
   if Result > offscreenBitmap.Height - 1 then Result := offscreenBitmap.Height - 1;
end;


procedure  TAnalogFrm.OnNewFrame;
begin
	DrawAnalogData;

	StatusBar1.Panels[0].Text := 'Frame: ' + IntToStr(CurrentFrameIndex + 1) +
   	'/' + IntToStr(mpFile.FrameCount);
end;


{********************************* PUBLIC *************************************}

procedure TAnalogFrm.DrawAnalogData;
var    rc: TRect;
begin
   DrawBackground;
   DrawTraces;
   rc := Rect(0, 0, offscreenBitmap.Width - 1, offscreenBitmap.Height - 1);
   Canvas.CopyRect(rc, offscreenBitmap.Canvas,
   	Rect(rc.Left, rc.Top - ToolBar1.Height, rc.Right, rc.Bottom - ToolBar1.Height));
end;


{****************************** FORM EVENTS ***********************************}

procedure TAnalogFrm.FormCreate(Sender: TObject);
begin
   offscreenBitmap := TBitmap.Create;
   offscreenBitmap.HandleType := bmDDB;
   PrevFrame1.ShortCut := ShortCut(VK_LEFT, []);
   NextFrame1.ShortCut := ShortCut(VK_RIGHT, []);
end;


procedure TAnalogFrm.FormShow(Sender: TObject);
begin
	if bInitialized then ResizeElements;
end;


procedure TAnalogFrm.FormActivate(Sender: TObject);
begin
   with Mainform do
   begin
       NewFile1.Enabled := True;
       OpenFile1.Enabled := True;
       if mpFile <> nil then
           FileAs1.Enabled := mpFile.IsMemoryFile
       else
           FileAs1.Enabled := False;
       FileInformation1.Enabled := True;
       Close1.Enabled := True;
   end;
end;


procedure TAnalogFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   action := caFree;
   bFormClosing := True;
	try
	   	if not Mainform.bAppClosing then mpFile.OnWndClose(self);
   finally
   	mpFile := nil;
   end;
end;


procedure TAnalogFrm.FormDestroy(Sender: TObject);
begin
   offscreenBitmap.Free;
end;


procedure TAnalogFrm.FormPaint(Sender: TObject);
var    rc: TRect;
begin
	if bFormClosing then Exit;
   rc := Canvas.ClipRect;
   with mpFile do
       if AnalogChEnabled[2] or AnalogChEnabled[3] then
       begin
           DrawBackground;
           DrawTraces;
           Canvas.CopyRect(rc, offscreenBitmap.Canvas,
           	Rect(rc.Left, rc.Top - ToolBar1.Height, rc.Right, rc.Bottom - ToolBar1.Height));
       end
       else
       begin
           Canvas.Brush.Color := clBlack;
           Canvas.FillRect(Rect(0, 0, ClientWidth - 1, ClientHeight - StatusBar1.Height - 1));
       end;
end;


procedure TAnalogFrm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var    currentCh: integer;
       iX, iY: integer;
begin
   currentCh := 3; iX := 0; iY := 0;

   with mpFile do
       if not AnalogChEnabled[2] and not AnalogChEnabled[3] then Exit;

   with mpFile do
   begin
       if AnalogChEnabled[2] and not AnalogChEnabled[3] then
       begin
           currentCh := 3;
           iX := MulDiv(X, cFramesInView * ChDataPtsPerFrame[2] - 1, ch2Rect.Right - ch2Rect.Left);
           iY := MulDiv(Y - ch2Rect.Top, - yExtents[2] - 1, ch2Rect.Bottom - ch2Rect.Top)
               +  yTops[2];
       end;

       if not AnalogChEnabled[2] and AnalogChEnabled[3] then
       begin
           currentCh := 4;
           iX := MulDiv(X, cFramesInView * ChDataPtsPerFrame[3] - 1, ch3Rect.Right - ch3Rect.Left);
           iY := MulDiv(Y - ch3Rect.Top, - yExtents[3] - 1, ch3Rect.Bottom - ch3Rect.Top)
               +  yTops[2];
       end;

       if AnalogChEnabled[2] and AnalogChEnabled[3] then
       begin
           if X < ClientWidth div 2 then
               currentCh := 3
           else
           begin
               currentCh := 4;
               X := X - ClientWidth div 2 - 1;
           end;

           if currentCh = 3 then
           begin
               iX := MulDiv(X, cFramesInView * ChDataPtsPerFrame[2] - 1, ch2Rect.Right - ch2Rect.Left);
               iY := MulDiv(Y - ch2Rect.Top, - yExtents[2] - 1, ch2Rect.Bottom - ch2Rect.Top)
                   +  yTops[2];
           end
           else
           begin
               iX := MulDiv(X, cFramesInView * ChDataPtsPerFrame[3] - 1, ch3Rect.Right - ch3Rect.Left);
               iY := MulDiv(Y - ch3Rect.Top, - yExtents[3] - 1, ch3Rect.Bottom - ch3Rect.Top)
                   +  yTops[3];
           end;
       end;

       StatusBar1.Panels[1].Text := 'Analog Ch' + IntToStr(currentCh - 2) + ': ' + mpFile.ChNames[currentCh - 1];
       StatusBar1.Panels[2].Text := 'T: ' + AnalogXToTime(currentCh - 1 {zero-based}, iX) +
       	' (Frame ' + IntToStr(iX div mpFile.ChDataPtsPerFrame[currentCh - 1] + CurrentFrameIndex + 1) + ')';
       StatusBar1.Panels[3].Text := 'Analog: ' + AnalogYToValue(currentCh - 1 {zero-based}, iY);
   end;
end;


procedure TAnalogFrm.FormResize(Sender: TObject);
begin
   ResizeElements;
   Invalidate; {necessary?}
	StatusBar1.Panels[0].Text := 'Frame: ' + IntToStr(CurrentFrameIndex + 1) +
   	'/' + IntToStr(mpFile.FrameCount);
end;

{-------------------------------- MENU EVENTS ---------------------------------}

procedure TAnalogFrm.ToolButton1Click(Sender: TObject);
begin
   while (CurrentFrameIndex > 0) and (ToolButton1.Down) do
   begin
       CurrentFrameIndex := CurrentFrameIndex - 1;
       Application.ProcessMessages;
   end;
   ToolButton1.Down := False;
end;


procedure TAnalogFrm.ToolButton2Click(Sender: TObject);
begin
   CurrentFrameIndex := CurrentFrameIndex - 1;
end;


procedure TAnalogFrm.ToolButton3Click(Sender: TObject);
begin
   ToolButton1.Down := False;
   ToolButton5.Down := False;
end;


procedure TAnalogFrm.ToolButton4Click(Sender: TObject);
begin
   CurrentFrameIndex := CurrentFrameIndex + 1;
end;


procedure TAnalogFrm.ToolButton5Click(Sender: TObject);
begin
   while (CurrentFrameIndex < mpFile.FrameCount - 1) and (ToolButton5.Down) do
   begin
       CurrentFrameIndex := CurrentFrameIndex + 1;
       Application.ProcessMessages;
   end;
   ToolButton5.Down := False;
end;


procedure TAnalogFrm.FirstFrame1Click(Sender: TObject);
begin
	CurrentFrameIndex := 0;
end;


procedure TAnalogFrm.LastFrame1Click(Sender: TObject);
begin
	CurrentFrameIndex := mpFile.FrameCount - 1;
end;


procedure TAnalogFrm.GotoFrame1Click(Sender: TObject);
var	s: string;
		newFrameIndex: integer;
begin
	s := '1';
	if InputQuery('Go to Frame', 'Enter Frame Index', s) then
   try
		newFrameIndex := StrToInt(s);
       if (newFrameIndex > 0) and (newFrameIndex <= mpFile.FrameCount) then
       	CurrentFrameIndex := newFrameIndex - 1
       else
			MessageDlg('Frame Index Out of Bounds.', mtError, [mbOK], 0);
   except
   	MessageDlg('Invalid Frame Index.', mtError, [mbOK], 0);
   end;
end;


procedure TAnalogFrm.CopytoBitmap1Click(Sender: TObject);
begin
	with Clipboard do
   begin
   	Open;
       Assign(offscreenBitmap);
       Close;
   end;
end;


procedure TAnalogFrm.CopyDatatoClipboard1Click(Sender: TObject);
var	fromFrame, toFrame: integer;
		savedCursor: TCursor;
begin
	with CopyAnalogDlg do
   begin
   	RadioButton1.Checked := mpFile.AnalogChEnabled[2];
   	RadioButton1.Enabled := mpFile.AnalogChEnabled[2];
   	RadioButton2.Checked := mpFile.AnalogChEnabled[3];
   	RadioButton2.Enabled := mpFile.AnalogChEnabled[3];

       SpinEdit1.Value := CurrentFrameIndex + 1;
       SpinEdit2.Value := CurrentFrameIndex + 1;

       if CopyAnalogDlg.ShowModal = mrOK then
       	begin
       		fromFrame := SpinEdit1.Value - 1;
           	toFrame := SpinEdit2.Value - 1;
				if 	(fromFrame >= 0) and (fromFrame <= mpFile.FrameCount - 1) and
               (toFrame >= 0) and (toFrame <= mpFile.FrameCount - 1) and
               	(fromFrame <= toFrame) then
                   begin
                   	savedCursor := Screen.Cursor;
                       try
                       	Screen.Cursor := crHourGlass;
           				mpFile.CopyChannelsToClipboard(
                   			RadioButton1.Checked and RadioButton1.Enabled,
                   			RadioButton2.Checked and RadioButton2.Enabled,
                   		fromFrame, toFrame);
                       finally
                       	Screen.Cursor := savedCursor;
                       end;
                   end
               else
               	MessageDlg('Invalid frame indices.', mtError, [mbOK], 0);
       	end;
   end;
end;


procedure TAnalogFrm.Axes1Click(Sender: TObject);
var	newYTop2, newYTop3, newYExtent2, newYExtent3: integer;
		bAccepted: boolean;
begin
	newyTop2 := 0;
	newyTop3 := 0;
	newyExtent2 := 0;
	newyExtent3 := 0;

	with AnYAxisDlg do
   begin
   	PhysUnitEdit1.Clip12Bits := (f_MAX_Y_VALUE < 32768);
       PhysUnitEdit2.Clip12Bits := (f_MAX_Y_VALUE < 32768);
       PhysUnitEdit3.Clip12Bits := (f_MAX_Y_VALUE < 32768);
       PhysUnitEdit4.Clip12Bits := (f_MAX_Y_VALUE < 32768);

   	if mpFile.AnalogChEnabled[2] then
       begin
       	PhysUnitEdit1.PhysUnit := mpFile.ChUnit[2];
       	PhysUnitEdit1.ConvFactor := mpFile.ChConvFactor[2] * FullScaleToVal(mpFile.ChInputRange[2]);
       	PhysUnitEdit1.Prefix := PrefixToExponent(mpFile.ChPrefix[2]);
       	PhysUnitEdit1.PhysOffset := mpFile.ChOffset[2];

       	PhysUnitEdit2.PhysUnit := mpFile.ChUnit[2];
       	PhysUnitEdit2.ConvFactor := mpFile.ChConvFactor[2] * FullScaleToVal(mpFile.ChInputRange[2]);
       	PhysUnitEdit2.Prefix := PrefixToExponent(mpFile.ChPrefix[2]);
       	PhysUnitEdit2.PhysOffset := mpFile.ChOffset[2];

       	PhysUnitEdit1.int16Value := yTops[2];
           PhysUnitEdit2.int16Value := yTops[2] - yExtents[2] + 1;
       end;

       if mpFile.AnalogChEnabled[3] then
       begin
       	PhysUnitEdit3.PhysUnit := mpFile.ChUnit[3];
       	PhysUnitEdit3.ConvFactor := mpFile.ChConvFactor[3] * FullScaleToVal(mpFile.ChInputRange[3]);
       	PhysUnitEdit3.Prefix := PrefixToExponent(mpFile.ChPrefix[3]);
       	PhysUnitEdit3.PhysOffset := mpFile.ChOffset[3];

       	PhysUnitEdit4.PhysUnit := mpFile.ChUnit[3];
       	PhysUnitEdit4.ConvFactor := mpFile.ChConvFactor[3] * FullScaleToVal(mpFile.ChInputRange[3]);
       	PhysUnitEdit4.Prefix := PrefixToExponent(mpFile.ChPrefix[3]);
       	PhysUnitEdit4.PhysOffset := mpFile.ChOffset[3];

       	PhysUnitEdit3.int16Value := yTops[3];
           PhysUnitEdit4.int16Value := yTops[3] - yExtents[3] + 1;
       end;

   	TabSheet1.Enabled := mpFile.AnalogChEnabled[2];
   	TabSheet2.Enabled := mpFile.AnalogChEnabled[3];
       if not mpFile.AnalogChEnabled[2] then PageControl1.ActivePage := TabSheet2
       	else PageControl1.ActivePage := TabSheet1;
   end;

   if AnYAxisDlg.ShowModal = mrOK then
   with AnYAxisDlg do
   begin
		bAccepted := True;

   	if mpFile.AnalogChEnabled[2] then
       begin
       	newyTop2 := PhysUnitEdit1.int16Value;
           newyExtent2 := newyTop2 - PhysUnitEdit2.int16Value + 1;
           if (newyTop2 > f_MAX_Y_VALUE) or (newyTop2 < -(f_MAX_Y_VALUE + 1)) or
           	(newyExtent2 < 10) or (newyExtent2 > 2 * (f_MAX_Y_VALUE + 1)) then
               bAccepted := False;
       end;

       if mpFile.AnalogChEnabled[3] then
       begin
       	newyTop3 := PhysUnitEdit3.int16Value;
           newyExtent3 := newyTop3 - PhysUnitEdit4.int16Value + 1;
           if (newyTop3 > f_MAX_Y_VALUE) or (newyTop3 < -(f_MAX_Y_VALUE + 1)) or
           	(newyExtent3 < 10) or (newyExtent3 > 2 * (f_MAX_Y_VALUE + 1)) then
               bAccepted := False;
       end;

   	if bAccepted then
   	begin
   		yTops[2] := newyTop2;
   		yTops[3] := newyTop3;
   		yExtents[2] := newyExtent2;
   		yExtents[3] := newyExtent3;
   		DrawAnalogData;
   	end
   	else
   		MessageDlg('Invalid Y Axes Values.', mtError, [mbOK], 0);
   end;

end;



procedure TAnalogFrm.NewAnalogChannelsWindow1Click(Sender: TObject);
begin
	mpFile.NewAnalogWnd;
end;


procedure TAnalogFrm.SaveTrace(chIndex: integer; const fname: string);
begin
end;


procedure TAnalogFrm.FramesinView1Click(Sender: TObject);
var	sFramesInView: string;
		newcFramesInView: integer;
begin
	sFramesInView := IntToStr(cFramesInView);
	if InputQuery('Frames in View', 'Enter number of frames for which to display analog data', sFramesInView) then
	try
   	newcFramesInView := StrToInt(sFramesInView);
       if (newcFramesInView >= 1) and (newcFramesInView <= mpFile.FrameCount) then
       begin
       	if cFramesInView <> newcFramesInView then
           begin
				cFramesInView := newcFramesInView;
               SetBufferLength;
               Invalidate;
               if cFramesInView = 1 then
               	Label1.Caption := IntToStr(cFramesInView) + ' frame in view'
               else
               	Label1.Caption := IntToStr(cFramesInView) + ' frames in view'
           end;
       end
       else
			MessageDlg('Invalid number of frames', mtError, [mbOK], 0);
   except
		MessageDlg('Invalid number of frames', mtError, [mbOK], 0);
   end;
end;

end.
