unit Gammafrmu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, ComCtrls, ExtCtrls, mpFileu;

type
  TGammaFrm = class(TForm)
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    CheckBox1: TCheckBox;
    TrackBar1: TTrackBar;
    TrackBar2: TTrackBar;
    Label1: TLabel;
    Label2: TLabel;
    Bevel1: TBevel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    RadioButton1: TRadioButton;
    RadioButton2: TRadioButton;
    RadioButton3: TRadioButton;
    RadioButton4: TRadioButton;
    procedure FormPaint(Sender: TObject);
    procedure TrackBar1Change(Sender: TObject);
    procedure TrackBar2Change(Sender: TObject);
    procedure RadioButton1Click(Sender: TObject);
  private
    { Private declarations }
    histoRect: TRect;
    oldBlackSliderPos, oldWhiteSliderPos: integer;
    procedure DrawBlackLevelLine(sliderPos: integer);
    procedure DrawHistogram;
    procedure DrawWhiteLevelLine(sliderPos: integer);
  public
    { Public declarations }
    bInitializing: boolean;
    mpFile: TMPFile;
    procedure	InitGUI(afile: TMPFile);
    function   SelectedCh: integer;
  end;

var
  GammaFrm: TGammaFrm;

implementation

{$R *.DFM}
const
   BORDER_OFFSET = 12;
   BIN_SIZE = 128; {for histogram}

procedure TGammaFrm.DrawBlackLevelLine(sliderPos: integer);
var    linePos: integer;
begin
   Canvas.Pen.Color := clRed;
   Canvas.Pen.Width := 1;
   Canvas.Pen.Style := psSolid;
   Canvas.Pen.Mode := pmXor;
   linePos := histoRect.Left + Muldiv(sliderPos, histoRect.Right - histoRect.Left,
       2047 + 128);
   Canvas.MoveTo(linePos, histoRect.Top);
   Canvas.LineTo(linePos, histoRect.Bottom);
   oldBlackSliderPos := sliderPos;
end;


procedure TGammaFrm.DrawHistogram;
var    histogram: array of integer;
       chIndex, i, maxValue, barValue, leftRect, rightRect: integer;
       barRect: TRect;
begin
   Canvas.Brush.Color := clWhite;
   Canvas.FillRect(histoRect);

   // Get the histogram
   chIndex := SelectedCh;
   SetLength(histogram, 2048 div BIN_SIZE + 1);
   (mpFile.Frames[mpFile.ActiveFrameIndex].channels[chIndex] as TVideoFrame).
       GetProfile(BIN_SIZE, histogram);

   maxValue := 0;
   for i := 0 to 2048 div BIN_SIZE do
       if histogram[i] > maxValue then maxValue := histogram[i];

   // Draws the rectangles
   Canvas.Brush.Color := clBlack;
   for i := 0 to 2048 div BIN_SIZE do
   begin
       barValue := histoRect.Bottom - Muldiv(histogram[i],
           histoRect.Bottom - histoRect.Top, maxValue);
       leftRect := histoRect.Left + Muldiv(i, histoRect.Right - histoRect.Left,
           2048 div BIN_SIZE + 1);
       rightRect := histoRect.Left + Muldiv(i + 1, histoRect.Right - histoRect.Left,
           2048 div BIN_SIZE + 1) - 1;
       barRect := Rect(leftRect, barValue, rightRect, histoRect.Bottom);
       Canvas.FillRect(barRect);
   end;
end;


procedure TGammaFrm.DrawWhiteLevelLine(sliderPos: integer);
var    linePos: integer;
begin
   Canvas.Pen.Color := clRed;
   Canvas.Pen.Width := 1;
   Canvas.Pen.Style := psSolid;
   Canvas.Pen.Mode := pmXor;
   linePos := histoRect.Left + Muldiv(sliderPos, histoRect.Right - histoRect.Left,
       2047 + 128);
   Canvas.MoveTo(linePos, histoRect.Top);
   Canvas.LineTo(linePos, histoRect.Bottom);
   oldWhiteSliderPos := sliderPos;
end;


procedure TGammaFrm.FormPaint(Sender: TObject);
begin
   histoRect := Rect(BORDER_OFFSET, TrackBar1.Top + TrackBar1.Height,
       ClientWidth - BORDER_OFFSET, TrackBar2.Top - 1);
   DrawHistogram;
   DrawBlackLevelLine(TrackBar1.Position);
   DrawWhiteLevelLine(TrackBar2.Position);
end;


procedure TGammaFrm.TrackBar1Change(Sender: TObject);
begin
   if bInitializing then Exit;
   Label1.Caption := 'Black Level = ' + IntToStr(TrackBar1.Position - 128);
   DrawBlackLevelLine(oldBlackSliderPos);
   DrawBlackLevelLine(TrackBar1.Position);
end;


procedure TGammaFrm.TrackBar2Change(Sender: TObject);
begin
   if bInitializing then Exit;
   Label1.Caption := 'White Level = ' + IntToStr(TrackBar2.Position - 128);
   DrawWhiteLevelLine(oldWhiteSliderPos);
   DrawWhiteLevelLine(TrackBar2.Position);
end;


procedure TGammaFrm.RadioButton1Click(Sender: TObject);
begin
   DrawHistogram;
   DrawBlackLevelLine(TrackBar1.Position);
   DrawWhiteLevelLine(TrackBar2.Position);
end;


procedure	TGammaFrm.InitGUI(afile: TMPFile);
begin
	with afile do
   begin
   	RadioButton1.Checked := (DefaultVideoChannel = 0);
       RadioButton2.Checked := (DefaultVideoChannel = 1);
   	RadioButton3.Checked := (DefaultVideoChannel = 2);
       RadioButton4.Checked := (DefaultVideoChannel = 3);

       if VideoChCount = 1 then
       begin
       	RadioButton1.Enabled := False;
           RadioButton2.Enabled := False;
           RadioButton3.Enabled := False;
           RadioButton4.Enabled := False;
       end
       else
       begin
       	RadioButton1.Enabled := ChEnabled[0];
           RadioButton2.Enabled := ChEnabled[1];
           RadioButton3.Enabled := ChEnabled[2];
           RadioButton4.Enabled := ChEnabled[3];
       end;
   end;
end;


function   TGammaFrm.SelectedCh: integer;
begin
	if RadioButton1.Checked then
   	Result := 0
   else if RadioButton2.Checked then
   	Result := 1
   else if RadioButton3.Checked then
   	Result := 2
   else if RadioButton4.Checked then
   	Result := 3
   else
   	Result := 0;
end;


end.
