unit mp_ippi;

//
// INTEL Integrated Performance Primitives libraries for image processing
//

interface

{$ALIGN ON}
{$MINENUMSIZE 4}

uses Windows;

const
  ippCpuUnknown = $00;
  ippCpuPP      = ippCpuUnknown+1;  // Intel(R) Pentium(R) Processor
  ippCpuPMX     = ippCpuUnknown+2;  // Intel(R) Pentium(R) Processor with MMX(TM) Technology
  ippCpuPPR     = ippCpuUnknown+3;  // Intel(R) Pentium(R) Pro Processor
  ippCpuPII     = ippCpuUnknown+4;  // Intel(R) Pentium(R) II Processor
  ippCpuPIII    = ippCpuUnknown+5;  // Intel(R) Pentium(R) III Processor
  ippCpuP4      = ippCpuUnknown+6;  // Intel(R) Pentium(R) 4 Processor
  ippCpuP4HT    = ippCpuUnknown+7;  // Intel(R) Pentium(R) 4 Processor with HT Technology
  ippCpuP4HT2   = ippCpuUnknown+8;  // Intel(R) Pentium(R) 4 processor with HT Technology
  ippCpuCentrino= ippCpuUnknown+9;  // Intel(R) Centrino(TM) mobile technology
  ippCpuCoreSolo= ippCpuUnknown+10; // Intel(R) Core(TM) Solo processor
  ippCpuCoreDuo = ippCpuUnknown+11; // Intel(R) Core(TM) Duo processor
  ippCpuITP     = $10;              // Intel(R) Itanium(R) Processor
  ippCpuITP2    = ippCpuITP+1;      // Intel(R) Itanium(R) 2 Processor
  ippCpuEM64T   = $20;              // Intel(R) 64 Instruction Set Architecture (ISA)
  ippCpuC2D     = ippCpuEM64T+1;    // Intel(R) Core(TM) 2 Duo processor
  ippCpuC2Q     = ippCpuEM64T+2;    // Intel(R) Core(TM) 2 Quad processor
  ippCpuSSE41   = $44;              // Processor supports Streaming SIMD Extensions 4.1 instruction set

	IPP_FFT_NODIV_BY_ANY = 8;
   IPP_FFT_DIV_BY_SQRTN = 4;
	ippAlgHintFast = 1;

type
	pIPP32f = ^IPP32f;
	IPP32f  = single;
	ippStatus = integer;

	PIppLibraryVersion = ^TIppLibraryVersion;
	TIppLibraryVersion = record
		Major           : Integer;
		Minor           : Integer;
		MajorBuild      : Integer;
		Build           : Integer;
		TargetCpu       : array [0..3] of Char;
		Name            : PChar;
		Version         : PChar;
		BuildDate       : PChar;
	end;

	TpIppiFFTSpec_R_32f = pointer;
	IppHintAlgorithm = integer;
	Ipp32f_C1R = single;

   Ipp32fc = record
   	re, im: Ipp32f;
   end;

	IppiSize = record
		width, height: integer;
	end;

	TippiGetLibVersion  = function : PIppLibraryVersion; stdcall;
	TippccGetLibVersion = function : PIppLibraryVersion; stdcall;

	TippiFFTInitAlloc_R_32f = function(
		var pFFTSpec: TpIppiFFTSpec_R_32f;
   	orderX,
   	orderY,
   	flag: integer;
   	hint: IppHintAlgorithm): ippStatus; stdcall;

	TippiFFTGetBufSize_R_32f = function(
		pFFTSpec: TpIppiFFTSpec_R_32f;
   	var buffSize: integer): ippStatus; stdcall;

	TippiWinHamming_32f_C1IR = function(
		var Src: Ipp32f_C1R;
   	srcDesStep: integer;
   	roiSize: IppiSize): ippStatus; stdcall;

	TippiFFTFwd_RToPack_32f_C1IR = function(
		var Src: Ipp32f;
   	srcDesStep: integer;
   	const pFFTSpec: TpIppiFFTSpec_R_32f;
   	pBuffer: PByte): ippStatus; stdcall;

	TippiMulPackConj_32f_C1IR = function(
		var Src: Ipp32f;
   	srcStep: integer;
   	var Dst: Ipp32f;
   	dstSrcStep: integer;
   	roiSize: IppiSize): ippStatus; stdcall;

	TippiFFTInv_PackToR_32f_C1IR = function(
		var Src: Ipp32f;
   	srcDesStep: integer;
   	pFFTSpec: TpIppiFFTSpec_R_32f;
   	pBuffer: PByte): ippStatus; stdcall;

	TippiMagnitudePack_32f_C1R = function(
		var Src: Ipp32f;
   	srcStep: integer;
		var Dst: Ipp32f;
   	dstStep: integer;
   	roiSize: IppiSize): ippStatus; stdcall;

	TippiFFTFree_R_32f = function(
		ppFFTSpec: TpIppiFFTSpec_R_32f): ippStatus; stdcall;


const
	ippiGetLibVersion  : TippiGetLibVersion   = nil;
	ippccGetLibVersion : TippccGetLibVersion  = nil;
	ippiFFTInitAlloc_R_32f : TippiFFTInitAlloc_R_32f  = nil;
	ippiFFTGetBufSize_R_32f : TippiFFTGetBufSize_R_32f  = nil;
	ippiWinHamming_32f_C1IR : TippiWinHamming_32f_C1IR  = nil;
	ippiFFTFwd_RToPack_32f_C1IR : TippiFFTFwd_RToPack_32f_C1IR  = nil;
	ippiMulPackConj_32f_C1IR : TippiMulPackConj_32f_C1IR  = nil;
	ippiFFTInv_PackToR_32f_C1IR : TippiFFTInv_PackToR_32f_C1IR  = nil;
	ippiMagnitudePack_32f_C1R : TippiMagnitudePack_32f_C1R  = nil;
	ippiFFTFree_R_32f : TippiFFTFree_R_32f  = nil;

var
	m_IPP_Loaded: integer;
	m_nLibName: Integer;
	m_nccLibName: Integer;
	m_LibExists: array[0..5] of boolean;
	m_ccLibExists: array[0..5] of boolean;

function ippGetCpuType: Integer; stdcall;
function InitDLL(flag: String; nThreads: Integer) : integer;
procedure FreeDLL;

implementation

uses SysUtils;

const
	lName = 'Path';
	size: Integer = 256;
	ippcoreDLL = 'ippcore-5.3.dll';
	libsName: array[0..5] of string=(
		'ippipx-5.3.dll','ippia6-5.3.dll','ippiw7-5.3.dll',
		'ippit7-5.3.dll','ippiv8-5.3.dll','ippip8-5.3.dll');
	cclibsName: array[0..5] of string = (
		'ippccpx-5.3.dll','ippcca6-5.3.dll','ippccw7-5.3.dll','ippcct7-5.3.dll',
		'ippccv8-5.3.dll','ippccp8-5.3.dll');

var
	Handle  : HINST;
	ccHandle  : HINST;
	IppiDLL : PChar = 'ippi-5.3.dll';
	IppccDLL : PChar = 'ippcc-5.3.dll';
	buffer: array [0..255] of Char;
	pathDir: string;

function ippGetCpuType;   external ippcoreDLL;

function InitDLL(flag: String; nThreads: Integer) : integer;
const
	sThreads = 'OMP_NUM_THREADS';
var
	buf: array[0..1] of Char;
	version: PIppLibraryVersion;
	str: string;
	CpuType: Integer;
begin
	if nThreads > 0 then
	begin
		GetEnvironmentVariable(sThreads, buf, 2);
		buf[0] := IntToStr(nThreads)[1];
		buf[1] := #0;
		SetEnvironmentVariable(sThreads, buf);
	end;

	if flag = '' then
	begin
		Handle := LoadLibrary(IppiDLL);
 		if Handle <> 0 then
		begin
			GetModuleFileName(Handle, @buffer, size);
			pathDir := ExtractFileDir(buffer)+'\';
			@ippiGetLibVersion := GetProcAddress(Handle,'ippiGetLibVersion');
			if ippiGetLibVersion <> nil then
			begin
				version := ippiGetLibVersion;
				str := version.Name;
				FreeLibrary(Handle);
				Handle := LoadLibrary(PChar(pathDir+str));
				if str = libsName[0] then m_nLibName := 0;
				if str = libsName[1] then m_nLibName := 1;
				if str = libsName[2] then m_nLibName := 2;
				if str = libsName[3] then m_nLibName := 3;
				if str = libsName[4] then m_nLibName := 4;
				if str = libsName[5] then m_nLibName := 5;
				CpuType := ippGetCpuType;
				m_LibExists[0] := FileExists(pathDir+libsName[0]);
				m_LibExists[1] := FileExists(pathDir+libsName[1]) and (CpuType>=ippCpuPIII);
				m_LibExists[2] := FileExists(pathDir+libsName[2]) and (CpuType>=ippCpuP4);
				m_LibExists[3] := FileExists(pathDir+libsName[3]) and ((CpuType >= ippCpuP4HT2) or (CpuType >= ippCpuCoreDuo) or (CpuType >=ippCpuCoreSolo) or (CpuType >=ippCpuEM64T));
				m_LibExists[4] := FileExists(pathDir+libsName[4]) and ((CpuType =ippCpuC2D) or (CpuType =ippCpuC2Q));
				m_LibExists[5] := FileExists(pathDir+libsName[5]) and (CpuType = ippCpuSSE41);
      			end;
    		end;
    		ccHandle := LoadLibrary(IppccDLL);
			if ccHandle <> 0 then
			begin
				GetModuleFileName(ccHandle, @buffer, size);
				pathDir := ExtractFileDir(buffer)+'\';
				@ippccGetLibVersion := GetProcAddress(ccHandle,'ippccGetLibVersion');
				if ippccGetLibVersion <> nil then
				begin
					version := ippccGetLibVersion;
					str := version.Name;
					FreeLibrary(ccHandle);
					ccHandle := LoadLibrary(PChar(pathDir+str));
					if str = cclibsName[0] then m_nccLibName := 0;
					if str = cclibsName[1] then m_nccLibName := 1;
					if str = cclibsName[2] then m_nccLibName := 2;
					if str = cclibsName[3] then m_nccLibName := 3;
					if str = cclibsName[4] then m_nccLibName := 4;
					if str = cclibsName[5] then m_nccLibName := 5;
					CpuType := ippGetCpuType;
					m_ccLibExists[0] := FileExists(pathDir+cclibsName[0]);
					m_ccLibExists[1] := FileExists(pathDir+cclibsName[1]) and (CpuType>=ippCpuPIII);
 					m_ccLibExists[2] := FileExists(pathDir+cclibsName[2]) and (CpuType>=ippCpuP4);
					m_ccLibExists[3] := FileExists(pathDir+cclibsName[3]) and ((CpuType >= ippCpuP4HT2) or (CpuType >= ippCpuCoreDuo) or (CpuType >=ippCpuCoreSolo) or (CpuType >=ippCpuEM64T));
					m_ccLibExists[4] := FileExists(pathDir+cclibsName[4]) and ((CpuType =ippCpuC2D) or (CpuType =ippCpuC2Q));
					m_ccLibExists[5] := FileExists(pathDir+cclibsName[5]) and (CpuType = ippCpuSSE41);
				end;
			end;
	end
	else
  	begin
   	FreeLibrary(Handle);
		//Handle := LoadLibrary(PChar(pathDir+libsName[flag]));
		Handle := LoadLibrary(PChar(pathDir+flag+'-5.3.dll'));
	end;

	if Handle = 0 then
	begin
 		result := -1;
       Exit;
	end;

   @ippiFFTInitAlloc_R_32f := GetProcAddress(Handle,'ippiFFTInitAlloc_R_32f');
   @ippiFFTGetBufSize_R_32f := GetProcAddress(Handle,'ippiFFTGetBufSize_R_32f');
   @ippiWinHamming_32f_C1IR := GetProcAddress(Handle,'ippiWinHamming_32f_C1IR');
   @ippiFFTFwd_RToPack_32f_C1IR := GetProcAddress(Handle,'ippiFFTFwd_RToPack_32f_C1IR');
   @ippiMulPackConj_32f_C1IR := GetProcAddress(Handle,'ippiMulPackConj_32f_C1IR');
   @ippiFFTInv_PackToR_32f_C1IR := GetProcAddress(Handle,'ippiFFTInv_PackToR_32f_C1IR');
   @ippiMagnitudePack_32f_C1R := GetProcAddress(Handle,'ippiMagnitudePack_32f_C1R');
   @ippiFFTFree_R_32f := GetProcAddress(Handle,'ippiFFTFree_R_32f');
	result := 0;
end;


procedure FreeDLL;
begin
	if Handle <> 0 then FreeLibrary(Handle);
	if ccHandle <> 0 then FreeLibrary(ccHandle);
end;


initialization
	m_IPP_Loaded := InitDLL('', -1);

finalization
	FreeDLL;

end.
