unit mpfileu;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Forms, Registry, StdCtrls, ActiveX,
       ComObj, mpviewu, vfw, mp_ippi;
       
{$ASSERTIONS ON}

type
   TMPFile = class;


   TFrame = class
   private
       chIndex, frameIndex: integer;
       mpFile: TMPFile;
       procedure   LoadFromFile(frameIndex: integer); virtual; abstract;
   public
       data: array of int16;
       constructor Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
   end;


   TVideoFrame = class(TFrame)
   private
       function    GetHeight: integer;
       function    GetPixels(x, y: integer): int16;
       function    GetWidth: integer;
       procedure   LoadFromFile(frameIndex: integer); override;
       procedure   SaveToFile;
   public
       procedure   CopyData(dest: TVideoFrame);
       procedure   CopyToDoubleArray(var dblArray: array of double);
       procedure   GammaCorrection(blackLevel, whiteLevel: integer);
       procedure   GetProfile(binSize: integer; var histogram: array of integer);

		constructor Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);

       property    Height: integer read GetHeight;
       property    Pixels[x, y: integer]: int16 read GetPixels;
       property    Width: integer read GetWidth;
   end;


   TAnalogFrame = class(TFrame)
   {x: index of analog sample in frame data}
   private
       function    GetIAnalogValue(x: integer): integer;
       function    GetfAnalogValue(x: integer): double;
       procedure   LoadFromFile(frameIndex: integer); override;
   public
       function    SampleToAbsoluteTime(x: integer): double; {in seconds}

		constructor Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);

       property    iAnalogValue[x: integer]: integer read GetIAnalogValue;
       property    fAnalogValue[x: integer]: double read GetfAnalogValue;
   end;


   TFrameCommentList = class(TStringList)
   private
       mpFile: TMPFile;
       procedure   SaveFrameComments;
       procedure   LoadFrameComments;
       function    GetFrameComment(frameIndex: integer): string;
       procedure   SetFrameComment(frameIndex: integer; const frameComment: string);
   public
       constructor Create(theMPFile: TMPFile);
       property    FrameComment[frameIndex: integer]: string read GetFrameComment write SetFrameComment;
   end;


   TFrameGroup = class
   private
       fWidth, fHeight, fFrameIndex: integer;
       mpFile: TMPFile;
       procedure   LoadFromFile(frameIndex: integer);
     {  procedure   SaveToFile;}
       procedure   SetFrameWidth(newWidth: integer);
       procedure   SetFrameHeight(newHeight: integer);
       procedure   SetFrameIndex(newIndex: integer);
   public
       comments: string;
       channels: array[0..MAX_CH - 1] of TFrame; {0..1: TVideoFrame; 2..3: TAnalogFrame}

       property    FrameIndex: integer read fFrameIndex write SetFrameIndex;
       property    FrameWidth: integer read fWidth write SetFrameWidth;
       property    FrameHeight: integer read fHeight write SetFrameHeight;
       constructor Create(mp_file: TMPFile);
       constructor CreateTempFrameGroup(mp_file: TMPFile);
       constructor CreateForOp(mp_file: TMPFile); {creates a frame group for operations: only channels[0] is valid}
       destructor  Destroy; override;
   end;


   //-------------------------------- FILE OBJECT ------------------------------

	TIntensityControl = (IC_NO_CONTROL, IC_LINEAR, IC_EXPONENTIAL);

   TMPFile = class
   private
   	bReadOnly: boolean;
       fActiveFrameIndex: integer;
       fAngle: integer;
       {fbLineForXScan, OBSOLETE}
       fColorScheme: TColorScheme;
       fbIsMemoryFile, fbClosing: boolean;
       fChConvFactors: array[0..MAX_CH -1] of double;
       fChDataPtsPerFrames: array[0..MAX_CH -1] of integer;
       fChEnabled, fIsVideoCh: array[0..MAX_CH - 1] of boolean;
       fChInputRanges: array[0..MAX_CH - 1] of TFullScaleVal;
       fChNames: array[0..MAX_CH - 1] of string;
       fChMaxPixelValues: array[0..MAX_CH - 1] of integer;
       fChOffsets: array[0..MAX_CH -1] of double;
       fChPrefixes: array[0..MAX_CH -1] of TPrefix;
       fChUnits: array[0..MAX_CH - 1] of string;
       fFilename: string;
       fFrameCount: integer;
       fFrameHeight,
       fFrameWidth: integer;
       fLineRepeatCount: integer;
       fMagnification: double;
       fPixelClock {in increments of 50 ns (= 20 MHz)}: integer;
       fResolution: TFrameResolution; {RESOLUTION_8_BITS, 12, 16 bits}
       fRotation: integer;
       fScanMode: TScanMode;
       fStackAveragingCount,
       fStackFrameCount: integer;
       fStackInterval: double;
       fStackRepeatCount,
       fXFrameOffset,
       fXStagePosition,
       fYFrameOffset,
       fYStagePosition: integer;
       fZStagePosition: double;
       sComments: string;
       fzDistance: double;
       fTravelDuration: string;
       fFastStackRepeatCount: integer;
       fInitialIntensity, fFinalIntensity: integer;
		fAtZDistance: double;
       fIntensityControl: TIntensityControl;

		RegionPts: array of TPoint; {for region scans}

       chStreams: array[0..MAX_CH -1] of IStream;
       propertyStorage: IPropertyStorage;
       propertySetStorage: IPropertySetStorage;
       rootStorage: IStorage;

       fCurrentFrameGroup: TFrameGroup;
       frameCommentList: TFrameCommentList;
       frameGroupList: TList;
       bDirty: boolean;

		disableMotionCorrection, {when computing motion correction, this flag is true}
       fbMotionCorrectionEnabled: boolean;
       FFTroiSize: IppiSize;
       pSpec: TpIppiFFTSpec_R_32f;
       refMotionChannel: integer;
       refFFTarray, frameFFTarray, magnitudeArray: array of single;
       poweroftwo_width, poweroftwo_height: integer;
       FFTBuffer: array of Byte; {used by Intel functions}
		refMotionRect: TRect;

     {  function    GetActiveFrameGroup: TFrameGroup;}
       procedure   AdjustFrameCount;
       procedure	FillBaseColors(chIndex: integer; const s: string);
       function    GetAnalogChCount: integer;
       function	GetAnalogChEnabled(chIndex: integer): boolean;
       function    GetAnalogWndCount: integer;
       function    GetAnalogWnds(wndIndex: integer): TForm;
       function    GetChAcquisitionRate(chIndex: integer): double;
       function    GetChConvFactor(chIndex: integer): double;
       function    GetChDataPtsPerFrame(chIndex: integer): integer;
       function    GetChEnabled(chIndex: integer): boolean;
       function    GetChInputRange(chIndex: integer): TFullScaleVal;
       function    GetChMaxPixelValues(chIndex: integer): integer;
       function    GetChNames(chIndex: integer): string;
       function    GetChUnit(chIndex: integer): string;
       function    GetChPrefix(chIndex: integer): TPrefix;
       function    GetChOffset(chIndex: integer): double;
      { function    GetCurrentFrame(chIndex: integer): TFrame;}
       function    GetDefaultVideoChannel: integer;
       function    GetFrameComment(frameIndex: integer): string;
       function    GetFrameCount: integer;
       function    GetFrameRate: double; {in frames per seconds}
       function	GetFrameSize: integer;
       function    GetFrames(frameIndex: integer): TFrameGroup;
       function	GetFullFrameSize: integer;
       function	GetIsVideoChannel(chIndex: integer): boolean;
       function	GetMaxPixelValue: integer;
       procedure	GetRegions;
       function	GetResolution: string;
       function    GetVideoChCount: integer;
       function	GetVideoChEnabled(chIndex: integer): boolean;
       function    GetViewerCount: integer;
       function    GetViewers(viewerIndex: integer): TForm;
		procedure   InitializeColors;
       procedure   LoadFrameGroup(newIndex: integer);
       procedure	LoadMotionCorrection;
       procedure   ReadPropertySet;
       procedure   SetActiveFrameIndex(newIndex: integer);
       procedure	SetMotionCorrectionEnabled(flag: boolean);
       procedure   SetFrameComment(frameIndex: integer; sComments: string);
       procedure   WritePropertySet;

   public
       viewerList, analogWndList: TList;
    	{custom LUTs}
    	baseColors: TBaseColorsArray;
    	negativeColors,
    	midRangeColors,
    	maxColors: TRangeColorsArray;
    	maxPixels: TMaxPixelsArray;
    	CustomColors: array[0..MAX_CH - 1, 0..MAX_FALSE_COLORS - 1] of TRGBTriple;
       motionBuffer: array of int16;
       motionDeltas: array of TPoint;

       procedure   AddFrameGroup(newFrameGroup: TFrameGroup);
       procedure   AverageFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
       procedure	BinaryOp(file1, file2: TMPFile; ch1, ch2, frame1, frame2: integer; gain1, gain2: double;
       				resultFrame: integer);
       procedure   Close(Sender: TObject);
       function	CreateEmptyFrame(fromFile: TMPFile): integer;
    	procedure	CopyChannelsToClipboard(bCh2, bCh3: boolean; fromFrame, toFrame: integer);
       procedure   CopyFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile); {from one file to a memory file, rect to rect}
       procedure	CloneROIs(ROIList: TROIList; fromIndex: integer); {duplicate each ROI but with opposite channel}
       procedure   DetectROIs(ROIList: TROIList; chIndex, fromFrame, toFrame, threshold, minArea,
       			templateFrom, templateTo: integer);
       function    DigitalToAnalog(chIndex, iValue: integer): double;
       procedure	DoBackgroundCorrection;
       procedure	FinishMotionCorrection;
       function	GetAverage(frameIndex, chIndex: integer; rc: TRect): integer;
       function    GetROIAverageValue(ROIList: TROIList; roiIndex, frameIndex: integer): integer;
       function    GetPixelValue(frameIndex, chIndex, xData, yData: integer): int16;
       function	GetMax(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
       function	GetMin(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
       function	GetPropertyValue(propName: string): string;
       function    IsOperationOK(chIndex, fromFrame, toFrame: integer; dest: TMPFile): boolean;
       procedure   MakeAVIMovie(const avifilename: string; chIndex, fromFrame, toFrame, newframeRate: integer;
						colorScheme: TColorScheme; fromViewer: TObject; bOverlayCh1on2: boolean);
       function    MakeTIFF(const avifilename: string; chIndex, fromFrame, toFrame: integer): boolean;
       procedure   NewAnalogWnd;
       procedure   NewViewer;
       procedure   OnNewFrames;
       procedure   OnWndClose(wnd: TForm); {when a viewer or analog wnd closes}
       procedure	PerformMotionCorrection(frameIndex: integer);
       procedure	PrepareMotionCorrection(refFrame, refCh: integer; refRect: TRect);
       procedure	SaveChannelsToFile(bASCII: boolean; fname: string; bCh2, bCh3: boolean; fromFrame, toFrame: integer);
       function   	SaveMotionCorrection: boolean;
		function	SizeOfFrameCompatible(width, height: integer): boolean;
       procedure   StackX(chIndex, fromFrame, toFrame, fromY, toY: integer; dest: TMPFile);
       procedure   StackY(chIndex, fromFrame, toFrame, fromX, toX: integer; dest: TMPFile);
       procedure   StackZ(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
       function    SaveAs(const sFilename: string): TFileErr;
       procedure   SubtractFrame(chIndex, plusFrameIndex, minusFrameIndex: integer; dest: TMPFile);

       constructor CreateFromTemplate(const sFilename: string; templateFile: TMPFile);
       constructor CreateFromFile(const sFilename: string);
       destructor  Destroy; override;

       property    ActiveFrameIndex: integer read fActiveFrameIndex write SetActiveFrameIndex;
       property    AnalogChCount: integer read GetAnalogChCount;
       property	AnalogChEnabled[chIndex: integer]: boolean read GetAnalogChEnabled;
       property    AnalogWndCount: integer read GetAnalogWndCount;
       property    AnalogWnds[wndIndex: integer]: TForm read GetAnalogWnds;
       property    Angle: integer read fAngle;
       property    ChAcquisitionRate[chIndex: integer]: double read GetChAcquisitionRate;
       property    ChConvFactor[chIndex: integer]: double read GetChConvFactor;
       property    ChDataPtsPerFrame[chIndex: integer]: integer read GetChDataPtsPerFrame;
       property    ChEnabled[chIndex: integer]: boolean read GetChEnabled;
       property    ChInputRange[chIndex: integer]: TFullScaleVal read GetChInputRange;
       property    ChNames[chIndex: integer]: string read GetChNames;
       property    ChUnit[chIndex: integer]: string read GetChUnit;
       property	ChMaxPixelValues[chIndex: integer]: integer read GetChMaxPixelValues;
       property    ChPrefix[chIndex: integer]: TPrefix read GetChPrefix;
       property    ChOffset[chIndex: integer]: double read GetChOffset;
       property    Closing: boolean read fbClosing; {when user closes file from menu}
       property	ColorScheme: TColorScheme read fColorScheme;
       property    Comments: string read sComments write sComments;
       property	DefaultVideoChannel: integer read GetDefaultVideoChannel;
       property    Filename: string read ffilename;
       property    FrameComment[frameIndex: integer]: string read GetFrameComment
                       write SetFrameComment;
       property    FrameCount: integer read GetFrameCount;
       property    FrameHeight: integer read fFrameHeight;
       property    FrameRate: double read GetFrameRate; {in frames per second}
       property	FrameSize: integer read GetFrameSize; {in samples}
       property    Frames[frameIndex: integer]: TFrameGroup read GetFrames;
       property    FrameWidth: integer read fFrameWidth;
       property	FullFrameSize: integer read GetFullFrameSize; {= Frame size + "forgotten pixels"}
       property    IsMemoryFile: boolean read fbIsMemoryFile;
       property	IsVideoChannel[chIndex: integer]: boolean read GetIsVideoChannel;
       property    LineRepeatCount: integer read fLineRepeatCount;
     {  property    LineForXScan: boolean read fbLineForXScan;}
       property    Magnification: double read fMagnification;
       property	MaxPixelValue: integer read GetMaxPixelValue; {255, 2047 or 32767}
       property	MotionCorrectionEnabled: boolean read fbMotionCorrectionEnabled write SetMotionCorrectionEnabled;
       property    PixelClock: integer read fPixelClock; {in ns}
       property	Resolution: string read GetResolution;
       property    Rotation: integer read fRotation;
       property    VideoChCount: integer read GetVideoChCount;
       property	VideoChEnabled[chIndex: integer]: boolean read GetVideoChEnabled;
       property    ViewerCount: integer read GetViewerCount;
       property    Viewers[viewerIndex: integer]: TForm read GetViewers;
       property    ScanMode: TScanMode read fScanMode;
       property    StackAveragingCount: integer read fStackAveragingCount;
       property    StackFrameCount: integer read fStackFrameCount;
       property    StackInterval: double read fStackInterval;
       property    StackRepeatCount: integer read fStackRepeatCount;
       property    XFrameOffset: integer read fXFrameOffset;
       property    XStagePosition: integer read fXStagePosition;
       property    YFrameOffset: integer read fYFrameOffset;
       property    YStagePosition: integer read fYStagePosition;
       property    ZStagePosition: double read fZStagePosition;
   end;


   //------------------------------- FILE LIST ---------------------------------

   {holds all the files}

   TFileList = class(TStringList)
   private
       function    IsMPFile(const sFilename: string): boolean;
       function    GetTemporaryFileName: string;
       function    Load(const sFilename: string): TFileErr;
   public
       dataDirectory: string;
       procedure   FillComboBoxWithWorkspaces(aComboBox: TComboBox);
       procedure   NotifyFileClosing(Sender: TMPFile);
       procedure   NewFile(templateFile: TMPFile);
       procedure   Open(const sFilename: string);
       procedure   SaveFileAs(mpFile: TMPFile; const newName: string);
       function    WorkspaceCount: integer;

       destructor  Destroy; override;
   end;


function BoolToString(b: Boolean): string;
{function LineForXScanToString(bLine: boolean): string; OBSOLETE}
{function StringToLineForXScan(s: string): boolean;}
function ScanModeToString(smode: TScanMode): string;
function StringToScanMode(s: string): TScanMode;
function StringToBool(s: string): Boolean;
function StringToInputRange(s: string): TFullScaleVal;
function StringToPrefix(s: string): TPrefix;
function IntensityControlToString(ic: TIntensityControl): string;

{******************************************************************************}
{*}                                                                          {*}
{*}                            IMPLEMENTATION                                {*}
{*}                                                                          {*}
{******************************************************************************}


uses mainfrm, Math, analogu, vieweru, dialogs, Clipbrd;

const
	sMotionCorrection = 'Motion Corrections';

	TIFF_TAG_COUNT = 22; {Includes ImageDescription tag}
   TIFF_ASCII = 2;
   TIFF_SHORT = 3;
   TIFF_LONG = 4;
   TIFF_RATIONAL = 5;

   TIFF_ImageDescription = 270;


   FILE_PROP_COUNT = 74; {74 default properties per file in property set}

type
	TTIFFHeader = packed record
		order,
   	signature: Word;
       IFDOffset: integer;
   end;


   TTagRecord = packed record
   	tagID, tagType: Word;
       count, dataOffset: integer;
   end;

   TIFD = packed record
   	count: Word;
       tagRecords: array[1..TIFF_TAG_COUNT] of TTagRecord;
       nextIDFoffset: integer;
       XNum, XDenom,
       YNum, YDenom: integer;
   end;
   {followed by image description ASCII string}

const
	TiffHeader: TTIFFHeader = (order: $4949; signature: 42; IFDOffset: 8);

   OFLAGS = STGM_DIRECT or STGM_READWRITE or STGM_CREATE or STGM_SHARE_EXCLUSIVE;
   RFLAGS = STGM_DIRECT or STGM_READ or STGM_SHARE_EXCLUSIVE;
   RWFLAGS = STGM_DIRECT or STGM_READWRITE or STGM_SHARE_EXCLUSIVE;

   FMTID_User_Defined_Properties: TGUID = '{D5CDD505-2E9C-101B-9397-08002B2CF9AE}';

	sTIFFPropNames: string =
   	'Channel: %d' + #13#10 +
   	'Resolution: %s' + #13#10 +
       'Scan Mode: %s' + #13#10 +
       'X Stage Position (microns): %d' + #13#10 +
       'Y Stage Position (microns): %d' + #13#10 +
       'Z Stage Position (microns): %f' + #13#10 +
       'Section Count: %d' + #13#10 +
       'z- Interval (microns): %f' + #13#10 +
       'Frames Per Section: %d' + #13#10 +
       'Stack Repeat Count: %d' + #13#10 +
       'Magnification: x%f' + #13#10 +
       'Rotation (degrees): %d' + #13#10 +
       'X Frame Offset (pixels): %d' + #13#10 +
       'Y Frame Offset (pixels): %d' + #13#10 +
       'Frame Rate (fps): %f';

   sPropNames: array[0..FILE_PROP_COUNT - 1] of string = (
   	'Resolution',
       'Scan Mode',
       'X Frame Offset',
       'Y Frame Offset',
       'Frame Width',
       'Frame Height',
       'Pixel Clock',
       'Channel name (1)', 'Enabled (1)', 'Input Range (1)',
       'Channel name (2)', 'Enabled (2)', 'Input Range (2)',
       'Channel name (3)', 'Enabled (3)', 'Input Range (3)', 'Channel Unit (3)',
       'Channel Prefix (3)', 'Conversion factor (3)', 'Offset (3)', 'Data Points Per Frame (3)',
       'Channel name (4)', 'Enabled (4)', 'Input Range (4)', 'Channel Unit (4)',
       'Channel Prefix (4)', 'Conversion factor (4)', 'Offset (4)', 'Data Points Per Frame (4)',
       'Frame Count',
       'X Position',
       'Y Position',
       'Z Position',
       'Section Count',
       'z- Interval',
       'Averaging Count',
       'Repeat Count',
       'Magnification',
       'Rotation',
       'Comments',
		'Is Video Channel (3)',
       'Is Video Channel (4)',
       'z- Distance',
       'Travel Duration',
       'Fast Stack Repeat Count',
       'Initial Intensity',
       'Final Intensity',
       'At Z- Distance',
       'Intensity Control',
       'Ch Max Pixel Value (1)',
       'Ch Max Pixel Value (2)',
       'Ch Max Pixel Value (3)',
       'Ch Max Pixel Value (4)', {53 props}
        slutType,
        'Base Color (1)', {55}
        'Base Color (2)',
        'Base Color (3)',
        'Base Color (4)',
        'Negative Color (1)', {59}
        'Negative Color (2)',
        'Negative Color (3)',
        'Negative Color (4)',
        'MidRange Color (1)',  {63}
        'MidRange Color (2)',
        'MidRange Color (3)',
        'MidRange Color (4)',
        'Max Color (1)',  {67}
        'Max Color (2)',
        'Max Color (3)',
        'Max Color (4)',
        'Max Pixel Value of LUT (1)', {71}
        'Max Pixel Value of LUT (2)',
        'Max Pixel Value of LUT (3)',
        'Max Pixel Value of LUT (4)'
        );

   sAlreadyLoaded = 'File %s is already opened in MPView.';


procedure MakeIFD(var IFD: TIFD);
begin
	with IFD do
   begin
   	count := TIFF_TAG_COUNT;

       tagRecords[1].tagID := 254; {NewSubtitleType}
       tagRecords[1].tagType := TIFF_LONG;
       tagRecords[1].count := 1;
       tagRecords[1].dataOffset := 0;

       tagRecords[2].tagID := 256; {ImageWidth}
       tagRecords[2].tagType := TIFF_LONG;
       tagRecords[2].count := 1;

       tagRecords[3].tagID := 257; {ImageLength}
       tagRecords[3].tagType := TIFF_LONG;
       tagRecords[3].count := 1;

       tagRecords[4].tagID := 258; {BitsPerSample}
       tagRecords[4].tagType := TIFF_SHORT;
       tagRecords[4].count := 1;
       tagRecords[4].dataOffset := 16;

       tagRecords[5].tagID := 259; {Compression}
       tagRecords[5].tagType := TIFF_SHORT;
       tagRecords[5].count := 1;
       tagRecords[5].dataOffset := 1; {No compression}

       tagRecords[6].tagID := 262; {PhotometricInterpretation}
       tagRecords[6].tagType := TIFF_SHORT;
       tagRecords[6].count := 1;
       tagRecords[6].dataOffset := 1; {BlackIsZero}

       tagRecords[7].tagID := 263; {Tresholding}
       tagRecords[7].tagType := TIFF_SHORT;
       tagRecords[7].count := 1;
       tagRecords[7].dataOffset := 1;

       tagRecords[8].tagID := 266; {FillOrder}
       tagRecords[8].tagType := TIFF_SHORT;
       tagRecords[8].count := 1;
       tagRecords[8].dataOffset := 1;

       tagRecords[9].tagID := 273; {StripOffsets}
       tagRecords[9].tagType := TIFF_LONG;
       tagRecords[9].count := 1;

       tagRecords[10].tagID := 274; {Orientation}
       tagRecords[10].tagType := TIFF_SHORT;
       tagRecords[10].count := 1;
       tagRecords[10].dataOffset := 1;

       tagRecords[11].tagID := 277; {SamplesPerPixel}
       tagRecords[11].tagType := TIFF_SHORT;
       tagRecords[11].count := 1;
       tagRecords[11].dataOffset := 1;

       tagRecords[12].tagID := 278; {RowsPerStrip}
       tagRecords[12].tagType := TIFF_LONG;
       tagRecords[12].count := 1;

       tagRecords[13].tagID := 279; {StripByteCounts}
       tagRecords[13].tagType := TIFF_LONG;
       tagRecords[13].count := 1;

       tagRecords[14].tagID := 280; {MinSampleValue}
       tagRecords[14].tagType := TIFF_SHORT;
       tagRecords[14].count := 1;
       tagRecords[14].dataOffset := 0;

       tagRecords[15].tagID := 281; {MaxSampleValue}
       tagRecords[15].tagType := TIFF_SHORT;
       tagRecords[15].count := 1;
       tagRecords[15].dataOffset := 65535;

       tagRecords[16].tagID := 282; {XResolution}
       tagRecords[16].tagType := TIFF_RATIONAL;
       tagRecords[16].count := 1;

       tagRecords[17].tagID := 283; {YResolution}
       tagRecords[17].tagType := TIFF_RATIONAL;
       tagRecords[17].count := 1;

       tagRecords[18].tagID := 284; {PlanarConfiguration}
       tagRecords[18].tagType := TIFF_SHORT;
       tagRecords[18].count := 1;
       tagRecords[18].dataOffset := 1; {Chunky}

       tagRecords[19].tagID := 290; {GrayResponseUnit}
       tagRecords[19].tagType := TIFF_SHORT;
       tagRecords[19].count := 1;
       tagRecords[19].dataOffset := 1;

       tagRecords[20].tagID := 296; {ResolutionUnit}
       tagRecords[20].tagType := TIFF_SHORT;
       tagRecords[20].count := 1;
       tagRecords[20].dataOffset := 2; {Inch}

       tagRecords[21].tagID := 339; {SampleFormat}
       tagRecords[21].tagType := TIFF_SHORT;
       tagRecords[21].count := 1;
       tagRecords[21].dataOffset := 1;

       tagRecords[22].tagID := TIFF_ImageDescription;
       tagRecords[22].tagType := TIFF_ASCII;

       XNum := 72;
       XDenom := 1;
       YNum := 72;
       YDenom := 1;
   end;
end;

{
OBSOLETE FUNCTIONS
function LineForXScanToString(bLine: boolean): string;
begin
   if bLine then Result := 'Line' else Result := 'Sine';
end;


function StringToLineForXScan(s: string): boolean;
begin
   Result := (s = 'Line');
end;}


function ScanModeToString(smode: TScanMode): string;
begin
   case smode of
       SM_MOVIE: Result := 'Movie';
       SM_STACK: Result := 'Image Stack';
     {  SM_STACKMOVIE: Result := 'Image Stack Movie';}
       SM_LINESCAN: Result := 'Line Scan';
       SM_REPEAT_LINESCAN: Result := 'Repeat Line Scan';
       SM_REGIONSCAN: Result := 'Region Scan';
   else Result := '';
   end
end;


function StringToColorScheme(const sScheme: string): TColorScheme;
begin
	if sScheme = 'Gray scale' then
   	Result := CS_GRAYSCALE
   else if sScheme = 'False colors' then
   	Result := CS_FALSECOLORS
   else if sScheme = 'Custom colors' then
   	Result := CS_CUSTOMLUT
   else
   	Result := CS_GRAYSCALE;
end;


function StringToScanMode(s: string): TScanMode;
begin
   if s = 'Movie' then Result := SM_MOVIE else
   if s = 'Image Stack' then Result := SM_STACK else
 {  if s = 'Image Stack Movie' then Result := SM_STACKMOVIE else}
   if s = 'Line Scan' then Result := SM_LINESCAN else
   if s = 'Repeat Line Scan' then Result := SM_REPEAT_LINESCAN else
   if s = 'Region Scan' then Result := SM_REGIONSCAN else
       Result := SM_MOVIE;
end;


function BoolToString(b: Boolean): string;
begin
   if b then Result := 'True' else Result := 'False';
end;


function StringToBool(s: string): Boolean;
begin
   Result := (s = 'True');
end;


function StringToInputRange(s: string): TFullScaleVal;
begin
   if s = Chr(177) + '42V' then Result := pm_42V else
   if s = Chr(177) + '20V' then Result := pm_20V else
   if s = Chr(177) + '10V' then Result := pm_10V else
   if s = Chr(177) + '5V' then Result := pm_5V else
   if s = Chr(177) + '2V' then Result := pm_2V else
   if s = Chr(177) + '1V' then Result := pm_1V else
   if s = Chr(177) + '0.5V' then Result := pm_0_5V else
   if s = Chr(177) + '0.2V' then Result := pm_0_2V else
       Result := pm_10V;
end;


function StringToPrefix(s: string): TPrefix;
begin
   if s = 'x' then Result := tpXENNO else
   if s = 'y' then Result := tpYOCTO else
   if s = 'z' then Result := tpZEPTO else
   if s = 'a' then Result := tpATTO else
   if s = 'f' then Result := tpFEMTO else
   if s = 'p' then Result := tpPICO else
   if s = 'n' then Result := tpNANO else
   if s = #181 then Result := tpMICRO else
   if s = 'm' then Result := tpMILLI else
   if s = '' then Result := tpUNITY else
   if s = 'k' then Result := tpKILO else
   if s = 'M' then Result := tpMEGA else
   if s = 'G' then Result := tpGIGA else
   if s = 'T' then Result := tpTERA else
   if s = 'P' then Result := tpPETA  else
   if s = 'E' then Result := tpECTA else
   if s = 'Z' then Result := tpZETTA else
   if s = 'Y' then Result := tpYOTTA else
   if s = 'X' then Result := tpXENNA else
       Result := tpUNITY;
end;

//-------------------------------- FRAME  --------------------------------------

constructor TFrame.Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
var	i: integer;
begin
   mpFile := mp_File;
   frameIndex := iFrameIndex;
   chIndex := iChIndex;
   SetLength(data, mpFile.FrameWidth * mpFile.FrameHeight);
   if mpFile.ScanMode = SM_REGIONSCAN then
   	for i := 0 to Length(data) - 1 do data[i] := 0;
end;

//------------------------------ VIDEO FRAME  ----------------------------------

function    TVideoFrame.GetHeight: integer;
begin
   Result := mpFile.FrameHeight;
end;


function    TVideoFrame.GetPixels(x, y: integer): int16;
begin
   Result := data[x + y * Width];
end;


function    TVideoFrame.GetWidth: integer;
begin
   Result := mpFile.FrameWidth;
end;


procedure   TVideoFrame.LoadFromFile(frameIndex: integer);
var    libNewPosition: Largeint;
       i, j, sourceFirstIndex, destFirstIndex: integer;
       pt: TPoint;
       tempData: array of int16;
begin
   OleCheck(mpFile.chStreams[chIndex].Seek(frameIndex * mpFile.FrameSize * SizeOf(int16)
   	, STREAM_SEEK_SET, libNewPosition));

   if not mpFile.MotionCorrectionEnabled or mpfile.DisableMotionCorrection then
   begin
   	{Copy directly to data buffer}
   	if mpFile.ScanMode <> SM_REGIONSCAN then
   		OleCheck(mpFile.chStreams[chIndex].Read(@data[0], mpFile.FrameSize * SizeOf(int16), nil))
   	else
   	begin
   		{Region scan: the frame is 512 x 512, we fill with data indexed by RegionPts}
			FillChar(data[0], mpFile.FrameWidth * mpFile.FrameHeight * SizeOf(int16), 0);
   		SetLength(tempData, mpFile.FrameSize);
   		OleCheck(mpFile.chStreams[chIndex].Read(@tempData[0], mpFile.FrameSize * SizeOf(int16), nil));
       	for i := 0 to mpFile.FrameSize - 1 do
       	begin
       		pt := mpFile.RegionPts[i];
           	data[pt.x + pt.y * mpFile.FrameWidth] := tempData[i];
       	end;
   	end;
   end
   else
   begin
		{Load frame in temporary buffer}
   	if mpFile.ScanMode <> SM_REGIONSCAN then
   		OleCheck(mpFile.chStreams[chIndex].Read(@mpFile.motionBuffer[0], mpFile.FrameSize * SizeOf(int16), nil))
   	else
   	begin
   		{Region scan: the frame is 512 x 512, we fill with data indexed by RegionPts}
			FillChar(mpFile.motionBuffer[0], mpFile.FrameWidth * mpFile.FrameHeight * SizeOf(int16), 0);
   		SetLength(tempData, mpFile.FrameSize);
   		OleCheck(mpFile.chStreams[chIndex].Read(@tempData[0], mpFile.FrameSize * SizeOf(int16), nil));
       	for i := 0 to mpFile.FrameSize - 1 do
       	begin
       		pt := mpFile.RegionPts[i];
           	mpFile.motionBuffer[pt.x + pt.y * mpFile.FrameWidth] := tempData[i];
       	end;
   	end;
		{***** Shift the frame for motion correction ******}
       with mpFile do
       begin
       	FillChar(data[0], FrameWidth * FrameHeight * SizeOf(int16), 0); {fills scan line with zeros}
			if motionDeltas[frameIndex].y >= 0 then
           	for j := 0 to FrameHeight - 1 - motionDeltas[frameIndex].y do
				begin
               	if motionDeltas[frameIndex].x >= 0 then
                   begin
                   	sourceFirstIndex := (j + motionDeltas[frameIndex].y) * FrameWidth
                       	+ motionDeltas[frameIndex].x;
                       destFirstIndex := j * FrameWidth;
               		Move(motionBuffer[sourceFirstIndex], data[destFirstIndex],
               			(FrameWidth - motionDeltas[frameIndex].x) * SizeOf(int16));
                   end
                   else
                   begin
                   	sourceFirstIndex := (j + motionDeltas[frameIndex].y) * FrameWidth;
                       destFirstIndex := j * FrameWidth - motionDeltas[frameIndex].x;
               		Move(motionBuffer[sourceFirstIndex], data[destFirstIndex],
               			(FrameWidth + motionDeltas[frameIndex].x) * SizeOf(int16));
                   end
               end
           else // motionDeltas[frameIndex].y < 0
           	for j := - motionDeltas[frameIndex].y to FrameHeight - 1  do
				begin
               	if motionDeltas[frameIndex].x >= 0 then
                   begin
                   	sourceFirstIndex := (j + motionDeltas[frameIndex].y) * FrameWidth
                       	+ motionDeltas[frameIndex].x;
                       destFirstIndex := j * FrameWidth;
               		Move(motionBuffer[sourceFirstIndex], data[destFirstIndex],
               			(FrameWidth - motionDeltas[frameIndex].x) * SizeOf(int16));
                   end
                   else
                   begin
                   	sourceFirstIndex := (j + motionDeltas[frameIndex].y) * FrameWidth;
                       destFirstIndex := j * FrameWidth - motionDeltas[frameIndex].x;
               		Move(motionBuffer[sourceFirstIndex], data[destFirstIndex],
               			(FrameWidth + motionDeltas[frameIndex].x) * SizeOf(int16));
                   end
               end;
       end; // with mpFile
   end; // mpFile.MotionCorrectionEnabled
end;


procedure   TVideoFrame.SaveToFile;
begin
   mpFile.chStreams[chIndex].Write(@data[0], Width * Height * SizeOf(int16), nil);
end;


procedure   TVideoFrame.CopyData(dest: TVideoFrame);
var    i: integer;
begin
   for i := 0 to Width * Height - 1 do
       dest.data[i] := data[i];
end;


procedure   TVideoFrame.GammaCorrection(blackLevel, whiteLevel: integer);
var    i: integer;
       maxPixelVal, pixelVal: int16;
begin
	maxPixelVal := mpFile.ChMaxPixelValues[chIndex];
   for i := 0 to Width * Height - 1 do
   begin
       pixelVal := data[i];
       if pixelVal > whiteLevel then
           pixelVal := maxPixelVal
       else if pixelVal < blackLevel then
           pixelVal := 0
       else
           pixelVal := int16(Muldiv(integer(pixelVal) - blackLevel, maxPixelVal,
               whiteLevel - blackLevel));
       data[i] := pixelVal;
   end;
end;


procedure    TVideoFrame.GetProfile(binSize: integer; var histogram: array of integer);
var    binCount, i, j: integer;
begin
   if (binSize <= 0) or (binSize > 1024) then Exit;
   binCount := DEFAULT_MAX_PIXEL_VALUE + 1 div binSize + 1;
  { SetLength(histogram, binCount);}
   for i := 0 to binCount - 1 do
       histogram[i] := 0;

   for i := 0 to Width * Height - 1 do
       for j := 0 to binCount - 1 do
       begin
           if j = 0 then
           begin
               if data[i] < 0 then histogram[0] := histogram[0] + 1;
           end
           else
               if (data[i] >= (j - 1) * binSize) and (data[i] < (j * binSize)) then
                   histogram[j] := histogram[j] + 1;
       end;
end;


procedure   TVideoFrame.CopyToDoubleArray(var dblArray: array of double);
var    i: integer;
begin
   for i := 0 to Width * Height - 1 do
       dblArray[i] := data[i];
end;


constructor TVideoFrame.Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
begin
	inherited Create(mp_File, ichIndex, iFrameIndex);
   SetLength(data, mpFile.FrameWidth * mpFile.FrameHeight);
end;

///------------------------------ ANALOG FRAME  ---------------------------------

function    TAnalogFrame.SampleToAbsoluteTime(x: integer): double; {in seconds}
begin
   Result := (frameIndex + x / mpFile.ChDataPtsPerFrame[chIndex]) / mpFile.FrameRate;
end;


function    TAnalogFrame.GetIAnalogValue(x: integer): integer;
begin
   Result := data[x];
end;


function    TAnalogFrame.GetfAnalogValue(x: integer): double;
begin
   Result := mpFile.DigitalToAnalog(chIndex, data[x]);
end;


procedure   TAnalogFrame.LoadFromFile(frameIndex: integer);
var    libNewPosition: Largeint;
begin
   mpFile.chStreams[chIndex].Seek(frameIndex * mpFile.ChDataPtsPerFrame[chIndex] * SizeOf(int16),
   	STREAM_SEEK_SET, libNewPosition);
   mpFile.chStreams[chIndex].Read(@data[0], mpFile.ChDataPtsPerFrame[chIndex] * SizeOf(int16), nil);
end;


constructor TAnalogFrame.Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
begin
	inherited Create(mp_File, ichIndex, iFrameIndex);
   SetLength(data, mpFile.fChDataPtsPerFrames[iChIndex]);
end;


//--------------------------- FRAME COMMENT LIST  ------------------------------

procedure   TFrameCommentList.SaveFrameComments;
var    commentStream: IStream;
       i, j: integer;
       s: string;
begin
   if Count > 0 then
       if Succeeded(mpFile.rootStorage.CreateStream('Frame comments', OFLAGS, 0, 0, commentStream)) then
       begin
           j := Count - 1;
           commentStream.Write(@j, SizeOf(j), nil);
           for i := 0 to Count - 1 do
           begin
               j := Integer(Objects[i]);
               commentStream.Write(@j, SizeOf(j), nil);
               j := Length(Strings[i]);
               commentStream.Write(@j, SizeOf(j), nil);
               s := Strings[i];
               commentStream.Write(@s[1], j, nil);
           end;
           commentStream := nil;
       end;
end;


procedure   TFrameCommentList.LoadFrameComments;
var    commentStream: IStream;
       i, frameIndex, stringLength, frameCount: integer;
       s: string;
begin
   if Succeeded(mpFile.rootStorage.OpenStream('Frame comments', nil, RFLAGS, 0, commentStream)) then
   begin
       commentStream.Read(@frameCount, SizeOf(frameCount), nil);
       if frameCount >= 0 then
       for i := 0 to frameCount do
       begin
           commentStream.Read(@frameIndex, SizeOf(frameIndex), nil);
           commentStream.Read(@stringLength, SizeOf(stringLength), nil);
           if stringLength > 0 then
           begin
               SetLength(s, stringLength);
               commentStream.Read(@s[1], stringLength, nil);
               AddObject(s, TObject(frameIndex));
           end;
       end;
       commentStream := nil;
   end;
end;


function    TFrameCommentList.GetFrameComment(frameIndex: integer): string;
begin
   Result := '';
   if (Count > 0) and (IndexOfObject(TObject(frameIndex)) >= 0) then
      Result := Strings[IndexOfObject(TObject(frameIndex))];
end;


procedure   TFrameCommentList.SetFrameComment(frameIndex: integer; const frameComment: string);
begin
   if IndexOfObject(TObject(frameIndex)) >= 0 then
       Strings[IndexOfObject(TObject(frameIndex))] := frameComment
   else
       AddObject(frameComment, TObject(frameIndex));
end;


constructor TFrameCommentList.Create(theMPFile: TMPFile);
begin
   mpFile := theMPFile;
end;


//------------------------------ FRAME GROUP  ----------------------------------

procedure   TFrameGroup.LoadFromFile(frameIndex: integer);
var    i: integer;
begin
   for i := 0 to MAX_CH - 1 do
       if channels[i] <> nil then
           channels[i].LoadFromFile(frameIndex);
end;


procedure   TFrameGroup.SetFrameIndex(newIndex: integer);
var    i: integer;
begin
   fFrameIndex := newIndex;
   for i := 0 to MAX_CH - 1 do
       if channels[i] <> nil then channels[i].frameIndex := newIndex;
end;


procedure   TFrameGroup.SetFrameWidth(newWidth: integer);
begin
   fWidth := newWidth;
end;


procedure   TFrameGroup.SetFrameHeight(newHeight: integer);
begin
   fHeight := newHeight;
   if channels[0] <> nil then SetLength(channels[0].data, fWidth * fHeight);
   if channels[1] <> nil then SetLength(channels[1].data, fWidth * fHeight);
   if mpFile.IsVideoChannel[2] and (channels[2] <> nil) then SetLength(channels[2].data, fWidth * fHeight);
   if mpFile.IsVideoChannel[3] and (channels[3] <> nil) then SetLength(channels[3].data, fWidth * fHeight);
   if mpFile.AnalogChEnabled[2] and (channels[2] <> nil) then SetLength(channels[2].data, mpFile.ChDataPtsPerFrame[2]);
   if mpFile.AnalogChEnabled[3] and (channels[3] <> nil) then SetLength(channels[3].data, mpFile.ChDataPtsPerFrame[3]);
end;


constructor TFrameGroup.Create(mp_file: TMPFile);
var    i: integer;
begin
   mpFile := mp_file;
   for i := 0 to MAX_CH - 1 do
       if mpFile.VideoChEnabled[i] then
       	channels[i] := TVideoFrame.Create(mpFile, i, fFrameIndex)
       else if mpFile.AnalogChEnabled[i] then
			channels[i] := TAnalogFrame.Create(mpFile, i, fFrameIndex);
end;


constructor TFrameGroup.CreateTempFrameGroup(mp_file: TMPFile);
begin
   mpFile := mp_file;
   channels[0] := TVideoFrame.Create(mpFile, 0, 0);
 {  SetLength(channels[0].data, width * height);}
end;


constructor TFrameGroup.CreateForOp(mp_file: TMPFile);
begin
   mpFile := mp_file;
   channels[0] := TVideoFrame.Create(mpFile, 0, fFrameIndex);
 {  SetLength(channels[0].data, width * height);}
end;


destructor  TFrameGroup.Destroy;
var    i: integer;
begin
   for i := 0 to MAX_CH - 1 do
       channels[i].Free;
   inherited Destroy;
end;


//-------------------------------- FILE OBJECT --------------------------------

//--------------------------- FILE OBJECT: PRIVATE -----------------------------

procedure   TMPFile.AdjustFrameCount;
var    statstg: TStatStg;
       validCh: integer;
begin
   if VideoChEnabled[0] then validCh := 0
   else if VideoChEnabled[1] then validCh := 1
   else if VideoChEnabled[2] then validCh := 2
   else validCh := 3;
   chStreams[validCh].Stat(statstg, STATFLAG_NONAME);
   fFrameCount := statstg.cbSize div (FrameWidth * FrameHeight * SizeOf(int16));
end;


procedure	TMPFile.FillBaseColors(chIndex: integer; const s: string);
begin
	baseColors[chIndex, 0] := (Pos('Red', s) > 0);
	baseColors[chIndex, 1] := (Pos('Green', s) > 0);
	baseColors[chIndex, 2] := (Pos('Blue', s) > 0);
end;


function    TMPFile.GetAnalogChCount: integer;
begin
   Result := 0;
   if fChEnabled[2] and not fIsVideoCh[2] then Result := 1;
   if fChEnabled[3] and not fIsVideoCh[3] then Result := Result + 1;
end;


function	TMPFile.GetAnalogChEnabled(chIndex: integer): boolean;
begin
	if (chIndex <> 2) and (chIndex <> 3) then
   	Result := False
   else
   	Result := fChEnabled[chIndex] and not fIsVideoCh[chIndex];
end;


function    TMPFile.GetAnalogWndCount: integer;
begin
   if analogWndList <> nil then
       Result := analogWndList.Count
   else
       Result := 0;
end;


function    TMPFile.GetAnalogWnds(wndIndex: integer): TForm;
begin
   Result := nil;
   if (wndIndex >= 0) and (wndIndex < analogWndList.Count) then
       Result := TForm(analogWndList.Items[wndIndex]);
end;


function    TMPFile.GetChAcquisitionRate(chIndex: integer): double;
begin
   Result := FrameRate / fChDataPtsPerFrames[chIndex];
end;


function    TMPFile.GetChConvFactor(chIndex: integer): double;
begin
   Result := fChConvFactors[chIndex];
end;


function    TMPFile.GetChDataPtsPerFrame(chIndex: integer): integer;
begin
   Result := fChDataPtsPerFrames[chIndex];
end;


function    TMPFile.GetChEnabled(chIndex: integer): boolean;
begin
   Result := fChEnabled[chIndex];
end;


function    TMPFile.GetChInputRange(chIndex: integer): TFullScaleVal;
begin
   Result := fChInputRanges[chIndex];
end;


function    TMPFile.GetChMaxPixelValues(chIndex: integer): integer;
begin
	Result := fChMaxPixelValues[chIndex];
end;


function    TMPFile.GetChNames(chIndex: integer): string;
begin
   Result := fChNames[chIndex];
end;


function    TMPFile.GetChUnit(chIndex: integer): string;
begin
   Result := fChUnits[chIndex];
end;


function    TMPFile.GetChPrefix(chIndex: integer): TPrefix;
begin
   Result := fChPrefixes[chIndex];
end;


function    TMPFile.GetChOffset(chIndex: integer): double;
begin
   Result := fChOffsets[chIndex];
end;


function    TMPFile.GetDefaultVideoChannel: integer;
begin
	if VideoChEnabled[0] then Result := 0
   else if VideoChEnabled[1] then Result := 1
   else if VideoChEnabled[2] then Result := 2
   else Result := 3;
end;


function    TMPFile.GetFrameComment(frameIndex: integer): string;
begin
   Result := frameCommentList.FrameComment[frameIndex];
end;


function    TMPFile.GetFrameCount: integer;
begin
   if IsMemoryFile then
       Result := frameGroupList.Count
   else
       Result := fFrameCount;
end;


function    TMPFile.GetFrameRate: double; {in frames per seconds}
begin
   {50 ns = 5e-8 s per pixel clock unit}
   if ScanMode <> SM_REGIONSCAN then
   	Result := 1 / (Muldiv(5, fFrameWidth, 4) * fFrameHeight * fPixelClock * BASE_CLOCK)
   else
   	Result := 1 / (FrameSize * fPixelClock * BASE_CLOCK);
end;


function	 TMPFile.GetFrameSize: integer;
begin
	if ScanMode <> SM_REGIONSCAN then
   	Result := FrameWidth * FrameHeight
   else
   	Result := Length(RegionPts);
end;


function    TMPFile.GetFrames(frameIndex: integer): TFrameGroup;
begin
   Result := nil;
   if (frameIndex >= 0) and (frameIndex < FrameCount) then
       if IsMemoryFile then
           Result := TFrameGroup(frameGroupList.Items[frameIndex])
       else
       begin
           ActiveFrameIndex := frameIndex;
           Result := fCurrentFrameGroup;
       end;
end;

{returns the size of the frame if turnaround points (20% of number of pixels in
line scans)are included}
function    TMPFile.GetFullFrameSize: integer;
begin
   case ScanMode of
		SM_MOVIE, SM_LINESCAN, SM_STACK, SM_FASTSTACK: Result := Muldiv(5, FrameSize, 4);
   	else Result := FrameSize;
   end;
end;


function	TMPFile.GetIsVideoChannel(chIndex: integer): boolean;
begin
    if (chIndex = 0) or (chIndex = 1) then
   	    Result := True
    else if (chIndex = 2) or (chIndex = 3) then
        Result := fIsVideoCh[chIndex]
    else
        Result := False;
end;


function	TMPFile.GetMaxPixelValue: integer;
begin
	case fResolution of
   	RESOLUTION_8_BITS: Result := 255;
       RESOLUTION_12_BITS: Result := DEFAULT_MAX_PIXEL_VALUE;
       else Result := 32767;
   end;
end;


procedure	TMPFile.GetRegions;
var	regionStream: IStream;
		statstg: TStatStg;
       regionSize: integer;
	    libNewPosition: Largeint;
begin
	{opens the stream 'Regions'}
   OleCheck(rootStorage.OpenStream('Regions', nil, RFLAGS, 0, regionStream));
   {gets the size of the stream}
   regionStream.Stat(statstg, STATFLAG_NONAME);
   regionSize := statStg.cbSize;
   OleCheck(regionStream.Seek(0, STREAM_SEEK_SET, libNewPosition));
   SetLength(RegionPts, regionSize div SizeOf(TPoint));
   OleCheck(regionStream.Read(@RegionPts[0], regionSize, nil));
end;


function	 TMPFile.GetResolution: string;
begin
	case fResolution of
   	RESOLUTION_8_BITS: Result := '8-bit';
       RESOLUTION_12_BITS: Result := '12-bit';
       else Result := '16-bit';
   end;
end;


function    TMPFile.GetVideoChCount: integer;
begin
   Result := 0;
   if fChEnabled[0] then Result := 1;
   if fChEnabled[1] then Result := Result + 1;
   if fChEnabled[2] and fIsVideoCh[2] then Result := Result + 1;
   if fChEnabled[3] and fIsVideoCh[3] then Result := Result + 1;
end;


function	TMPFile.GetVideoChEnabled(chIndex: integer): boolean;
begin
	if (chIndex = 0) or (chIndex = 1) then
		Result := fChEnabled[chIndex]
   else
		Result := fChEnabled[chIndex] and  fIsVideoCh[chIndex];
end;


function    TMPFile.GetViewerCount: integer;
begin
   Result := viewerList.Count;
end;


function    TMPFile.GetViewers(viewerIndex: integer): TForm;
begin
   if (viewerIndex >= 0) and (viewerIndex < viewerList.Count) then
       Result := TForm(viewerList.Items[viewerIndex])
   else
       Result := nil;
end;


procedure    TMPFile.InitializeColors;
var	i: integer;
begin
   baseColors[0][0] := True; {Red}
   baseColors[0][1] := False;
   baseColors[0][2] := False;

   baseColors[1][0] := False;
   baseColors[1][1] := True;
   baseColors[1][2] := False;

   baseColors[2][0] := False;
   baseColors[2][1] := False;
   baseColors[2][2] := True;

   baseColors[3][0] := False;
   baseColors[3][1] := True;
   baseColors[3][2] := True;

   negativeColors[0].rgbtRed := 0; negativeColors[0].rgbtGreen := 0; negativeColors[0].rgbtBlue := 255;
   midRangeColors[0].rgbtRed := 0; midRangeColors[0].rgbtGreen := 255; midRangeColors[0].rgbtBlue := 0;
   maxColors[0].rgbtRed := 255; maxColors[0].rgbtGreen := 0; maxColors[0].rgbtBlue := 0;

   negativeColors[1].rgbtRed := 0; negativeColors[1].rgbtGreen := 255; negativeColors[1].rgbtBlue := 0;
   midRangeColors[1].rgbtRed := 0; midRangeColors[1].rgbtGreen := 255; midRangeColors[1].rgbtBlue := 255;
   maxColors[1].rgbtRed := 255; maxColors[1].rgbtGreen := 0; maxColors[1].rgbtBlue := 0;

   negativeColors[2].rgbtRed := 0; negativeColors[2].rgbtGreen := 255; negativeColors[2].rgbtBlue := 0;
   midRangeColors[2].rgbtRed := 0; midRangeColors[2].rgbtGreen := 255; midRangeColors[2].rgbtBlue := 255;
   maxColors[2].rgbtRed := 255; maxColors[2].rgbtGreen := 0; maxColors[2].rgbtBlue := 0;

   negativeColors[3].rgbtRed := 0; negativeColors[3].rgbtGreen := 255; negativeColors[3].rgbtBlue := 0;
   midRangeColors[3].rgbtRed := 0; midRangeColors[3].rgbtGreen := 255; midRangeColors[3].rgbtBlue := 255;
   maxColors[3].rgbtRed := 255; maxColors[3].rgbtGreen := 0; maxColors[3].rgbtBlue := 0;

	for i := 0 to MAX_CH - 1 do MaxPixels[i] := 2047;
end;

procedure   TMPFile.LoadFrameGroup(newIndex: integer);
begin
   fCurrentFrameGroup.LoadFromFile(newIndex);
end;


procedure	 TMPFile.LoadMotionCorrection;
var	motionStream: IStream;
	    libNewPosition: Largeint;
begin
	if m_IPP_Loaded = 0 then {makes sure the Intel libraries are present}
   if Succeeded(rootStorage.OpenStream(sMotionCorrection, nil, RFLAGS, 0, motionStream)) then
   begin
       SetLength(motionBuffer, FrameSize);
   	SetLength(motionDeltas, FrameCount);
   	if Succeeded (motionStream.Seek(0, STREAM_SEEK_SET, libNewPosition)) then
   		if Succeeded(motionStream.Read(@motionDeltas[0], FrameCount * SizeOf(TPoint), nil)) then
           	fbMotionCorrectionEnabled := True;
   end;
end;


type
   TPropSpecArray = array[0..1000] of TPropSpec;
   TPropVariantArray = array[0..1000] of TPropVariant;
   TpPropSpecArray = ^TPropSpecArray;
   TpPropVariantArray = ^TPropVariantArray;

const	RGBTriple_Black: TRGBTriple = ();


function StringToRGBTriple(const s: string): TRGBTriple;
type	TQuadByte = packed record
			a,b,c,d: Byte
		end;
var	i: integer;
begin
	try
   	i := StrToInt('$' + s);
       Result.rgbtBlue := TQuadByte(i).a;
       Result.rgbtGreen := TQuadByte(i).b;
       Result.rgbtRed := TQuadByte(i).c;
	except
   	Result := RGBTriple_Black;
   end;
end;


function IntensityControlToString(ic: TIntensityControl): string;
begin
	if ic = IC_NO_CONTROL then
   	Result := 'No Control'
   else if ic = IC_LINEAR then
   	Result := 'Linear'
   else
   	Result := 'Exponential';
end;


function StringToIntensityControl(s: string): TIntensityControl;
begin
	if s = 'No Control' then
   	Result := IC_NO_CONTROL
   else if s = 'Linear' then
   	Result := IC_LINEAR
   else
   	Result := IC_EXPONENTIAL;
end;



procedure   TMPFile.ReadPropertySet;
var    ps: TpPropSpecArray;
       pv: TpPropVariantArray;
       i: integer;
       s: string;
begin
   ps := nil; pv := nil;
   try
       GetMem(ps, FILE_PROP_COUNT * SizeOf(TPropSpec));
       GetMem(pv, FILE_PROP_COUNT * SizeOf(TPropVariant));
       for i := 0 to FILE_PROP_COUNT - 1 do
       begin
           ps^[i].ulKind := PRSPEC_LPWSTR;
           ps^[i].lpwstr := StringToOLEStr(sPropNames[i]);
       end;
       OleCheck(propertyStorage.ReadMultiple(FILE_PROP_COUNT, @ps[0], @pv[0]));

       if (pv^[0].pwszVal <> nil) and (WideCharToString(pv^[0].pwszVal) = '16-bit') then
			fResolution := RESOLUTION_16_BITS
       else fResolution := RESOLUTION_12_BITS;

       if fResolution = RESOLUTION_12_BITS then
       	for i := 0 to MAX_CH - 1 do fChMaxPixelValues[i] := DEFAULT_MAX_PIXEL_VALUE
       else
       	for i := 0 to MAX_CH - 1 do fChMaxPixelValues[i] := 32767;

		if pv^[1].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[1].pwszVal); {Scan Mode}
       fScanMode := StringToScanMode(s);
{       s := WideCharToString(pv^[2].pwszVal);} {Line Scan Mode}
{       fbLineForXScan := StringToLineForXScan(s);}

		if pv^[2].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[2].pwszVal); {X Frame Offset'}
       if s = '' then s := '0'; fXFrameOffset := StrToInt(s);

		if pv^[3].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[3].pwszVal); {Y Frame Offset'}
       if s = '' then s := '0'; fYFrameOffset := StrToInt(s);

		if pv^[4].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[4].pwszVal); {Frame Width'}
       if s = '' then s := '0'; fFrameWidth := StrToInt(s);

		if pv^[5].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[5].pwszVal); {Frame Height}
       if s = '' then s := '0'; fFrameHeight := StrToInt(s);

		if pv^[6].pwszVal = nil then s := '20' else
       s := WideCharToString(pv^[6].pwszVal); {Pixel Clock}
       if s = '' then s := '0'; fPixelClock := StrToInt(s);

		if pv^[7].pwszVal = nil then fChNames[0] := 'Channel 1' else
       fChNames[0] := WideCharToString(pv^[7].pwszVal); {Channel name (1)}

		if pv^[8].pwszVal = nil then fChEnabled[0] := False else
       fChEnabled[0] := StringToBool(WideCharToString(pv^[8].pwszVal)); {Enabled (1)}

		if pv^[9].pwszVal = nil then fChInputRanges[0] := pm_10V else
       fChInputRanges[0] := StringToInputRange(WideCharToString(pv^[9].pwszVal)); {Input Range  (1)}

		if pv^[10].pwszVal = nil then fChNames[1] := 'Channel 2' else
       fChNames[1] := WideCharToString(pv^[10].pwszVal); {Channel name (2)}

		if pv^[11].pwszVal = nil then fChEnabled[1] := False else
       fChEnabled[1] := StringToBool(WideCharToString(pv^[11].pwszVal)); {Enabled (2)}

		if pv^[12].pwszVal = nil then fChInputRanges[1] := pm_10V else
       fChInputRanges[1] := StringToInputRange(WideCharToString(pv^[12].pwszVal)); {Input Range (2)}

		if pv^[13].pwszVal = nil then fChNames[2] := 'Channel 3' else
       fChNames[2] := WideCharToString(pv^[13].pwszVal); {Channel name (3)}

		if pv^[14].pwszVal = nil then fChEnabled[2] := False else
       fChEnabled[2] := StringToBool(WideCharToString(pv^[14].pwszVal)); {Enabled (3)}

		if pv^[15].pwszVal = nil then fChInputRanges[2] := pm_10V else
       fChInputRanges[2] := StringToInputRange(WideCharToString(pv^[15].pwszVal)); {Input Range (3)}

		if pv^[16].pwszVal = nil then fChInputRanges[2] := pm_10V else
       fChUnits[2] := WideCharToString(pv^[16].pwszVal); {Channel Unit (3)}

		if pv^[17].pwszVal = nil then fChPrefixes[2] := tpUNITY else
       fChPrefixes[2] := StringToPrefix(WideCharToString(pv^[17].pwszVal)); {Channel Prefix (3)}

		if pv^[18].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[18].pwszVal); {Conversion factor (3)}
       if s = '' then s := '0'; fChConvFactors[2] := StrToFloat(s);

		if pv^[19].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[19].pwszVal); {Offset (3)}
       if s = '' then s := '0'; fChOffsets[2] := StrToFloat(s);


		if pv^[20].pwszVal = nil then s := '' else
       s := WideCharToString(pv^[20].pwszVal); {Data Points Per Frame (3)}
       if s = '' then s := '0'; fChDataPtsPerFrames[2] := StrToInt(s);

		if pv^[21].pwszVal = nil then fChNames[3] := 'Channel 4' else
       fChNames[3] := WideCharToString(pv^[21].pwszVal); {Channel name (4)}

		if pv^[22].pwszVal = nil then fChEnabled[3] := False else
       fChEnabled[3] := StringToBool(WideCharToString(pv^[22].pwszVal)); {Enabled (4)}

		if pv^[23].pwszVal = nil then fChInputRanges[3] := pm_10V else
       fChInputRanges[3] := StringToInputRange(WideCharToString(pv^[23].pwszVal)); {Input Range (4)}

		if pv^[24].pwszVal = nil then fChUnits[3] := 'V' else
       fChUnits[3] := WideCharToString(pv^[24].pwszVal); {Channel Unit (4)}

		if pv^[25].pwszVal = nil then fChPrefixes[3] := tpUNITY else
       fChPrefixes[3] := StringToPrefix(WideCharToString(pv^[25].pwszVal)); {Channel Prefix (4)}

		if pv^[26].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[26].pwszVal); {Conversion factor (4)}
       if s = '' then s := '0'; fChConvFactors[3] := StrToFloat(s);

		if pv^[27].pwszVal = nil then s := '0' else
       s := WideCharToString(pv^[27].pwszVal); {Offset (4)}
       if s = '' then s := '0'; fChOffsets[3] := StrToFloat(s);

		if pv^[28].pwszVal = nil then s := '0' else
       s := WideCharToString(pv^[28].pwszVal); {Data Points Per Frame (4)}
       if s = '' then s := '0'; fChDataPtsPerFrames[3] := StrToInt(s);

		if pv^[29].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[29].pwszVal); {Frame Count'}
       if s = '' then s := '0'; fFrameCount := StrToInt(s);

		if pv^[30].pwszVal = nil then s := '0' else
       s := WideCharToString(pv^[30].pwszVal); {X Position'}
       if s = '' then s := '0'; fXStagePosition := StrToInt(s);

		if pv^[31].pwszVal = nil then s := '0' else
       s := WideCharToString(pv^[31].pwszVal); {Y Position'}
       if s = '' then s := '0'; fYStagePosition := StrToInt(s);

		if pv^[32].pwszVal = nil then s := '0' else
       s := WideCharToString(pv^[32].pwszVal); {Z Position'}
       if s = '' then s := '0'; fZStagePosition := Round(StrToFloat(s));

		if pv^[33].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[33].pwszVal); {Stack Count'}
       if s = '' then s := '0'; fStackFrameCount := StrToInt(s);

		if pv^[34].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[34].pwszVal); {z- Interval'}
       if s = '' then s := '0'; fStackInterval := StrToFloat(s);

		if pv^[35].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[35].pwszVal); {Averaging Count'}
       if s = '' then s := '0'; fStackAveragingCount := StrToInt(s);

		if pv^[36].pwszVal = nil then s := '1' else
       s := WideCharToString(pv^[36].pwszVal); {Repeat Count'}
       if s = '' then s := '0'; fStackRepeatCount := StrToInt(s);

		if pv^[37].pwszVal = nil then s := 'x1.0' else
       s := WideCharToString(pv^[37].pwszVal); {Magnification - remove 'x'}
       s := Copy(s, 2, Length(s) - 1); if s = '' then s := '1'; fMagnification := StrToFloat(s);

		if pv^[38].pwszVal = nil then s := '0' else
       s := WideCharToString(pv^[38].pwszVal); {Rotation'}
       if s = '' then s := '0'; fRotation := StrToInt(s);

		if pv^[39].pwszVal = nil then sComments := '' else
       sComments := WideCharToString(pv^[39].pwszVal); {Comments}

		if pv^[40].pwszVal = nil then fIsVideoCh[2] := False else
       	fIsVideoCh[2] := StringToBool(
           	WideCharToString(pv^[40].pwszVal)); {Is Video Channel (3)}

		if pv^[41].pwszVal = nil then fIsVideoCh[3] := False else
       	fIsVideoCh[3] := StringToBool(
           	WideCharToString(pv^[41].pwszVal)); {Is Video Channel (4)}

		if pv^[42].pwszVal = nil then fzDistance := 0 else
       	fzDistance := StrToFloat(
           	WideCharToString(pv^[42].pwszVal)); {z- Distance}

		if pv^[43].pwszVal = nil then fTravelDuration := '0 s' else
       	fTravelDuration := WideCharToString(pv^[43].pwszVal); {Travel Duration}

		if pv^[44].pwszVal = nil then fFastStackRepeatCount := 0 else
       	fFastStackRepeatCount := StrToInt(
           	WideCharToString(pv^[44].pwszVal)); {Fast Stack Repeat Count}

		if pv^[45].pwszVal = nil then fInitialIntensity := 0 else
       	fInitialIntensity := StrToInt(
           	WideCharToString(pv^[45].pwszVal)); {Initial Intensity}

		if pv^[46].pwszVal = nil then fFinalIntensity := 0 else
       	fFinalIntensity := StrToInt(
           	WideCharToString(pv^[46].pwszVal)); {Final Intensity}

		if pv^[47].pwszVal = nil then fAtZDistance := 0 else
       	fAtZDistance := StrToFloat(
           	WideCharToString(pv^[47].pwszVal)); {At Z- Distance}

		if pv^[48].pwszVal = nil then fIntensityControl := IC_NO_CONTROL else
       	fIntensityControl := StringToIntensityControl(
           	WideCharToString(pv^[48].pwszVal)); {Intensity Control}

		for i := 0 to MAX_CH - 1 do
			if pv^[49 + i].pwszVal <> nil then
       		try
       			fChMaxPixelValues[i] := StrToInt(WideCharToString(pv^[49 + i].pwszVal));
       		except
					fChMaxPixelValues[i] := DEFAULT_MAX_PIXEL_VALUE;
       		end;

		if pv^[53].pwszVal = nil then fColorScheme := CS_GRAYSCALE else
       	fColorScheme := StringToColorScheme(
           	WideCharToString(pv^[53].pwszVal)); {LUT type}

		for i := 0 to MAX_CH - 1 do
			if pv^[54 + i].pwszVal <> nil then
				FillBaseColors(i, WideCharToString(pv^[54 + i].pwszVal));

		for i := 0 to MAX_CH - 1 do
			if pv^[58 + i].pwszVal = nil then
           	negativeColors[i] := RGBTriple_Black
           else
           	negativeColors[i] := StringToRGBTriple(WideCharToString(pv^[58 + i].pwszVal));

		for i := 0 to MAX_CH - 1 do
			if pv^[62 + i].pwszVal = nil then
           	midRangeColors[i] := RGBTriple_Black
           else
           	midRangeColors[i] := StringToRGBTriple(WideCharToString(pv^[62 + i].pwszVal));

		for i := 0 to MAX_CH - 1 do
			if pv^[66 + i].pwszVal = nil then
           	maxColors[i] := RGBTriple_Black
           else
           	maxColors[i] := StringToRGBTriple(WideCharToString(pv^[66 + i].pwszVal));

		for i := 0 to MAX_CH - 1 do
			if pv^[70 + i].pwszVal = nil then
           	maxPixels[i] := 2047
           else
           	try
               	s := WideCharToString(pv^[70 + i].pwszVal);
           		maxPixels[i] := StrToInt(s);
                   if (maxPixels[i] <= 0) or (maxPixels[i] > 2047) then maxPixels[i] := 2047;
               except
               	maxPixels[i] := 2047;
               end;
   finally
       if ps <> nil then Freemem(ps);
       if pv <> nil then Freemem(pv);
   end;
end;


procedure   TMPFile.SetActiveFrameIndex(newIndex: integer);
begin
   if (newIndex >= 0) and (newIndex < FrameCount) then
       if newIndex <> fActiveFrameIndex then
       begin
           if not IsMemoryFile then LoadFrameGroup(newIndex);
           fActiveFrameIndex := newIndex;
       end;
end;


procedure    TMPFile.SetFrameComment(frameIndex: integer; sComments: string);
begin
   frameCommentList.FrameComment[frameIndex] := sComments;
end;



procedure    TMPFile.SetMotionCorrectionEnabled(flag: boolean);
begin
	if IsMemoryFile then flag := False;
   fbMotionCorrectionEnabled := flag;
	if flag then
   begin
       SetLength(motionBuffer, FrameSize);
       SetLength(motionDeltas, FrameCount);
       FillChar(motionDeltas[0], FrameCount * SizeOf(TPoint), 0);
   end;
end;


procedure   TMPFile.WritePropertySet;
var    ps: TpPropSpecArray;
       pv: TpPropVariantArray;
       i: integer;
       sPropValues: array[0..FILE_PROP_COUNT - 1] of string;
begin
	sPropValues[0] := Resolution;
   sPropValues[1] := ScanModeToString(ScanMode); {Scan Mode}
   sPropValues[2] := IntToStr(XFrameOffset);  {X Frame Offset'}
   sPropValues[3] := IntToStr(YFrameOffset);  {Y Frame Offset'}
   sPropValues[4] := IntToStr(FrameWidth);  {Frame Width'}
   sPropValues[5] := IntToStr(FrameHeight);  {Frame Height}
   sPropValues[6] := IntToStr(PixelClock);  {Pixel Clock}
   sPropValues[7] := ChNames[0]; {Channel name (1)}
   sPropValues[8] := BoolToString(ChEnabled[0]); {Enabled (1)}
   sPropValues[9] := InputRangeToString(ChInputRange[0]); {Input Range  (1)}
   sPropValues[10] := ChNames[1]; {Channel name (2)}
   sPropValues[11] := BoolToString(ChEnabled[1]); {Enabled (2)}
   sPropValues[12] := InputRangeToString(ChInputRange[1]); {Input Range (2)}
   sPropValues[13] := ChNames[2]; {Channel name (3)}
   sPropValues[14] := BoolToString(ChEnabled[2]); {Enabled (3)}
   sPropValues[15] := InputRangeToString(ChInputRange[2]); {Input Range (3)}
   sPropValues[16] := ChUnit[2]; {Channel Unit (3)}
   sPropValues[17] := PrefixToString(ChPrefix[2]); {Channel Prefix (3)}
   sPropValues[18] := FloatToStr(ChConvFactor[2]); {Conversion factor (3)}
   sPropValues[19] := FloatToStr(ChOffset[2]); {Offset (3)}
   sPropValues[20] := IntToStr(ChDataPtsPerFrame[2]); {Data Points Per Frame (3)}
   sPropValues[21] := ChNames[3]; {Channel name (4)}
   sPropValues[22] := BoolToString(ChEnabled[3]); {Enabled (4)}
   sPropValues[23] := InputRangeToString(ChInputRange[3]); {Input Range (4)}
   sPropValues[24] := ChUnit[3]; {Channel Unit (4)}
   sPropValues[25] := PrefixToString(ChPrefix[3]); {Channel Prefix (4)}
   sPropValues[26] := FloatToStr(ChConvFactor[3]); {Conversion factor (4)}
   sPropValues[27] := FloatToStr(ChOffset[3]);{Offset (4)}
   sPropValues[28] := IntToStr(ChDataPtsPerFrame[3]); {Data Points Per Frame (4)}

   sPropValues[29] := IntToStr(FrameCount); {Frame Count'}
   sPropValues[30] := IntToStr(XStagePosition); {X Position'}
   sPropValues[31] := IntToStr(YStagePosition); {Y Position'}
   sPropValues[32] := FloatToStr(ZStagePosition); {Z Position'}
   sPropValues[33] := IntToStr(StackFrameCount); {Stack Count'}
   sPropValues[34] := FloatToStr(StackInterval); {z- Interval'}

   sPropValues[35] := IntToStr(StackAveragingCount); {Averaging Count'}
   sPropValues[36] := IntToStr(StackRepeatCount);  {Repeat Count'}
   sPropValues[37] := 'x' + FloatToStr(Magnification); {Magnification'}
   sPropValues[38] := IntToStr(Rotation);  {Rotation'}
   sPropValues[39] := sComments; {Comments}

   sPropValues[40] := BoolToString(fIsVideoCh[2]); {Is Video Channel (3)}
   sPropValues[41] := BoolToString(fIsVideoCh[3]); {s Video Channel (4)}
   sPropValues[42] := FloatToStr(fzDistance); {z- Distance}
   sPropValues[43] := fTravelDuration; {Travel Duration}
   sPropValues[44] := IntToStr(fFastStackRepeatCount); {Fast Stack Repeat Count}
   sPropValues[45] := IntToStr(fInitialIntensity); {Initial Intensity}
   sPropValues[46] := IntToStr(fFinalIntensity); {Final Intensity}
   sPropValues[47] := FloatToStr(fAtZDistance); {At Z- Distance}
   sPropValues[48] := IntensityControlToString(fIntensityControl); {Intensity Control}

   for i := 0 to MAX_CH - 1 do
   	sPropValues[49 + i] := IntToStr(fChMaxPixelValues[i]);

   ps := nil; pv := nil;
   try
       GetMem(ps, FILE_PROP_COUNT * SizeOf(TPropSpec));
       GetMem(pv, FILE_PROP_COUNT * SizeOf(TPropVariant));
       for i := 0 to FILE_PROP_COUNT - 1 do
       begin
           ps^[i].ulKind := PRSPEC_LPWSTR;
           ps^[i].lpwstr := StringToOLEStr(sPropNames[i]);
           pv^[i].vt := VT_LPWSTR;
           pv^[i].pwszVal := StringToOLEStr(sPropValues[i]);
       end;
       OleCheck(propertyStorage.WriteMultiple(FILE_PROP_COUNT, @ps[0], @pv[0], 2));
   finally
       if ps <> nil then Freemem(ps);
       if pv <> nil then Freemem(pv);
   end;
end;


//--------------------------- FILE OBJECT: PUBLIC ------------------------------

procedure   TMPFile.AddFrameGroup(newFrameGroup: TFrameGroup);
begin
   if FrameCount = 0 then
   begin
       {allows overriding frame width and frame height when no frame in file}
       fFrameHeight := newFrameGroup.frameHeight;
       fFrameWidth := newFrameGroup.frameWidth; {sets the size of the data array}
   end;
   if (fFrameHeight = newFrameGroup.frameHeight) and (fFrameWidth = newFrameGroup.frameWidth) then
   begin
       newFrameGroup.FrameIndex := FrameCount;
       frameGroupList.Add(newFrameGroup);
   end;
end;


procedure   TMPFile.AverageFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       avgCount: integer;
       avgArray: array of integer;
       i, j: integer;
begin
   avgCount := toFrame - fromFrame + 1;

   SetLength(avgArray, FrameWidth * FrameHeight);
   for i := 0 to FrameWidth * FrameHeight - 1 do avgArray[i] := 0;

   newFrameGroup := TFrameGroup.CreateForOp(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := FrameHeight;

   for j := fromFrame to toFrame do
       for i := 0 to FrameWidth * FrameHeight - 1 do
           avgArray[i] := avgArray[i] + Frames[j].channels[chIndex].data[i];

   for i := 0 to FrameWidth * FrameHeight - 1 do
       newFrameGroup.channels[0].data[i] := avgArray[i] div avgCount;

   dest.AddFrameGroup(newFrameGroup);
end;


procedure	 TMPFile.BinaryOp(file1, file2: TMPFile; ch1, ch2, frame1, frame2: integer; gain1, gain2: double;
       				resultFrame: integer);
var	i, j: integer;
		dblOpResult: double;
       iOpResult: int16;
       frame1Obj, frame2Obj, resultFrameObj: TVideoFrame;
begin
	frame1Obj := file1.Frames[frame1].channels[ch1] as TVideoFrame;
	frame2Obj := file2.Frames[frame2].channels[ch2] as TVideoFrame;
   resultFrameObj := Frames[resultFrame].channels[0] as TVideoFrame;
	for j := 0 to FrameHeight - 1 do
   	for i := 0 to FrameWidth - 1 do
       begin
       	dblOpResult := gain1 * frame1Obj.Pixels[i, j] + gain2 * frame2Obj.Pixels[i, j];
       	if dblOpResult > DEFAULT_MAX_PIXEL_VALUE then
           	iOpResult := DEFAULT_MAX_PIXEL_VALUE
           else if dblOpResult < -DEFAULT_MAX_PIXEL_VALUE - 1 then
           	iOpResult := -DEFAULT_MAX_PIXEL_VALUE - 1
           else
           	iOpResult := Round(dblOpResult);
			resultFrameObj.data[i + j * FrameWidth] := iOpResult;
       end;
   OnNewFrames;
end;


procedure   TMPFile.Close(Sender: TObject);
var    i: integer;
begin
   fbClosing := True;
   {closes all the windows}
   if ViewerCount > 0 then
       for i := 0 to ViewerCount - 1 do
           Viewers[i].Close;
   if AnalogWndCount > 0 then
       for i := 0 to AnalogWndCount - 1 do
           AnalogWnds[i].Close;
   Free;
end;


function	 TMPFile.CreateEmptyFrame(fromFile: TMPFile): integer;
var    newFrameGroup: TFrameGroup;
begin
	newFrameGroup := TFrameGroup.CreateForOp(self);
   newFrameGroup.FrameWidth := fromFile.FrameWidth;
   newFrameGroup.FrameHeight := fromFile.FrameHeight;
	AddFrameGroup(newFrameGroup);
	Result := FrameCount - 1;
   OnNewFrames; {update each time we project - looks nicer!}
end;



procedure TMPFile.CopyChannelsToClipboard(bCh2, bCh3: boolean; fromFrame, toFrame: integer);
type   TCharArray = array[0..Maxint div 2 - 1] of WideChar;
var    chIndex, rowCount, frameIndex, sampleIndex: integer;
		timeVal, sampleVal: double;
       memhandle: THandle;
       pString: ^TCharArray;
       swLength, stringindex,maxcharcount: integer;
       s1: string;
       sw: array[0..179] of WideChar;
begin
	if not(bCh2 or bCh3) then Exit;
   if bCh2 then chIndex := 2 else chIndex := 3;

   rowCount := (toFrame - fromFrame + 1) * ChDataPtsPerFrame[chIndex];

	if rowCount <= 65536 then
   begin
       Clipboard.Open;
       Clipboard.Clear;

   	try
       {fills string with data; 11 digits for each column - each digit is a wide char, 2 columns}
       {number of rows is (toFrame - fromFrame + 1) * ChDataPtsPerFrame[]}
       {frameIndex: 0 to toFrame - fromFrame}
       {time in ms for each frame: (fromFrame + frameIndex) * FrameSize * PixelClock * 1e-6 }
       {sampleIndex: 0 to ChDataPtsPerFrame[] - 1}
       {for each sample: sampleIndex * FrameSize * PixelClock * 1e-6 / ChDataPtsPerFrame[]}
       maxcharcount := SizeOf(WideChar)* (
           			(2 {columns} * 20 {20 digits} + 3 {Tab, CR, LF}) * rowCount
               		+ 1 {Null}
               	); {wide char}

       	memHandle := GlobalAlloc(
       		GMEM_MOVEABLE or GMEM_DDESHARE or GMEM_ZEROINIT,
               maxcharcount);

           if memHandle <> 0 then
           begin
       		pString := GlobalLock(memHandle);
       		stringindex := 0;

       		for frameIndex := 0 to toFrame - fromFrame do
           	begin
           		ActiveFrameIndex := fromFrame + frameIndex;
       			for sampleIndex := 0 to ChDataPtsPerFrame[chIndex]- 1 do
           		begin
       				timeVal := 1000 * (sampleIndex / ChDataPtsPerFrame[chIndex] + fromFrame + frameIndex) / FrameRate;
           			{timeVal in ms; pixel clock is in increments of 50 ns = 5e-5 ms(20 MHz)}
           	   	{	timeVal := (FrameSize * PixelClock * 5e-5) *
               		(sampleIndex / ChDataPtsPerFrame[chIndex] + fromFrame + frameIndex);}
               		{sampleVal}
                   	sampleVal := ChConvFactor[chIndex] * FullScaleToVal(ChInputRange[chIndex]) *
           				Frames[ActiveFrameIndex].channels[chIndex].data[sampleIndex] /
           				(f_MAX_Y_VALUE + 1) + ChOffset[chIndex];

                       s1 := Format('%.6f', [timeVal]) + TAB + Format('%g', [sampleVal]) + CRLF;

           			swLength := Length(s1);
                       StringToWideChar(s1, @sw, SizeOf(WideChar) * (swLength + 1));
                       if stringIndex < maxcharcount div 2 - swLength then
                       begin
           		 		Move(sw[0], pString^[stringIndex], SizeOf(WideChar) * swLength);
           				stringIndex := stringIndex + swLength; {points to the next character}
                       end;
           		end;
           	end;

       		Clipboard.SetAsHandle(CF_UNICODETEXT, memHandle);
           end
           else
   			MessageDlg('Insufficient memory to export data; try saving data to disk as text file.' + CRLF +
                  '(Menu: File\Save\Analog Data As...)', mtError, [mbOK], 0);
   	finally
       	Clipboard.Close;
   	end;
   end {rowcount <= }
   else
   	MessageDlg('Too many data samples to export; try saving data to disk as text file.' + CRLF +
                  '(Menu: File\Save\Analog Data As...)', mtError, [mbOK], 0);
end;


procedure   TMPFile.CopyFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile); {from one file to a memory file, rect to rect}
var    newFrameGroup: TFrameGroup;
       i, j: integer;
begin
   for j := fromFrame to toFrame do
   begin
       newFrameGroup := TFrameGroup.CreateForOp(dest);
       newFrameGroup.FrameWidth := FrameWidth;
       newFrameGroup.FrameHeight := FrameHeight;
       for i := 0 to FrameWidth * FrameHeight - 1 do
           newFrameGroup.channels[0].data[i] := Frames[j].channels[chIndex].data[i];
       dest.AddFrameGroup(newFrameGroup);
   end;
end;


procedure	TMPFile.CloneROIs(ROIList: TROIList; fromIndex: integer); {duplicate each ROI but with opposite channel}
var 	i, j, oldCount, chIndex: integer;
		oldROI, newROI: TSimpleROI;
begin
	with ROIList do
   	if fromIndex <= Count - 1 then
       begin
       	oldCount := Count;
       	for i := fromIndex to oldCount - 1 do
           begin
           	oldROI := TSimpleROI(Items[i]);

               for chIndex := 0 to MAX_CH - 1 do
					if IsVideoChannel[chIndex] and (oldROI.Channel <> chIndex) then
                   begin
               		if oldROI is TRectangularROI then
               		begin
               			newROI := TRectangularROI.Create;
                   		(newROI as TRectangularROI).roiRect := (oldROI as TRectangularROI).roiRect;
               		end
                       else if oldROI is TEllipticalROI then
               		begin
               			newROI := TEllipticalROI.Create;
                   		(newROI as TEllipticalROI).roiRect := (oldROI as TEllipticalROI).roiRect;
                           (newROI as TEllipticalROI).FindPixels;
               		end
               		else
               		begin
               			newROI := TObjectROI.Create;
                   		if oldROI.PixelCount > 0 then
                   		for j := 0 to oldROI.PixelCount - 1 do
								(newROI as TObjectROI).AddPt(oldROI.Pixels[j]);
               		end;
               		newROI.fChannel := chIndex;
               		newROI.fROIIndex := ROIList.Count + 1;
               		Add(newROI);
                   end;
           end;
       end;
end;


const
   ALREADY_SCANNED = -32768;


procedure  TMPFile.DetectROIs(ROIList: TROIList; chIndex, fromFrame, toFrame, threshold, minArea,
		templateFrom, templateTo: integer);
var	tempFrameGroup: TFrameGroup;
	    i, j: integer;
	    objectROI: TObjectROI;
       avgCount: integer;
       avgArray: array of integer;

   {nested recursive procedure to find ROI in all adjacent points}
   procedure FindROIObject(i, j: integer; var aROIObject: TObjectROI);
   var	pixelValue: int16;
   begin
	    if (i < 0) or (i >= FrameWidth) or (j < 0) or (j >= FrameHeight) then Exit;
	    pixelValue := tempFrameGroup.channels[0].data[j * FrameWidth + i];
       if pixelValue >= threshold then
	    begin
		    tempFrameGroup.channels[0].data[j * FrameWidth + i] := ALREADY_SCANNED;
		    if aROIObject = nil then
           begin
               aROIObject := TObjectROI.Create;
               aROIObject.fChannel := chIndex;
               aROIObject.fROIIndex := ROIList.Count + 1;
           end;
           aROIObject.AddPt(Point(i, j));
	        FindROIObject(i - 1, j - 1, aROIObject);
	        FindROIObject(i    , j - 1, aROIObject);
	        FindROIObject(i + 1, j - 1, aROIObject);
	        FindROIObject(i - 1, j    , aROIObject);
	        FindROIObject(i + 1, j    , aROIObject);
	        FindROIObject(i - 1, j + 1, aROIObject);
	        FindROIObject(i    , j + 1, aROIObject);
	        FindROIObject(i + 1, j + 1, aROIObject);
	    end;
   end;

begin
	// create a temporary frame group
	tempFrameGroup := TFrameGroup.CreateTempFrameGroup(self);
   tempFrameGroup.FrameWidth := FrameWidth;
   tempFrameGroup.FrameHeight := FrameHeight;

	// average frames from templateFrom to templateTo onto tempFrameGroup

   avgCount := templateTo - templateFrom + 1;

   SetLength(avgArray, FrameWidth * FrameHeight);
   for i := 0 to FrameWidth * FrameHeight - 1 do avgArray[i] := 0;

   for j := templateFrom to templateTo do
       for i := 0 to FrameWidth * FrameHeight - 1 do
           avgArray[i] := avgArray[i] + Frames[j].channels[chIndex].data[i];

   for i := 0 to FrameWidth * FrameHeight - 1 do
       tempFrameGroup.channels[0].data[i] := avgArray[i] div avgCount;


	// search for objects in tempFrameGroup recursively
	for j := 0 to FrameHeight - 1 do
		for i := 0 to FrameWidth - 1 do
       begin
			objectROI := nil;
			FindROIObject(i, j, objectROI);
           {we found a new object}
			if objectROI <> nil then
               ROIList.Add(objectROI);
		end;

	// clean up; we eliminate all ObjectROIs that are not large enough

	if ROIList.Count > 0 then
		for i := ROIList.Count - 1 downto 0 do
			if ROIList.ROIs[i] is TObjectROI then
				if (ROIList.ROIs[i] as TObjectROI).PixelCount < minArea then
					ROIList.Delete(i);

	tempFrameGroup.Free;
end;


function    TMPFile.DigitalToAnalog(chIndex, iValue: integer): double;
begin
   Result := ChOffset[chIndex] + ChConvFactor[chIndex] * FullScaleToVal(ChInputRange[chIndex]) *
       (iValue / (MaxPixelValue + 1));
end;


procedure	TMPFile.DoBackgroundCorrection;
var	ch, i, j, pixelIndex: integer;
		pixelVal, minValue: int16;
begin
	if FrameCount = 0 then Exit;
   for ch := 0 to MAX_CH - 1 do
   	if VideoChEnabled[ch] then
       begin
       	minValue := 32767;
           for i := 0 to FrameWidth - 1 do
           	for j := 0 to FrameHeight - 1 do
               begin
               	pixelVal := Frames[ActiveFrameIndex].channels[ch].data[i + j * FrameWidth];
                   if pixelVal < minValue then minValue := pixelVal;
               end;
           if minValue < 0 then
           	for i := 0 to FrameWidth - 1 do
               	for j := 0 to FrameHeight - 1 do
                   begin
                   	pixelIndex := i + j * FrameWidth;
                   	pixelVal := Frames[ActiveFrameIndex].channels[ch].data[pixelIndex];
                       pixelVal := pixelVal - minValue;
                       Frames[ActiveFrameIndex].channels[ch].data[pixelIndex] := pixelVal;
                   end;
       end;
end;


procedure	TMPFile.FinishMotionCorrection;
begin
	{terminates FFT operations}
	ippiFFTFree_R_32f(pSpec);
	disableMotionCorrection := False;
end;


function	TMPFile.GetAverage(frameIndex, chIndex: integer; rc: TRect): integer;
var	i, j: integer;
begin
	Result := 0;
   for i := rc.Left to rc.Right do
   	for j := rc.Top to rc.Bottom do
       	Result := Result + Frames[frameIndex].channels[chIndex].data[i + j * FrameWidth];
   Result := Result div ((rc.Right - rc.Left + 1) * (rc.Bottom - rc.Top + 1));
end;


function	TMPFile.GetMax(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
var	pixelVal, i, j: integer;
begin
    Result := -32768;
    for i := rc.Left to rc.Right do
        for j := rc.Top to rc.Bottom do
        begin
            pixelVal := Frames[frameIndex].channels[chIndex].data[i + j * FrameWidth];
            if pixelVal > Result then
            begin
                Result := pixelVal;
                x := i;
                y := j;
            end;
        end;
end;


function	TMPFile.GetMin(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
var	pixelVal, i, j: integer;
begin
	Result := 32768;
   for i := rc.Left to rc.Right do
   	for j := rc.Top to rc.Bottom do
       begin
       	pixelVal := Frames[frameIndex].channels[chIndex].data[i + j * FrameWidth];
           if pixelVal < Result then
           begin
           	Result := pixelVal;
           	x := i;
               y := j;
           end;
       end;
end;


function   TMPFile.GetROIAverageValue(ROIList: TROIList; roiIndex, frameIndex: integer): integer;
var	i, j: integer;
	theROI: TSimpleROI;
	roiPixel: TPoint;
begin
	Result := 0;
	theROI := ROIList.ROIs[roiIndex];
	for i := 0 to theROI.PixelCount - 1 do
	begin
		roiPixel := theROI.Pixels[i];
		j := roiPixel.x + roiPixel.y * FrameWidth;
		Result := Result + Frames[frameIndex].channels[theROI.Channel].data[j];
	end;
	Result := Result div theROI.PixelCount;
end;


function    TMPFile.GetPixelValue(frameIndex, chIndex, xData, yData: integer): int16;
begin
   Result := 0;
   if (frameIndex >= 0) and (frameIndex < FrameCount) and IsVideoChannel[chIndex]
   and (xData >= 0) and (xData < FrameWidth) and (yData >= 0)
       and (yData < FrameHeight) then
           Result := Frames[frameIndex].channels[chIndex].data[xData + yData * FrameWidth];
end;


function	TMPFile.GetPropertyValue(propName: string): string;
var    ps: TpPropSpecArray;
       pv: TpPropVariantArray;
begin
   ps := nil; pv := nil;
   try
       GetMem(ps, SizeOf(TPropSpec));
       GetMem(pv, SizeOf(TPropVariant));
       ps^[0].ulKind := PRSPEC_LPWSTR;
       ps^[0].lpwstr := StringToOLEStr(propName);
       try
       	OleCheck(propertyStorage.ReadMultiple(1, @ps[0], @pv[0]));
			Result := WideCharToString(pv^[0].pwszVal);
       except
           Result := 'No match';
       end;
   finally
       if ps <> nil then Freemem(ps);
       if pv <> nil then Freemem(pv);
   end;
end;


function    TMPFile.IsOperationOK(chIndex, fromFrame, toFrame: integer; dest: TMPFile): boolean;
begin
   Result := ChEnabled[chIndex] and (fromFrame >= 0) and (fromFrame < FrameCount) and (toFrame >= 0) and
       (toFrame < FrameCount) and (fromFrame <= toFrame) and dest.IsMemoryFile;
end;


type   EAVIError = class(Exception);
       TRGBTripleArray = array[0..Maxint div 8] of TRGBTriple;
       TpRGBTripleArray = ^TRGBTripleArray;


procedure   TMPFile.MakeAVIMovie(const avifilename: string; chIndex, fromFrame, toFrame, newframeRate: integer;
	colorScheme: TColorScheme; fromViewer: TObject; bOverlayCh1on2: boolean);
var    aviFile: IAVIFile;
       aviStream: IAVIStream;
       bitmapInfo: TBITMAPINFO;
       pAviBitmap: TpRGBTripleArray;
       spCharFileName: array[0..255] of WideChar;
       spCharStreamName: array[0..63] of WideChar;
       streamInfo: TAVIStreamInfo;
       i, j, k, maxPixVal: integer;
       pixel0, pixel1: TRGBTriple;
       hr: HResult;
const  sStreamName = '2-photon laser confocal microscope video stream';
begin
	maxPixVal := ChMaxPixelValues[chIndex]; {caching value}

   if AVIFileOpen(aviFile,
       StringToWideChar(aviFilename, @spCharFileName, 255),
       OF_CREATE or OF_SHARE_EXCLUSIVE or OF_WRITE, nil) <> 0 then
           raise EAVIError.Create('Error in creating AVI file');
   FillChar(streamInfo, SizeOf(streamInfo), 0);
   StringToWideChar(sStreamName, @spCharStreamName, 63);
   with streamInfo do
   begin
       fccType := streamtypeVIDEO;
       fccHandler := 0;
       dwScale := 1;
       dwRate := Round(newframeRate);
       if dwRate > 24 then dwRate := 24; // bound to 24 fps
       dwSuggestedBufferSize := FrameWidth * FrameHeight;
       rcFrame := Rect(0, 0, FrameWidth - 1, FrameHeight - 1);
       for i := 0 to 63 do
       szName[i] := spCharStreamName[i];
   end;

   if AVIFileCreateStream(aviFile, aviStream, streamInfo) <> 0 then
       raise EAVIError.Create('Cannot create AVI stream');

   FillChar(bitmapInfo, SizeOf(TBITMAPINFO), 0);
   with bitmapInfo.bmiHeader do
   begin
       biSize := SizeOf(TBITMAPINFOHEADER);
       biWidth := FrameWidth;
       biHeight := FrameHeight; // we invert bitmap in the loop
       biPlanes := 1;
       biBitCount := 24;
       biCompression := BI_RGB;
   end;

   if AVIStreamSetFormat(aviStream, 0, @bitmapInfo, SizeOf(TBITMAPINFOHEADER)) <> 0 then
       raise EAVIError.Create('Cannot set format of video stream');

   GetMem(pAviBitmap, FrameHeight * FrameWidth * SizeOf(TRGBTriple));

   i := fromFrame;

   repeat
       for j := 0 to FrameHeight * FrameWidth - 1 do
       begin
           k := Frames[i].Channels[chIndex].data[j];
           if k < 0 then k := 0;
           k := Muldiv(k, maxPixVal, MAX_FALSE_COLORS -  1);
           if k > MAX_FALSE_COLORS -  1 then k := MAX_FALSE_COLORS - 1;

           case colorScheme of
           	CS_GRAYSCALE:	pAviBitmap^[j] := Mainform.grayScaleTable[k];
               CS_FALSECOLORS: pAviBitmap^[j] := Mainform.falseColorTable[k];
               CS_CUSTOMLUT:	pAviBitmap^[j] := (fromViewer as TViewerFrm).mpFile.CustomColors[chIndex][k];
           end;

           {Overlay Ch1 on Ch2}
           if bOverlayCh1on2 and (chIndex = 1) and VideoChEnabled[0] then
           begin
           	pixel1 := pAviBitmap^[j];

               {pixel for Ch1 (index of 0)}
           	k := Frames[i].Channels[0].data[j];
           	if k < 0 then k := 0;
           	k := Muldiv(k, maxPixVal, MAX_FALSE_COLORS -  1);
           	if k > MAX_FALSE_COLORS -  1 then k := MAX_FALSE_COLORS - 1;
           	case colorScheme of
           		CS_GRAYSCALE:	pixel0 := Mainform.grayScaleTable[k];
               	CS_FALSECOLORS: pixel0 := Mainform.falseColorTable[k];
               	CS_CUSTOMLUT:	pixel0 := (fromViewer as TViewerFrm).mpFile.CustomColors[0][k];
           	end;

               pixel1.rgbtBlue := pixel1.rgbtBlue or pixel0.rgbtBlue;
               pixel1.rgbtGreen := pixel1.rgbtGreen or pixel0.rgbtGreen;
               pixel1.rgbtRed := pixel1.rgbtRed or pixel0.rgbtRed;
               pAviBitmap^[j] := pixel1;
           end;
       end;

       hr := AVIStreamWrite(aviStream, i - fromFrame, 1, pAviBitmap,
           FrameWidth * FrameHeight * SizeOf(TRGBTriple), AVIIF_KEYFRAME, nil, nil);
       i := i + 1;
   until not Succeeded(hr) or (i > toFrame);

   Freemem(pAviBitmap, FrameHeight * FrameWidth * SizeOf(TRGBTriple) );

   {instead of using the AVI API functions to release the stream and file,
   we do it the Delphi way by making them nil}
   aviStream := nil;
   aviFile := nil;

   if not Succeeded(hr) then
       raise EAVIError.Create('Error writing frame to file');
end;


function  TMPFile.MakeTIFF(const avifilename: string; chIndex, fromFrame, toFrame: integer): boolean;
var	Stream     : TFileStream;
       i, j, k, l, sLength, maxPixVal: integer;
       pixelValue: int16;
       wPixelValues: array of Word;
       videoFrame: TFrame;
       ifd: TIFD;
       sImageDescription: string;
begin
	Result := True;
   maxPixVal := ChMaxPixelValues[chIndex]; {just caching this value}
	Stream := TFileStream.Create(avifilename, fmCreate Or fmShareExclusive);
   try
   	sImageDescription := Format(sTIFFPropNames, [
   		chIndex + 1,
   		Resolution,
       	ScanModeToString(ScanMode),
       	XStagePosition,
       	YStagePosition,
       	ZStagePosition,
       	StackFrameCount,
       	StackInterval,
       	StackAveragingCount,
       	StackRepeatCount,
       	Magnification,
       	Rotation,
       	XFrameOffset,
       	YFrameOffset,
       	FrameRate ]);


   	{Write TIFF file header}
       Stream.Write(tiffHeader, SizeOf(tiffHeader));

       {Write the IFDs}
       MakeIFD(ifd);
       sLength := Length(sImageDescription) + 1; {includes the NULL finishing the string}

       with ifd do
       begin
       	tagRecords[22].Count := sLength;
       	tagRecords[2].dataOffset := FrameWidth;
       	tagRecords[3].dataOffset := FrameHeight;
       	tagRecords[12].dataOffset := FrameHeight; {RowsPerStrip}
       	tagRecords[13].dataOffset := FrameHeight * FrameWidth * SizeOf(Word); {StripByteCounts}
       end;

   	for i := fromFrame to toFrame do
       	with ifd do
       	begin
            	{XResolution}
       		ifd.tagRecords[16].dataOffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
               	- 4 * SizeOf(integer) - sLength;
               {YResolution}
       		ifd.tagRecords[17].dataOffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
               	- 2 * SizeOf(integer) - sLength;

               {ImageDescription}
				ifd.tagRecords[22].dataOffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
               	- sLength;

               {StripOffsets}
       		ifd.tagRecords[9].dataOffset := SizeOf(tiffHeader) + (toFrame - fromFrame + 1) * (SizeOf(ifd) + sLength) +
               	FrameWidth * FrameHeight * SizeOf(Word) * (i - fromFrame);

		       	if i <> toFrame then
               	nextIDFoffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
       	    else
           		nextIDFoffset := 0;

           	Stream.Write(ifd, SizeOf(ifd));

               {Write the description string}
               Stream.Write(sImageDescription[1], sLength);
           end;

		{Write frame data for each frame}
		SetLength(wPixelValues, FrameHeight * FrameWidth);
   	for i := fromFrame to toFrame do
       begin
           videoFrame := Frames[i].Channels[chIndex];
       	for j := 0 to FrameHeight - 1 do
           begin
               for k := 0 to FrameWidth - 1 do
               begin
               	l := j * FrameWidth + k;
               	pixelValue := videoFrame.data[l];

                   {truncate value and normalize to 0..65535}
                   if pixelValue < 0 then
                   begin
                   	wPixelValues[l] := 0;
                       Result := False;
                   end
                   else if pixelValue > maxPixVal then
                   begin
						wPixelValues[l] := 65535;
					end
                   else
                   	wPixelValues[l] := Muldiv(pixelValue, 65535, maxPixVal);
               end;
           end;
			Stream.Write(wPixelValues[0], FrameHeight * FrameWidth * SizeOf(Word));
        end;
   finally
   	Stream.Free;
   end;
end;


procedure   TMPFile.NewAnalogWnd;
var    analogFrm: TAnalogFrm;
begin
   analogFrm := TAnalogFrm.Create(Mainform);
   analogFrm.Initialize(self);
   analogFrm.Show;
   analogWndList.Add(analogFrm);
   analogFrm.FormResize(nil); {forces painting}
end;


procedure   TMPFile.NewViewer;
var    viewerFrm: TViewerFrm;
begin
   viewerFrm := TViewerFrm.Create(Mainform);
   viewerFrm.Initialize(self, viewerList.Add(viewerFrm) + 1);
   viewerFrm.CurrentFrameIndex := 0;
   viewerFrm.Show;
end;


procedure   TMPFile.OnNewFrames;
var    i: integer;
begin
   if FrameCount <= 0 then Exit;
   if viewerList.Count > 0 then
       for i := 0 to viewerList.Count - 1 do
           TViewerFrm(viewerList.Items[i]).CurrentFrameIndex := FrameCount - 1;
end;


procedure   TMPFile.OnWndClose(wnd: TForm); {when a viewer or analog wnd closes}
var    i: integer;
begin
   if not Closing and not Mainform.bAppClosing then
       if AnalogWndCount + ViewerCount = 1 then
           {the last window standing for this file: if we close, file object is free}
           Free
       else
       begin
           if analogWndList <> nil then
           begin
               i := analogWndList.IndexOf(wnd);
               if i >= 0 then analogWndList.Remove(wnd);
           end;
           i := viewerList.IndexOf(wnd);
           if i >= 0 then viewerList.Remove(wnd);
       end;
end;


function PowerOfTwo(value: integer): integer;
begin
	Result := 1;
	if value > 0 then
   	while Result < value do
       	Result := Result * 2;
end;


function ExponentofTwo(value: integer): integer;
var	poweroftwoval: integer;
begin
	Result := 0; poweroftwoval := 1;
   while poweroftwoval < value do
   begin
   	Result := Result + 1;
       poweroftwoval := poweroftwoval * 2;
   end;
end;


procedure	TMPFile.PerformMotionCorrection(frameIndex: integer);
var	i, j, k,l: integer;
		maxValue, corrValue: single;
       maxValuePt: TPoint;
       ippStatusCode: ippStatus;
begin
   {loads frame, convert to singles in frameFFTarray}
	LoadFrameGroup(frameIndex); {this will load without any motion correction}
	fActiveFrameIndex := frameIndex;

   {cleans all values from frameFFTarray from prev call to PerformMotionCorrection}
	FillChar(frameFFTarray[0], poweroftwo_width * poweroftwo_height * SizeOf(single), 0);
   with fCurrentFrameGroup.channels[refMotionChannel] do
   	for j := 0 to refMotionRect.Bottom - refMotionRect.Top do
       	for i := 0 to refMotionRect.Right - refMotionRect.Left do
           	frameFFTarray[j * poweroftwo_width + i]
               	:= data[(j + refMotionRect.Top) * FrameWidth + refMotionRect.Left + i];

   {Apply Hamming window to frameFFTarray}
	ippStatuscode := ippiWinHamming_32f_C1IR(frameFFTarray[0], poweroftwo_width * SizeOf(single), FFTroiSize);
	Assert(ippStatuscode = 0, 'ippiWinHammingSep_32f_C1IR: ' + IntToStr(ippStatuscode));

   {FFT of frameFFTarray}
	ippStatuscode := ippiFFTFwd_RToPack_32f_C1IR(frameFFTarray[0], poweroftwo_width * SizeOf(single), pSpec, @FFTBuffer[0]);
	Assert(ippStatuscode = 0, 'ippiFFTFwd_RToPack_32f_C1IR: ' + IntToStr(ippStatuscode));

   {multiply refFFTarray with conjugate of frameFFTarray, result in frameFFTarray}
	ippStatuscode := ippiMulPackConj_32f_C1IR(refFFTarray[0], poweroftwo_width * SizeOf(single),
   						frameFFTarray[0], poweroftwo_width * SizeOf(single), FFTroiSize);
	Assert(ippStatuscode = 0, 'ippiMulPackConj_32f_C1IR: ' + IntToStr(ippStatuscode));

   {calculate magnitude of frameFFTarray}
	ippStatuscode := ippiMagnitudePack_32f_C1R(frameFFTarray[0], poweroftwo_width * SizeOf(single),
   	magnitudeArray[0], poweroftwo_width * SizeOf(single), FFTroiSize);
	Assert(ippStatuscode = 0, 'ippiMagnitudePack_32f_C1IR: ' + IntToStr(ippStatuscode));

   {Divide all elements of frameFFTarray by its magnitude: takes packing into account}
   for i := 0 to (poweroftwo_width div - 1) div 2 do
   	for j := 0 to poweroftwo_height - 1 do
       begin
           l := j * poweroftwo_width + i;
           k := l + i + 1;
           if magnitudeArray[l] <> 0 then
           begin
           	frameFFTarray[k] := frameFFTarray[k] / magnitudeArray[l];
           	frameFFTarray[k+1] := frameFFTarray[k+1] / magnitudeArray[l];
           end;
       end;

	// values on side columns
   for i := 0 to poweroftwo_height div 2 do
 	// special cases: values in corner
		if i = 0 then
       begin
       	if magnitudeArray[0] <> 0 then frameFFTarray[0] := frameFFTarray[0] / magnitudeArray[0];
       	if magnitudeArray[poweroftwo_width div 2] <> 0 then frameFFTarray[poweroftwo_width - 1] := frameFFTarray[0] / magnitudeArray[poweroftwo_width div 2];
       end
       else if i = poweroftwo_height div 2 then
       begin
       	if magnitudeArray[i] <> 0 then frameFFTarray[(poweroftwo_height - 1)*poweroftwo_width] := frameFFTarray[0] / magnitudeArray[i];
			l := poweroftwo_width div 2 + (poweroftwo_height div 2) * poweroftwo_width;
           k := poweroftwo_width * poweroftwo_height - 1;
       	if magnitudeArray[l] <> 0 then frameFFTarray[k] := frameFFTarray[k] / magnitudeArray[l];
       end
       else
       begin
       	j := i * poweroftwo_width;
           if magnitudeArray[j] <> 0 then
           begin
           	k := (2 * (i - 1) + 1) * poweroftwo_width; l := k + poweroftwo_width;
				frameFFTarray[k] := frameFFTarray[k] / magnitudeArray[j];
				frameFFTarray[l] := frameFFTarray[l] / magnitudeArray[j];
			end;
           j := j + poweroftwo_width div 2;
           if magnitudeArray[j] <> 0 then
           begin
           	k := (2 * (i - 1) + 1) * poweroftwo_width + poweroftwo_width - 1;
               l := k + poweroftwo_width;
				frameFFTarray[k] := frameFFTarray[k] / magnitudeArray[j];
				frameFFTarray[l] := frameFFTarray[l] / magnitudeArray[j];
			end;
       end;


	{Fourier transform back}
	ippStatuscode := ippiFFTInv_PackToR_32f_C1IR(frameFFTarray[0], poweroftwo_width * SizeOf(single), pSpec, @FFTBuffer[0]);
	Assert(ippStatuscode = 0, 'ippiFFTInv_PackToR_32f_C1IR: ' + IntToStr(ippStatuscode));

   {finds maximum in frame}
   maxValue := 0;
	motionDeltas[frameIndex] := Point(0,0);

   for j := 0 to refMotionRect.Bottom - refMotionRect.Top do
   	for i := 0 to refMotionRect.Right - refMotionRect.Top do
       	try
           	corrValue := frameFFTarray[j * poweroftwo_width + i];
       		if corrValue > maxValue then
           	begin
   				maxValue := corrValue;
                   maxValuePt := Point(i,j);
           	end;
           except
           end;

   if maxValuePt.x > poweroftwo_width div 2 then maxValuePt.x := maxValuePt.x - poweroftwo_width;
   if maxValuePt.y > poweroftwo_height div 2 then maxValuePt.y := maxValuePt.y - poweroftwo_height;

   motionDeltas[frameIndex] := Point(-maxValuePt.x, -maxValuePt.y);
end;


procedure	TMPFile.PrepareMotionCorrection(refFrame, refCh: integer; refRect: TRect);
var	motionFFTBufferSize, i, j: integer;
		ippStatuscode: ippStatus;
begin
//
//			PHASE CORRELATION
//
	disableMotionCorrection := True; {prevents loading frames and adjusting for motion}
	MotionCorrectionEnabled := True; {sizes buffers etc...}
	refMotionChannel := refCh;
   refMotionRect := refRect;
   {determines power of two sizes: poweroftwo_width, poweroftwo_height}
   poweroftwo_width := PowerOfTwo(refRect.Right - refRect.Left + 1);
   poweroftwo_height := PowerOfTwo(refRect.Bottom - refRect.Top + 1);

   {initializes FFT operation}
   FFTroiSize.width := poweroftwo_width;
   FFTroiSize.height := poweroftwo_height;

	ippStatuscode := ippiFFTInitAlloc_R_32f(pSpec, ExponentOfTwo(refMotionRect.Right - refMotionRect.Left + 1),
       ExponentOfTwo(refMotionRect.Bottom - refMotionRect.Top + 1), IPP_FFT_NODIV_BY_ANY, ippAlgHintFast);
	Assert(ippStatuscode = 0, 'ippiFFTInitAlloc_R_32f: ' + IntToStr(ippStatuscode));

	{finds size of buffers}
	ippStatuscode := ippiFFTGetBufSize_R_32f(pSpec, motionFFTBufferSize);
	Assert(ippStatuscode = 0, 'ippiFFTGetBufSize_R_32f: ' + IntToStr(ippStatuscode));

   SetLength(FFTBuffer, motionFFTBufferSize);

   {creates single arrays, refFFTarray, frameFFTarray, magnitudeArray, fills with zeros}
   SetLength(refFFTarray, poweroftwo_width * poweroftwo_height);
   SetLength(frameFFTarray, poweroftwo_width * poweroftwo_height);
   SetLength(magnitudeArray, poweroftwo_width * poweroftwo_height);
	FillChar(refFFTarray[0], poweroftwo_width * poweroftwo_height * SizeOf(single), 0);

   {loads reference frame, convert to singles in refFFTarray}
	LoadFrameGroup(refFrame); {this will load without any motion correction}
	fActiveFrameIndex := refFrame;

   with fCurrentFrameGroup.channels[refMotionChannel] do
   	for j := 0 to refMotionRect.Bottom - refMotionRect.Top do
       	for i := 0 to refMotionRect.Right - refMotionRect.Left do
           	refFFTarray[j * poweroftwo_width + i]
               	:= data[(j + refMotionRect.Top) * FrameWidth + refMotionRect.Left + i];

   {Apply Hamming window to lines and columns of refFFTarray}
	ippStatuscode := ippiWinHamming_32f_C1IR(refFFTarray[0], poweroftwo_width * SizeOf(single), FFTroiSize);
	Assert(ippStatuscode = 0, 'ippiWinHammingSep_32f_C1IR: ' + IntToStr(ippStatuscode));

   {FFT of refFFTarray}
	ippStatuscode := ippiFFTFwd_RToPack_32f_C1IR(refFFTarray[0], poweroftwo_width * SizeOf(single), pSpec, @FFTBuffer[0]);
	Assert(ippStatuscode = 0, 'ippiFFTFwd_RToPack_32f_C1IR: ' + IntToStr(ippStatuscode));
end;


procedure	TMPFile.SaveChannelsToFile(bASCII: boolean; fname: string; bCh2, bCh3: boolean; fromFrame, toFrame: integer);
var    chIndex, frameIndex, sampleIndex: integer;
		timeVal, sampleVal: double;
		fileStream: TFileStream;
       swLength: integer;
       s1: string;
       sw: array[0..179] of WideChar;
begin
	if not(bCh2 or bCh3) then Exit;
   if bCh2 then chIndex := 2 else chIndex := 3;

   fileStream := TFileStream.Create(fname, fmCreate or fmShareExclusive);

   try
   	fileStream.Seek(0, soFromBeginning);
   {fills string with data; 11 digits for each column - each digit is a wide char, 2 columns}
   {number of rows is (toFrame - fromFrame + 1) * ChDataPtsPerFrame[]}
   {frameIndex: 0 to toFrame - fromFrame}
   {time in ms for each frame: (fromFrame + frameIndex) * FrameSize * PixelClock * 1e-6 }
   {sampleIndex: 0 to ChDataPtsPerFrame[] - 1}
   {for each sample: sampleIndex * FrameSize * PixelClock * 1e-6 / ChDataPtsPerFrame[]}

       for frameIndex := 0 to toFrame - fromFrame do
       begin
       	ActiveFrameIndex := fromFrame + frameIndex;
           for sampleIndex := 0 to ChDataPtsPerFrame[chIndex]- 1 do
           begin
           	timeVal := (sampleIndex / ChDataPtsPerFrame[chIndex] + fromFrame + frameIndex) / FrameRate;
               {timeVal in s; pixel clock is in increments of 50 ns = 5e-5 ms(20 MHz)}
               {	timeVal := (FrameSize * PixelClock * 5e-5) *
               (sampleIndex / ChDataPtsPerFrame[chIndex] + fromFrame + frameIndex);}
               {sampleVal}
               sampleVal := ChConvFactor[chIndex] * FullScaleToVal(ChInputRange[chIndex]) *
               	Frames[ActiveFrameIndex].channels[chIndex].data[sampleIndex] /
                   (f_MAX_Y_VALUE + 1) + ChOffset[chIndex];

               s1 := Format('%.6f', [timeVal]) + TAB + Format('%g', [sampleVal]) + CRLF;

               if bASCII then
               	fileStream.Write(s1[1], Length(s1))
               else
               begin
               	swLength := 2 * Length(s1); {Unicode business}
               	StringToWideChar(s1, @sw, swLength + 1);
               	fileStream.Write(sw, swLength);
               end;
           end;
       end;

   finally
   	fileStream.Free;
   end;
end;


function	TMPFile.SaveMotionCorrection: boolean;
var	motionStream: IStream;
		cbWritten: longint;
       libNewPosition: Largeint;
begin
	Result := False;
	if not bReadOnly then
   	if Length(motionDeltas) = FrameCount then
		if Succeeded(RootStorage.CreateStream(sMotionCorrection, OFLAGS, 0, 0, motionStream)) then
       with motionStream do
       	if Succeeded(Seek(0, STREAM_SEEK_SET, libNewPosition)) then
           	if Succeeded(Write(@motionDeltas[0], Length(motionDeltas) * SizeOf(TPoint), @cbWritten)) then
               	Result := True;
end;


procedure   TMPFile.StackX(chIndex, fromFrame, toFrame, fromY, toY: integer; dest: TMPFile);{creates new frame}
var    newFrameGroup: TFrameGroup;
       i, j, k: integer;
       value: int16;
       pDataElt: ^int16;
begin
   newFrameGroup := TFrameGroup.CreateForOp(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := toFrame - fromFrame + 1;

   dest.AddFrameGroup(newFrameGroup);
   for i := fromFrame to toFrame do
   begin
       ActiveFrameIndex := i;
       for j := 0 to FrameWidth - 1 do
       begin
           pDataElt := @(newFrameGroup.Channels[0].data[(i - fromFrame) * FrameWidth + j]);
           pDataElt^ := 0;
           for k := fromY to toY do
           begin
               value := Frames[ActiveFrameIndex].channels[chIndex].data[k * FrameWidth + j];
               if value > DEFAULT_MAX_PIXEL_VALUE then value := DEFAULT_MAX_PIXEL_VALUE;
               if value > pDataElt^ then pDataElt^ := value;
           end;
       end;
       dest.OnNewFrames; {update each time we project - looks nicer!}
   end;
end;


function	TMPFile.SizeOfFrameCompatible(width, height: integer): boolean;
begin
	if FrameCount = 0 then
   	Result := True
   else
   	if (width = FrameWidth) and (height = FrameHeight) then
       	Result := True
       else
       begin
       	MessageDlg( 'Operation not possible.' + CRLF +
       				'Destination workspace has different frame size.',
                       mtError, [mbOK], 0);
       	Result := False;
       end;
end;


procedure   TMPFile.StackY(chIndex, fromFrame, toFrame, fromX, toX: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       i, j, k: integer;
       value: int16;
       pDataElt: ^int16;
begin
   newFrameGroup := TFrameGroup.CreateForOp(dest);
   newFrameGroup.FrameWidth := FrameHeight;
   newFrameGroup.FrameHeight := toFrame - fromFrame + 1;

   dest.AddFrameGroup(newFrameGroup);
   for i := fromFrame to toFrame do
   begin
       ActiveFrameIndex := i;
       for j := 0 to FrameHeight - 1 do
       begin
           pDataElt := @newFrameGroup.Channels[0].data[(i - fromFrame) * FrameHeight + j];
           pDataElt^ := 0;
           for k := fromX to toX do
           begin
               value := Frames[ActiveFrameIndex].channels[chIndex].data[j * FrameWidth + k];
               if value > DEFAULT_MAX_PIXEL_VALUE then value := DEFAULT_MAX_PIXEL_VALUE;
               if value > pDataElt^ then pDataElt^ := value;
           end;
       end;
       dest.OnNewFrames; {update each time we project - looks nicer!}
   end;
end;


procedure   TMPFile.StackZ(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       i, j, k: integer;
       value: int16;
       pDataElt: ^int16;
begin
   newFrameGroup := TFrameGroup.CreateForOp(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := FrameHeight;

   for i := 0 to FrameWidth * FrameHeight - 1 do
       newFrameGroup.Channels[0].data[i] := 0;

   dest.AddFrameGroup(newFrameGroup);
   for i := fromFrame to toFrame do
   begin
       ActiveFrameIndex := i;
       for j := 0 to FrameHeight - 1 do
           for k := 0 to FrameWidth - 1 do
           begin
               pDataElt := @newFrameGroup.Channels[0].data[j * FrameWidth + k];
               value := Frames[ActiveFrameIndex].channels[chIndex].data[j * FrameWidth + k];
               if value > DEFAULT_MAX_PIXEL_VALUE then value := DEFAULT_MAX_PIXEL_VALUE;
               if value > pDataElt^ then pDataElt^ := value;
           end;
       dest.OnNewFrames; {update each time we project - looks nicer!}
   end;
end;


function   TMPFile.SaveAs(const sFilename: string): TFileErr;
var    hr: HResult;
       i: integer;
begin
   // create the file on disk
   hr := StgCreateDocFile(StringToOLEStr(sFilename), OFLAGS, 0, rootStorage);
   if Succeeded(hr) then
   begin
       propertySetStorage := rootStorage as IPropertySetStorage;
       propertySetStorage.Create(FMTID_User_Defined_Properties,
           FMTID_User_Defined_Properties, PROPSETFLAG_DEFAULT, OFLAGS, propertyStorage);
       rootStorage.SetClass(GUID_MPD);

       // save user information
       WritePropertySet;
       frameCommentList.SaveFrameComments;

       // change status of file and filename
       fFilename := sFilename;
       fbIsMemoryFile := False;

       fFrameCount := frameGroupList.Count;

       // opens the channels
       OLECheck(rootStorage.CreateStream('Ch0', OFLAGS, 0, 0, chStreams[0]));

       // save the frame list
       for i := 0 to frameGroupList.Count - 1 do
           (TFrameGroup(frameGroupList.Items[i]).channels[0] as TVideoFrame).SaveToFile;

       // create the current frame group (at frame Index 0)
       fActiveFrameIndex := 0;
       fCurrentFrameGroup := TFrameGroup(frameGroupList.Items[0]);

       // destroys the frame list (except first frame at index 0)
       if frameGroupList.Count> 1 then
           for i := 1 to frameGroupList.Count - 1 do
               TFrameGroup(frameGroupList.Items[i]).Free;
       frameGroupList.Free;
       frameGroupList := nil;

       Result := feOK;

       // notify all the viewers as the name change
       if viewerList.Count > 0 then
           for i := 0 to viewerList.Count - 1 do
               TForm(viewerList.Items[i]).Caption := 'Video Channels - ' +
               ExtractFileName(fFilename);
   end
   else
       Result := StgErrToFileErr(hr);
end;


procedure   TMPFile.SubtractFrame(chIndex, plusFrameIndex, minusFrameIndex: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       i: integer;
       value, minValue: int16;
begin
   newFrameGroup := TFrameGroup.CreateForOp(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := FrameHeight;

   for i := 0 to FrameWidth * FrameHeight - 1 do
       newFrameGroup.channels[0].data[i] := Frames[plusFrameIndex].channels[chIndex].data[i];

   minValue := DEFAULT_MAX_PIXEL_VALUE;
   for i := 0 to FrameWidth * FrameHeight - 1 do
   begin
       value := newFrameGroup.channels[0].data[i] - Frames[minusFrameIndex].channels[chIndex].data[i];
       if value < minValue then minValue := value;
       if value > DEFAULT_MAX_PIXEL_VALUE then value := DEFAULT_MAX_PIXEL_VALUE;
       newFrameGroup.channels[0].data[i] := value;
   end;

   {bring minimal pixel level to 0}
   if minValue < 0 then
   for i := 0 to FrameWidth * FrameHeight - 1 do
   begin
       value := newFrameGroup.channels[0].data[i] - minValue;
       if value > DEFAULT_MAX_PIXEL_VALUE then value := DEFAULT_MAX_PIXEL_VALUE;
       newFrameGroup.channels[0].data[i] := value;
   end;

   dest.AddFrameGroup(newFrameGroup);
end;


constructor TMPFile.CreateFromTemplate(const sFilename: string; templateFile: TMPFile);
var    i: integer;
begin
   {a worskpace file has always one video channel and no analog channel}
   fResolution := templateFile.fResolution;
   fActiveFrameIndex := -1;
   fAngle := templateFile.Angle;
   fbIsMemoryFile := True;
   fChEnabled[0] := True;
   fChEnabled[1] := False;
   fChEnabled[2] := False;
   fChEnabled[3] := False;
   for i := 0 to MAX_CH - 1 do
   begin
       fChConvFactors[i] := templateFile.ChConvFactor[i];
       fChDataPtsPerFrames[i] := templateFile.ChDataPtsPerFrame[i];
       fChInputRanges[i] := templateFile.ChInputRange[i];
       fChNames[i] := templateFile.ChNames[i];
       fChPrefixes[i] := templateFile.ChPrefix[i];
       fChUnits[i] := templateFile.ChUnit[i];
       fChOffsets[i] := templateFile.ChOffset[i];
       fIsVideoCh[i] := False;
       fChMaxPixelValues[i] := templateFile.ChMaxPixelValues[i];
   end;
   fFilename := sFilename;
   fFrameCount := 0;
   fFrameHeight := templateFile.FrameHeight;
   fFrameWidth := templateFile.FrameWidth;
   fLineRepeatCount := templateFile.LineRepeatCount;
   fMagnification := templateFile.Magnification;
   fPixelClock := templateFile.PixelClock;
   fRotation := templateFile.Rotation;
   fScanMode := SM_MOVIE; {a workspace is always a movie}
   fStackAveragingCount := templateFile.LineRepeatCount;
   fStackFrameCount := templateFile.StackFrameCount;
   fStackInterval := templateFile.StackInterval;
   fStackRepeatCount := templateFile.StackRepeatCount;
   fXFrameOffset := templateFile.XFrameOffset;
   fXStagePosition := templateFile.XStagePosition;
   fYFrameOffset := templateFile.YFrameOffset;
   fYStagePosition := templateFile.YStagePosition;
   fZStagePosition := templateFile.ZStagePosition;

   {color information here}
   baseColors := templateFile.baseColors;
   negativeColors := templateFile.negativeColors;
   midRangeColors := templateFile.midRangeColors;
   maxColors := templateFile.maxColors;
	MaxPixels := templateFile.MaxPixels;

   frameCommentList := TFrameCommentList.Create(self);
   frameGroupList := TList.Create;
   bDirty := False;
   viewerList := TList.Create;
   NewViewer;
end;


constructor TMPFile.CreateFromFile(const sFilename: string);
var    i: integer;
       hr: HResult;
begin
   {opens the storage}
	InitializeColors;
   hr := StgOpenStorage(StringToOLEStr(sFilename), nil, RWFLAGS, nil, 0, rootStorage);
   if Failed(hr) then
   begin
   	hr := StgOpenStorage(StringToOLEStr(sFilename), nil, RFLAGS, nil, 0, rootStorage);
       if Succeeded(hr) then bReadOnly := True;
	end;

   if Succeeded(hr) then
   begin
       propertySetStorage := rootStorage as IPropertySetStorage;
       if not Succeeded(propertySetStorage.Open(FMTID_User_Defined_Properties,
        RFLAGS, propertyStorage)) then MessageDlg('Cannot open property set', mtError, [mbOK], 0);

       ReadPropertySet;
       for i := 0 to MAX_CH - 1 do
           if chEnabled[i] then
               rootStorage.OpenStream(StringToOleStr('Ch' + IntToStr(i)), nil, RFLAGS,
                   0, chStreams[i]);

       AdjustFrameCount;
       fFilename := sFilename;

       if ScanMode = SM_REGIONSCAN then GetRegions;

       LoadMotionCorrection;

       fCurrentFrameGroup := TFrameGroup.Create(self);
       fActiveFrameIndex := -1; ActiveFrameIndex := 0; {forces loading}

       frameCommentList := TFrameCommentList.Create(self);
       frameCommentList.LoadFrameComments;

       viewerList := TList.Create;
       NewViewer;

       if (AnalogChCount > 0) and (ScanMode <> SM_STACK) and (ScanMode <> SM_FASTSTACK) then
			{just to be on the safe side, if the analog streams are empty, we bail}
       	if (chStreams[2] <> nil) or (chStreams[3] <> nil) then
       	begin
           	analogWndList := TList.Create;
           	NewAnalogWnd;
       	end;
   end
   else
       MessageDlg(sFilename + Chr(13) + Chr(10) + FileErrToStr(StgErrToFileErr(hr)), mtError, [mbOK], 0);
end;


destructor  TMPFile.Destroy;
var    i: integer;
begin
   fCurrentFrameGroup.Free;
   frameGroupList.Free;

	if analogWndList <> nil then
   begin
   	if analogWndList.Count > 0 then
       	for i := 0 to analogWndList.Count - 1 do
       	begin
           	TAnalogFrm(analogWndList.Items[i]).mpFile := nil;
           	TAnalogFrm(analogWndList.Items[i]).Close;
           	if Mainform.bAppClosing then Application.ProcessMessages;
       	end;
   	analogWndList.Free;
   end;

   {a lot of Windows voodoo here - why it seems to work beats me}
   if viewerList.Count > 0 then
       for i := 0 to viewerList.Count - 1 do
       begin
           TViewerFrm(viewerList.Items[i]).mpFile := nil;
           TViewerFrm(viewerList.Items[i]).Close;
           if Mainform.bAppClosing then Application.ProcessMessages;
       end;
   viewerList.Free;

   frameCommentList.Free;
  { for i := 0 to MAX_CH - 1 do channels[i] := nil;}

   propertyStorage := nil;
   propertySetStorage := nil;
   rootStorage := nil;
   Mainform.FileList.NotifyFileClosing(self);
end;


{********************************* TFileList **********************************}

{------------------------------- TFileList: private ---------------------------}

function    TFileList.IsMPFile(const sFilename: string): boolean;
{var clsid: TCLSID;}
begin
  { Result := False;}
   {if Succeeded(StgIsStorageFile(StringToOLEStr(sFilename)) then}
  { if Succeeded(GetClassFile(StringToOLEStr(sFilename), clsid)) then
       Result := IsEqualGUID(clsid, GUID_MPD);}
   Result := True;
end;


function    TFileList.GetTemporaryFileName: string;
var    i: integer;
begin
   i := 1;
   Result := 'Workspace ' + IntToStr(i);
   while IndexOf(Result) >= 0 do
   begin
       i := i + 1;
       Result := 'Workspace ' + IntToStr(i);
   end;
end;


function    TFileList.Load(const sFilename: string): TFileErr;
begin
   if not IsMPFile(sFilename) then
       Result := feNotAMPFile
   else if IndexOf(sFilename) >= 0 then
      Result := feFileAlreadyOpen
   else
   begin
       AddObject(sFilename, TMPFile.CreateFromFile(sFilename));
       Result := feOK;
   end;
end;


{------------------------------- TFileList: public ---------------------------}

procedure   TFileList.FillComboBoxWithWorkspaces(aComboBox: TComboBox);
var    i: integer;
begin
   aComboBox.Clear;
   if Count > 0 then
   begin
       for i := 0 to Count - 1 do
           if TMPFile(Objects[i]).IsMemoryFile then
               aComboBox.Items.AddObject(TMPFile(Objects[i]).Filename, Objects[i]);
       if aComboBox.Items.Count > 0 then aComboBox.ItemIndex := 0;
   end;
end;


procedure   TFileList.NewFile(templateFile: TMPFile);
var    tempFileName: string;
begin
   tempFileName := GetTemporaryFileName;
   AddObject(tempFileName, TMPFile.CreateFromTemplate(tempFileName, templateFile));
end;


procedure   TFileList.NotifyFileClosing(Sender: TMPFile);
var    i: integer;
begin
   if not Mainform.bAppClosing then
   begin
       i := IndexOfObject(Sender);
       if i >= 0 then Delete(i);
   end;
end;


function    TFileList.WorkspaceCount: integer;
var    i: integer;
begin
   Result := 0;
   if Count > 0 then
       for i := 0 to Count - 1 do
           if TMPFile(Objects[i]).IsMemoryFile then Result := Result + 1;
end;


procedure   TFileList.SaveFileAs(mpFile: TMPFile; const newName: string);
var    i: integer;
       fe: TFileErr;
begin
   i := IndexOfObject(mpFile);
   if i >= 0 then
       if TMPFile(Objects[i]).IsMemoryFile then
       begin
           fe := TMPFile(Objects[i]).SaveAs(newName);
           if fe = feOK then
               {swaps names}
               Strings[i] := newName
           else
               MessageDlg('Cannot save ' + Strings[i] + ' as' + CRLF +
                       newName + CRLF +
                       FileErrToStr(fe), mtError, [mbOK], 0);
       end;
end;


procedure   TFileList.Open(const sFilename: string);
var    fe: TFileErr;
begin
   if IndexOf(sFilename) >= 0 then
       MessageDlg(Format(sAlreadyLoaded, [sFilename]), mtInformation, [mbOK], 0)
   else
   begin
       fe := Load(sFilename);
       if fe <> feOK then
           MessageDlg(FileErrToStr(fe), mtError, [mbOK], 0)
       else
           dataDirectory := ExtractFilePath(sFilename);
   end;
end;


destructor  TFileList.Destroy;
var    i: integer;
begin
   if Count > 0 then
       for i := 0 to Count - 1 do
           TMPFile(Objects[i]).Free;
   inherited Destroy;
end;




end.
