unit mpviewu;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Forms, StdCtrls, COMObj,
  ActiveX, Registry;


const
   MAX_ANALOG_CH_COUNT = 2;
   MAX_VIDEO_CH_COUNT = 2;
   MAX_CH = 4;
   f_MAX_Y_VALUE = 2047;

   BASE_CLOCK = 5e-8; {50 ns base clock = 20 MHz clock frequency}
   {the GUID for multiphoton data files}
   GUID_MPD: TGUID = '{5BC02769-74F0-47DF-929E-2E5D3630D9B5}';

   CR = Chr(13);
   TAB = Chr(9);
   CRLF = #13#10;

   MAX_FALSE_COLORS = 2048;
   MAX_GRAY_LEVELS = 256;

   DEFAULT_MAX_PIXEL_VALUE = 2047;

  // TypeLibrary Major and minor versions for Matlab
  MLAppMajorVersion = 1;
  MLAppMinorVersion = 0;

  LIBID_MLApp: TGUID = '{554F6052-79D4-11D4-B067-009027BA5F81}';

  IID_IMLApp: TGUID = '{669CEC92-6E22-11CF-A4D6-00A024583C19}';
  IID_IMLEval: TGUID = '{27BEA9CE-A20C-430F-87C2-1CC8BA31C3A8}';
  DIID_DIMLApp: TGUID = '{669CEC93-6E22-11CF-A4D6-00A024583C19}';
  DIID_DIMLEval: TGUID = '{6E813920-23FE-4D6D-91F8-56FAB06C5D13}';
  CLASS_MLApp_: TGUID = '{554F6053-79D4-11D4-B067-009027BA5F81}';
  CLASS_MLEval: TGUID = '{00854C9D-4827-4BC5-8A7D-770E696DF6A3}';

  slutType = 'Lut Type';


type
   int16 = smallint;
   int32 = integer;


   TFrameResolution = (RESOLUTION_8_BITS, RESOLUTION_12_BITS, RESOLUTION_16_BITS);
   TChannelResolution = TFrameResolution;
   TPrecision = (PREC_8_BIT, PREC_10_BIT, PREC_12_BIT, PREC_14_BIT, PREC_16_BIT);
	TFullScaleVal = (pm_42V, pm_20V, pm_10V, pm_5V, pm_2V, pm_1V, pm_0_5V, pm_0_2V);
	TInputRange = Set Of TFullScaleVal;

   TPrefix = (	tpXENNO, tpYOCTO, tpZEPTO, tpATTO, tpFEMTO, tpPICO, tpNANO,
   		tpMICRO, tpMILLI, tpUNITY, tpKILO, tpMEGA, tpGIGA, tpTERA, tpPETA,
           tpECTA, tpZETTA, tpYOTTA, tpXENNA, tpZERO, tpNONE );

   TScanMode = (SM_MOVIE, SM_STACK, SM_LINESCAN, SM_REPEAT_LINESCAN, SM_REGIONSCAN,
   	SM_FASTSTACK);

   TRGBTripleArray = array[0..Maxint div 8] of TRGBTriple;
   TpRGBTripleArray = ^TRGBTripleArray;

	TColorScheme = (CS_GRAYSCALE, CS_FALSECOLORS, CS_CUSTOMLUT);


{*****************************************************************************}
{*									FILES								  	  *}
{*****************************************************************************}

   TFileErr = 	   (feOK,
   				feCannotFindFile,
                   fePathNotFound,
                   feTooManyFilesOpened,
                   feAccessDenied,
                   feBadFileType,
                   feBadVersion,
                   feForceConversion,
                   feDiskFull,
                   feFileIsNotStorage,
   				feOutOfMemory,
                   feBadDiskDrive,
                   feCannotReadFile,
                   feUnknownError,
                   feBadData,
                   feUnexpectedEOF,
                   feShareViolation,
                   feInvalidHandle,
                   feFileNotAssigned,
                   feFileNotOpen,
                   feFileNotOpenForInput,
                   feFileNotOpenForOutput,
                   feInvalidInput,
                   feInvalidName,
                   feNotAMPFile,
                   feFileAlreadyOpen
                   );

 	TStorageNameErr = (sneOK, sneInvalidName, sneTooLong, sneInvalidChar, sneNumChar,
   	sneReservedName, sneAlreadyExists);

   TBaseColorsArray = array[0..MAX_CH - 1 {channel index}, 0..2 {RGB}] of boolean;
   TRangeColorsArray = array[0..MAX_CH - 1] of TRGBTriple;
   TMaxPixelsArray = array[0..MAX_CH - 1] of integer;

  //******************************** GRAPHICS *********************************

   TMinMaxPt = record
       MaxPt, MinPt: TPoint;
   end;

   TTrace = array of TMinMaxPt;
   TpTrace = ^TTrace;

//*************************** REGION OF INTEREST ****************************

TSimpleROI = class
private
	function    GetCenter: TPoint; virtual; abstract;
	function    GetPixels(pixindex: integer): TPoint; virtual; abstract;
	function    GetPixelCount: integer; virtual; abstract;
public
	fchannel, fROIIndex: integer;
	procedure   Draw(aBitmap: TBitmap; xOffset: integer); virtual; abstract;
   function    PtInROI(aPt: TPoint): boolean; virtual; abstract;

	property    Channel: integer read fChannel;
	property    Center: TPoint read GetCenter;
	property    PixelCount: integer read GetPixelCount;
	property    Pixels[pixindex: integer]: TPoint read GetPixels;
end;



TRectangularROI = class(TSimpleROI)
private
	function    GetCenter: TPoint; override;
	function    GetPixels(pixindex: integer): TPoint; override;
	function    GetPixelCount: integer; override;
public
	roiRect: TRect;
	procedure   Draw(aBitmap: TBitmap; xOffset: integer); override;
	function    PtInROI(pt: TPoint): boolean; override;
end;


TEllipticalROI = class(TSimpleROI)
private
	ellipticalRegion: THandle;
   fPixelCount: integer;
   pixelArray: array of TPoint;
	function    GetCenter: TPoint; override;
	function    GetPixels(pixindex: integer): TPoint; override;
	function    GetPixelCount: integer; override;
public
	roiRect: TRect;
	procedure   Draw(aBitmap: TBitmap; xOffset: integer); override;
   procedure	FindPixels;
	function    PtInROI(pt: TPoint): boolean; override;

	destructor  Destroy; override;
end;


TObjectROI = class(TSimpleROI)
private
	ptsList: TList;
	function    GetCenter: TPoint; override;
	function    GetPixels(pixindex: integer): TPoint; override;
	function    GetPixelCount: integer; override;
public
	procedure   Draw(aBitmap: TBitmap; xOffset: integer); override;
	function    PtInROI(pt: TPoint): boolean; override;
	procedure   AddPt(pt: TPoint);

	constructor Create;
	destructor  Destroy; override;
end;


TROIList = class(TList)
public
	procedure   AddRectangularROI(chIndex: integer; rc: TRect);
   procedure	AddEllipticalROI(chIndex: integer; rc: TRect);
	function    GetROIs(roiIndex: integer): TSimpleROI;
	function    PtInObject(aPt: TPoint): TObjectROI;
   function    ROIOfPt(chIndex: integer; aPt: TPoint): integer;
	procedure   Clear; override;
	destructor  Destroy; override;

	property    ROIs[roiIndex: integer]: TSimpleROI read GetROIs;
end;


// ============================ Registry Methods ===============================

function  StrParse(var fullStr: string; delim: Char): string;
procedure SavePosToRegistry(aForm: TForm; regini: TRegistryIniFile; const section, entry: string);
procedure RestorePosFromRegistry(aForm: TForm; regini: TRegistryIniFile; const section, entry: string;
	bMainForm: boolean);
function GetLongFilePath(const FileName : String):string;

{*****************************************************************************}
{*							Transfer functions								  *}
{*****************************************************************************}

function PrefixToString(prefix: TPrefix): string;
function PrefixToExponent(prefix: TPrefix): integer;
function PrefixToFactor(prefix: TPrefix): double;
function ExpToPrefixString(exponent: integer): string;
function GetPrefixFromValue(value: double): TPrefix;

{*****************************************************************************}
{*							List Box Functions								  *}
{*****************************************************************************}

{this procedure initializes a list box with prefixes and unit}
procedure FillUnitListBox(listBox: TComboBox; sUnit: string);

{this function returns the index of the prefix in a list box filled with prefix-unit}
function UnitPrefixToListBoxIndex(prefix: integer): integer;
function ListBoxIndexToUnitPrefix(listBoxIndex: integer): integer;

procedure FillInputRangeListBox(listBox: TComboBox); {fill a list box with ns, us, etc...}
function InputRangeToString(inputRange: TFullScaleVal): string;

function FullScaleToVal(fs: TFullScaleVal): double;


// ============================ File I/O Methods ===============================

function  IOErrToFileErr(code: integer): TFileErr;
function  FileErrToStr(fe: TFileErr): string;
function  StgErrToFileErr(hr: HResult): TFileErr;


// ============================ Miscellaneous ==================================

function  FindCommonRegion(start1, end1, start2, end2: integer;
   var commonStart, commonEnd: integer): boolean;

procedure NormalizeRect(var rect: TRect);
function  PrecisionToString(aPrecision: TPrecision): string;
function  PointStrictlyInRect(const aPt: TPoint; aRect: TRect): boolean;
function  ScanModeToString(smode: TScanMode): string;
procedure MakeRectFromPts(left, top, right, bottom: integer; var rectarray: array of TPoint);


{******************************************************************************}
{*}                                                                          {*}
{*}                            IMPLEMENTATION                                {*}
{*}                                                                          {*}
{******************************************************************************}

uses Math, inifiles, Dialogs, Shlobj ;

resourcestring

	// FILE I/O ERRORS
	sIOErr_FileNotFound				= 'File not found.';
	sIOErr_PathNotFound				= 'Path not found.';
	sIOErr_TooManyOpenFiles			= 'Too many open files.';
	sIOErr_AccessDenied				= 'File access denied.';
	sIOErr_InvalidHandle			= 'Invalid file handle.';
	sIOErr_NotEnoughMemory			= 'Insufficient memory for this operation.';
	sIOErr_InvalidFileAccessCode	= 'Invalid file access code.';
	sIOErr_InvalidData				= 'Invalid data.';
	sIOErr_NotEnoughStorage			= 'Not enough storage.';
	sIOErr_InvalidDrive				= 'Invalid drive.';
	sIOErr_CannotWrite				= 'Error writing to file.';
	sIOErr_CannotRead				= 'Error reading from file.';
	sIOErr_SharingViolation			= 'Share violation error.';
	sIOErr_EOF						= 'Disk read error, read past end of file.';
	sIOErr_DiskFull					= 'Disk write error, disk full.';
	sIOErr_FileNotAssigned			= 'File not assigned.';
	sIOErr_NotACompoundFile			= 'Invalid file.';
	sIOErr_InvalidName				= 'Invalid file name.';
	sIOErr_Unexpected				= 'Unexpected error.';
   sIOErr_InvalidFileType			= 'Invalid file type.';
   sIOErr_InvalidVersion			= 'Invalid file version.';
	sIOErr_FileIsNotStorage			= 'The file is not a compound file.';
   sIOErr_BadData					= 'Corrupted data in file.';
	sIOErr_ForceConversion			= 'Forced file conversion.';
	sIOErr_FileNotOpen				= 'File not opened.';
	sIOErr_FileNotOpenForInput		= 'File not opened for input.';
	sIOErr_FileNotOpenForOutput		= 'File not opened for output.';
	sIOErr_InvalidInput				= 'Invalid input.';
   sIOErr_NotAMPFile               = 'The file is not a MP data file.';
   sIOErr_FileAlreadyOpen          = 'The file is already opened.';

   // Storage name error
   sSNE_InvalidName				= 'Invalid name.';
   sSNE_TooLong					= 'Name too long.';
   sSNE_InvalidChar				= 'Invalid character in name.';
	sSNE_NumChar					= 'Invalid numerical character in name.';
   sSNE_ReservedName				= 'Reserved name.';
   sSNE_AlreadyExist				= 'Name already exists.';

type
   EMPConfig = class(Exception);
   TPropSpecArray = array[0..1000] of TPropSpec;
   TpPropSpecArray = ^TPropSpecArray;
   TPropVariantArray = array[0..1000] of TPropVariant;
   TpPropVariantArray = ^TPropVariantArray;
   TStatPropStgArray = array[0..1000] of TStatPropStg;
   TpStatPropStgArray = ^TStatPropStgArray;


const
   FMTID_User_Defined_Properties: TGUID = '{D5CDD505-2E9C-101B-9397-08002B2CF9AE}';
   sAlreadyLoaded = '%s' + CRLF + 'is already open.';

function  PrecisionToString(aPrecision: TPrecision): string;
begin
   case aPrecision of
       PREC_8_BIT: Result := '8-bit';
       PREC_10_BIT: Result := '10-bit';
       PREC_12_BIT: Result := '12-bit';
       PREC_14_BIT: Result := '14-bit';
       else Result := '16-bit';
   end;
end;


procedure MakeRectFromPts(left, top, right, bottom: integer; var rectarray: array of TPoint);
begin
   rectarray[0].x := left; rectarray[0].y := top;
   rectarray[1].x := right; rectarray[1].y := top;
   rectarray[2].x := right; rectarray[2].y := bottom;
   rectarray[3].x := left; rectarray[3].y := bottom;
   rectarray[4].x := left; rectarray[4].y := top;
end;

// ============================ Registry Methods ===============================

function StrParse(var fullStr: string; delim: Char): string;
var	delimPos: integer;
begin
	delimPos := Pos(delim, fullStr);
   if delimPos > 0 then
   begin
   	Result := Copy(fullStr,1,Pred(delimPos));
       fullStr := Copy(fullStr,Succ(delimPos),Length(fullStr));
   end
   else
   	Result := fullStr;
end;


procedure SavePosToRegistry(aForm: TForm; regini: TRegistryIniFile; const section, entry: string);
var	buffer: array[0..79] of Char;
		windowPlacement: TWindowPlacement;
begin
	windowPlacement.Length := SizeOf(windowPlacement);
   GetWindowPlacement(aForm.Handle, @windowPlacement);
   WVSPrintf(buffer, '%i,%i,%i,%i,%i,%i,%i,%i,%i,%i,%i', @windowPlacement);
   regini.WriteString(section, entry, StrPas(buffer));
end;


procedure RestorePosFromRegistry(aForm: TForm; regini: TRegistryIniFile; const section, entry: string;
	bMainForm: boolean);
var	buffer: string;
		windowPlacement: TWindowPlacement;
begin
	buffer := regini.ReadString(section, entry, '');
   FillChar(windowPlacement, SizeOf(windowPlacement), 0);
	windowPlacement.Length := SizeOf(windowPlacement);
   if buffer <> '' then
   begin
       StrToIntDef(StrParse(buffer, ','), 0);
       with windowPlacement do
       begin
       	flags := StrToInt(StrParse(buffer, ','));
       	showCmd := StrToInt(StrParse(buffer, ','));
       	ptMinPosition.x := StrToInt(StrParse(buffer, ','));
       	ptMinPosition.y := StrToInt(StrParse(buffer, ','));
       	ptMaxPosition.x := StrToInt(StrParse(buffer, ','));
       	ptMaxPosition.y := StrToInt(StrParse(buffer, ','));
       	rcNormalPosition.Left := StrToInt(StrParse(buffer, ','));
       	rcNormalPosition.Top := StrToInt(StrParse(buffer, ','));
       	rcNormalPosition.Right := StrToInt(StrParse(buffer, ','));
       	rcNormalPosition.Bottom := StrToInt(StrParse(buffer, ','));

           case ShowCmd of
           	sw_showMinimized,
               sw_showminnoactive,
               sw_minimize:
               	aForm.WindowState := wsMinimized;
               sw_showmaximized:
               	aForm.WindowState := wsMaximized;
			end;
       end;
		SetWindowPlacement(aForm.Handle, @windowPlacement);
   end
   else
   	if bMainForm then
   	with windowPlacement do
		begin
       	showCmd := SW_SHOW;
           if not SystemParametersInfo(SPI_GETWORKAREA, 0, @rcNormalPosition, 0) then
	           rcNormalPosition := Rect(0, 0, 799, 599);
			SetWindowPlacement(aForm.Handle, @windowPlacement);
		end;
end;


{*****************************************************************************}
{*							Transfer functions								  *}
{*****************************************************************************}


function PrefixToString(prefix: TPrefix): string;
begin
	case prefix of
   tpXENNO: Result := 'x';
   tpYOCTO: Result := 'y';
   tpZEPTO: Result := 'z';
   tpATTO: Result := 'a';
   tpFEMTO: Result := 'f';
   tpPICO: Result := 'p';
   tpNANO: Result := 'n';
   tpMICRO: Result := #181;
   tpMILLI: Result := 'm';
   tpUNITY : Result := '';
   tpKILO: Result := 'k';
   tpMEGA: Result := 'M';
   tpGIGA: Result := 'G';
   tpTERA: Result := 'T';
   tpPETA: Result := 'P';
   tpECTA: Result := 'E';
   tpZETTA: Result := 'Z';
   tpYOTTA: Result := 'Y';
   tpXENNA: Result := 'X';
   tpNONE: Result := '';
   else
   	Result := '';
   end;
end;

function PrefixToFactor(prefix: TPrefix): double;
begin
	case prefix of
   tpXENNO: Result := 1E-27;
   tpYOCTO: Result := 1E-24;
   tpZEPTO: Result := 1E-21;
   tpATTO: Result := 1E-18;
   tpFEMTO: Result := 1E-15;
   tpPICO: Result := 1E-12;
   tpNANO: Result := 1E-9;
   tpMICRO: Result := 1E-6;
   tpMILLI: Result := 1E-3;
   tpUNITY : Result := 1;
   tpKILO: Result := 1E+3;
   tpMEGA: Result := 1E+6;
   tpGIGA: Result := 1E+9;
   tpTERA: Result := 1E+12;
   tpPETA: Result := 1E+15;
   tpECTA: Result := 1E+18;
   tpZETTA: Result := 1E+21;
   tpYOTTA: Result := 1E+24;
   tpXENNA: Result := 1E+27;
   tpNONE: Result := 1;
   else
   	Result := 1;
   end;
end;


function PrefixToExponent(prefix: TPrefix): integer;
begin
	case prefix of
   tpXENNO: Result := -27;
   tpYOCTO: Result := -24;
   tpZEPTO: Result := -21;
   tpATTO: Result := -18;
   tpFEMTO: Result := -15;
   tpPICO: Result := -12;
   tpNANO: Result := -9;
   tpMICRO: Result := -6;
   tpMILLI: Result := -3;
   tpUNITY : Result := 0;
   tpKILO: Result := +3;
   tpMEGA: Result := +6;
   tpGIGA: Result := +9;
   tpTERA: Result := +12;
   tpPETA: Result := +15;
   tpECTA: Result := +18;
   tpZETTA: Result := +21;
   tpYOTTA: Result := +24;
   tpXENNA: Result := +27;
   tpNONE: Result := 0;
   else
   	Result := 0;
   end;
end;

function ExpToPrefixString(exponent: integer): string;
begin
	case exponent of
   -27: Result := 'x';
   -24: Result := 'y';
   -21: Result := 'z';
   -18: Result := 'a';
   -15: Result := 'f';
   -12: Result := 'p';
   -9: Result := 'n';
   -6: Result := #181;
   -3: Result := 'm';
   0 : Result := '';
   3: Result := 'k';
   6: Result := 'M';
   9: Result := 'G';
   12: Result := 'T';
   15: Result := 'P';
   18: Result := 'E';
   21: Result := 'Z';
   24: Result := 'Y';
   27: Result := 'X';
   else
   	Result := '';
   end;
end;

{This function chooses the most appropriate prefix for the value
if value is too small, returns tpZERO}
function GetPrefixFromValue(value: double): TPrefix;
begin
	value := Abs(value);
   if value <= 1E-30 then
   	Result := tpZERO
	else if value < 1E-24 then
   	Result := tpXENNO
	else if value < 1E-21 then
   	Result := tpYOCTO
	else if value < 1E-18 then
   	Result := tpZEPTO
	else if value < 1E-15 then
   	Result := tpATTO
	else if value < 1E-12 then
   	Result := tpFEMTO
	else if value < 1E-9 then
   	Result := tpPICO
	else if value < 1E-6 then
   	Result := tpNANO
	else if value < 1E-3 then
   	Result := tpMICRO
	else if value < 1E-0 then
   	Result := tpMILLI
	else if value < 1E+3 then
   	Result := tpUNITY
	else if value < 1E+6 then
   	Result := tpKILO
	else if value < 1E+9 then
   	Result := tpMEGA
	else if value < 1E+12 then
   	Result := tpGIGA
	else if value < 1E+15 then
   	Result := tpTERA
	else if value < 1E+18 then
   	Result := tpPETA
	else if value < 1E+21 then
   	Result := tpECTA
	else if value < 1E+24 then
   	Result := tpZETTA
	else if value < 1E+27 then
   	Result := tpYOTTA
	else if value < 1E+30 then
   	Result := tpXENNA
	else
   	Result := tpXENNA;
end;

{*****************************************************************************}
{*							List Box Functions								  *}
{*****************************************************************************}


procedure FillUnitListBox(listBox: TComboBox; sUnit: string);
begin
	with listBox, listBox.Items do
   begin
   	Clear;
		Add('x' + sUnit);
		Add('y' + sUnit);
		Add('z' + sUnit);
		Add('a' + sUnit);
		Add('f' + sUnit);
		Add('p' + sUnit);
		Add('n' + sUnit);
		Add(#181 + sUnit);
		Add('m' + sUnit);
		Add(sUnit);
		Add('k' + sUnit);
		Add('M' + sUnit);
		Add('G' + sUnit);
		Add('T' + sUnit);
		Add('P' + sUnit);
		Add('E' + sUnit);
		Add('Z' + sUnit);
		Add('Y' + sUnit);
		Add('X' + sUnit);
   end;
end;


function UnitPrefixToListBoxIndex(prefix: integer): integer;
begin
	Result := (27 + prefix) div 3;
end;


function ListBoxIndexToUnitPrefix(listBoxIndex: integer): integer;
begin
	Result := listBoxIndex * 3 - 27;
end;


procedure FillTimeListBox(listBox: TComboBox); {fill a list box with ns, us, etc...}
begin
	with listBox, listBox.Items do
   begin
   	Clear;
       Add('ns');
       Add(Chr(181) + 's');
       Add('ms');
       Add('s');
       Add('min');
       Add('hr');
   end;
end;


procedure FillInputRangeListBox(listBox: TComboBox);
var    i: TFullScaleVal;
begin
	with listBox, listBox.Items do
   begin
   	Clear;
       for i := pm_42V to pm_0_2V do
           Add(InputRangeToString(i));
   end;
end;


function FullScaleToVal(fs: TFullScaleVal): double;
begin
	case fs of
       pm_42V: Result := 42.0;
       pm_20V: Result := 20.0;
       pm_10V: Result := 10.0;
       pm_5V: Result := 5.0;
       pm_2V: Result := 2.0;
       pm_1V: Result := 1.0;
       pm_0_5V: Result := 0.5;
       pm_0_2V: Result := 0.2;
	else
   	Result := 0;
   end;
end;


function InputRangeToString(inputRange: TFullScaleVal): string;
begin
	case inputRange of
       pm_42V: Result := Chr(177) + '42V';
       pm_20V: Result := Chr(177) + '20V';
       pm_10V: Result := Chr(177) + '10V';
       pm_5V: Result := Chr(177) + '5V';
       pm_2V: Result := Chr(177) + '2V';
       pm_1V: Result := Chr(177) + '1V';
       pm_0_5V: Result := Chr(177) + '0.5V';
       pm_0_2V: Result := Chr(177) + '0.2V';
	else
   	Result := '';
   end;
end;

function ScanModeToString(smode: TScanMode): string;
begin
   case smode of
       SM_MOVIE: Result := 'Movie';
       SM_STACK: Result := 'Image Stack';
       SM_LINESCAN: Result := 'Line Scan';
       SM_REPEAT_LINESCAN: Result := 'Repeat Line Scan';
       SM_REGIONSCAN: Result := 'Region Scan';
   else Result := 'Fast Stack';
   end
end;


function IOErrToFileErr(code: integer): TFileErr;
begin
	case code of
   	2: Result :=  feCannotFindFile;
       3: Result :=  fePathNotFound;
       4: Result :=  feTooManyFilesOpened;
       5: Result :=  feAccessDenied;
       6: Result :=  feInvalidHandle;
       8: Result :=  feOutOfMemory;
       100: Result :=  feUnexpectedEOF;
       101: Result :=  feDiskFull;
		102: Result :=  feFileNotAssigned;
       103: Result :=  feFileNotOpen;
       104: Result :=  feFileNotOpenForInput;
       105: Result :=  feFileNotOpenForOutput;
       106: Result :=  feInvalidInput;
       else
       	Result := feUnknownError;
   end;
end;


function FileErrToStr(fe: TFileErr): string;
begin
	case fe of
		feCannotFindFile: Result := sIOErr_FileNotFound;
		fePathNotFound: Result := sIOErr_PathNotFound;
		feTooManyFilesOpened: Result := sIOErr_TooManyOpenFiles;
		feAccessDenied: Result := sIOErr_AccessDenied;
		feBadFileType: Result := sIOErr_InvalidFileType;
		feBadVersion: Result := sIOErr_InvalidVersion;
		feForceConversion: Result := sIOErr_ForceConversion;
		feDiskFull: Result := sIOErr_DiskFull;
		feFileIsNotStorage: Result := sIOErr_FileIsNotStorage;
		feOutOfMemory: Result := sIOErr_NotEnoughMemory;
		feBadDiskDrive: Result := sIOErr_InvalidDrive;
		feCannotReadFile: Result := sIOErr_CannotRead;
		feUnknownError: Result := sIOErr_Unexpected;
		feBadData: Result := sIOErr_BadData;
		feUnexpectedEOF: Result := sIOErr_EOF;
		feShareViolation: Result := sIOErr_SharingViolation;
       feInvalidHandle: Result := sIOErr_InvalidHandle;
       feFileNotAssigned: Result := sIOErr_FileNotAssigned;
		feFileNotOpen: Result := sIOErr_FileNotOpen;
       feFileNotOpenForInput: Result := sIOErr_FileNotOpenForInput;
       feFileNotOpenForOutput: Result := sIOErr_FileNotOpenForOutput;
       feInvalidInput: Result := sIOErr_InvalidInput;
       feInvalidName: Result := sIOErr_InvalidName;
       feNotAMPFile: Result := sIOErr_NotAMPFile;
       feFileAlreadyOpen: Result := sIOErr_FileAlreadyOpen;
       else
       Result := sIOErr_Unexpected;
   end;
end;


// *****************************************************************************
//
//                             ROI Methods
//
// *****************************************************************************


// *************************** Rectangular ROI *********************************


function    TRectangularROI.GetCenter: TPoint;
begin
	Result.x := (roiRect.Right + roiRect.Left) div 2;
	Result.y := (roiRect.Bottom + roiRect.Top) div 2;
end;


function    TRectangularROI.GetPixels(pixindex: integer): TPoint;
var	rWidth, yOffset: integer;
begin
	rWidth := roiRect.Right - roiRect.Left + 1;
	yOffset := pixIndex div rWidth;
	Result.x := roiRect.Left + pixindex - yOffset * rWidth;
	Result.y := roiRect.Top + yOffset;
end;


function    TRectangularROI.GetPixelCount: integer;
begin
	Result := (roiRect.Right - roiRect.Left + 1) * (roiRect.Bottom - roiRect.Top + 1);
end;


procedure   TRectangularROI.Draw(aBitmap: TBitmap; xOffset: integer);
{var	s: string;
       sWidth, sHeight: integer;}
begin
	if fChannel = 0 then xOffset := 0;
  	with aBitmap.Canvas do
	begin
		Pen.Color := clWhite;
		Pen.Style := psDot;
		Pen.Width := 1;
		Pen.Mode := pmCopy;
		Polyline([Point(roiRect.Left + xOffset, roiRect.Top),
			 Point(roiRect.Right + xOffset, roiRect.Top),
			  Point(roiRect.Right + xOffset, roiRect.Bottom),
			  Point(roiRect.Left + xOffset, roiRect.Bottom),
             Point(roiRect.Left + xOffset, roiRect.Top)]);

      { TextFlags := 0;
		Font.Name := 'Arial';
		Font.Color := clWhite;
		Font.Size := 10;
		s := IntToStr(fROIIndex);
		sWidth := TextWidth(s);
		sHeight := TextHeight(s);
		TextOut(Center.x + xOffset - sWidth div 2, Center.y - sHeight div 2, s);}
	end;
end;


function TRectangularROI.PtInROI(pt: TPoint): boolean;
begin
   Result := PointStrictlyInRect(pt, roiRect);
end;

type TPPoint = ^TPoint;

function    TObjectROI.GetCenter: TPoint;
var	i: integer;
	ppt: TPPoint;
begin
	Result.x := 0;
	Result.y := 0;
	if ptsList.Count > 0 then
	begin
		for i := 0 to ptsList.Count - 1 do
		begin
			ppt := TPPoint(ptsList.Items[i]);
			Result.x := Result.x + ppt^.x;
			Result.y := Result.y + ppt^.y;
		end;
		Result.x := Result.x div ptsList.Count;
		Result.y := Result.y div ptsList.Count;
	end;
end;


// *************************** Elliptical ROI *********************************

function    TEllipticalROI.GetCenter: TPoint;
begin
	Result.x := (roiRect.Right + roiRect.Left) div 2;
	Result.y := (roiRect.Bottom + roiRect.Top) div 2;
end;


function    TEllipticalROI.GetPixels(pixindex: integer): TPoint;
begin
	if pixindex < PixelCount then
   	Result := pixelArray[pixindex]
   else
   	Result := Point(0, 0);
end;


function    TEllipticalROI.GetPixelCount: integer;
begin
	Result := fPixelCount;
end;


procedure	 TEllipticalROI.FindPixels;
var	i, j, nextPt: integer;
begin
	ellipticalRegion := CreateEllipticRgnIndirect(roiRect);
	fPixelCount := 0;
   for i := roiRect.Top to roiRect.Bottom do
   	for j := roiRect.Left to roiRect.Right do
       	if PtInRegion(ellipticalRegion, j, i) then
           	fPixelCount := fPixelCount + 1;
   SetLength(pixelArray, fPixelCount);
   nextPt := 0;
   for i := roiRect.Top to roiRect.Bottom do
   	for j := roiRect.Left to roiRect.Right do
       	if PtInRegion(ellipticalRegion, j, i) then
           begin
           	pixelArray[nextPt] := Point(j, i);
           	nextPt := nextPt + 1;
           end;
end;


procedure   TEllipticalROI.Draw(aBitmap: TBitmap; xOffset: integer);
begin
	if fChannel = 0 then xOffset := 0;
  	with aBitmap.Canvas do
	begin
		Pen.Color := clWhite;
		Pen.Style := psDot;
		Pen.Width := 1;
		Pen.Mode := pmCopy;
		Arc(roiRect.Left + xOffset,
       	roiRect.Top,
       	roiRect.Right + xOffset,
           roiRect.Bottom,
           roiRect.Left + xOffset,
       	roiRect.Top,
           roiRect.Left + xOffset,
       	roiRect.Top);
	end;
end;


function    TEllipticalROI.PtInROI(pt: TPoint): boolean;
begin
	Result := PtInRegion(ellipticalRegion, pt.x, pt.y);
end;


destructor  TEllipticalROI.Destroy;
begin
	DeleteObject(ellipticalRegion);
   inherited Destroy;
end;

// *************************** Object ROI *********************************

function    TObjectROI.GetPixels(pixindex: integer): TPoint;
var	ppt: TPPoint;
begin
	ppt := TPPoint(ptsList.Items[pixindex]);
	Result.x := ppt^.x;
	Result.y := ppt^.y;
end;


function    TObjectROI.GetPixelCount: integer;
begin
	Result := ptsList.Count;
end;


procedure   TObjectROI.Draw(aBitmap: TBitmap; xOffset: integer);
var   {	s: string;
	sWidth, sHeight: integer;}
	pLine: TpRGBTripleArray;
	pt : TPoint;
	pPixel : ^TRGBTriple;
   i: integer;
begin
	if fChannel = 0 then xOffset := 0;
	if ptsList.Count < 0 then Exit;
	for i := 0 to PixelCount - 1 do
	begin
		pt := Pixels[i];
		pLine := aBitmap.ScanLine[pt.y];
		pPixel := @pLine[xOffset + pt.x];
		pPixel^.rgbtBlue := 255;
		pPixel^.rgbtGreen := 255;
		pPixel^.rgbtRed := 255;
	end;

 {	with aBitmap.Canvas do
	begin
		Font.Name := 'Arial';
		Font.Color := clBlack;
		Font.Size := 10;
		s := IntToStr(fROIIndex);
		sWidth := TextWidth(s);
		sHeight := TextHeight(s);
		TextOut(Center.x + xOffset - sWidth div 2, Center.y - sHeight div 2, s);
	end; }
end;



function    TObjectROI.PtInROI(pt: TPoint): boolean;
var	i: integer;
   ppt: TPPoint;
begin
	Result := False;
	i := 0;
	if ptsList.Count > 0 then
		while not Result and (i < ptsList.Count) do
		begin
			ppt := TPPoint(ptsList.Items[i]);
			if (ppt^.x = pt.x) and (ppt^.y = pt.y) then
				Result := True
			else
				i := i + 1;
		end;
end;


procedure   TObjectROI.AddPt(pt: TPoint);
var	ppt: TPPoint;
begin
	GetMem(ppt, SizeOf(TPoint));
	ppt^.x := pt.x;
	ppt^.y := pt.y;
	ptsList.Add(ppt);
end;


constructor TObjectROI.Create;
begin
	ptsList := TList.Create;
	ptsList.Capacity := 1000;
end;


destructor  TObjectROI.Destroy;
var	i: integer;
begin
	if ptsList.Count >= 0 then
		for i := 0 to ptsList.Count - 1 do
			Freemem(ptsList.Items[i], SizeOf(TPoint));
	ptsList.Free;
	inherited Destroy;
end;




procedure   TROIList.AddRectangularROI(chIndex: integer; rc: TRect);
var	aRectROI: TRectangularROI;
begin
	NormalizeRect(rc);
	aRectROI := TRectangularROI.Create;
	aRectROI.fchannel := chIndex;
	aRectROI.fROIIndex := Count + 1;
	aRectROI.roiRect := rc;
	Add(aRectROI);
end;


procedure   TROIList.AddEllipticalROI(chIndex: integer; rc: TRect);
var	aEllROI: TEllipticalROI;
begin
	NormalizeRect(rc);
	aEllROI := TEllipticalROI.Create;
	aEllROI.fchannel := chIndex;
	aEllROI.fROIIndex := Count + 1;
	aEllROI.roiRect := rc;
   aEllROI.FindPixels; {populates the array of pixels}
	Add(aEllROI);
end;


function    TROIList.GetROIs(roiIndex: integer): TSimpleROI;
begin
	Result := TSimpleROI(Items[roiIndex]);
end;


function    TROIList.PtInObject(aPt: TPoint): TObjectROI;
var	i: integer;
begin
	Result := nil;
	if Count> 0 then
		for i := 0 to Count - 1 do
			if ROIs[i] is TObjectROI then
				if ROIs[i].PtInROI(aPt) then
					Result := TObjectROI(ROIs[i]);
end;


function    TROIList.ROIOfPt(chIndex: integer; aPt: TPoint): integer;
var    i: integer;
begin
   Result := -1;  i := 0;
   while (i < Count) and (Result = -1) do
   begin
       if (ROIs[i].fchannel = chIndex) and ROIs[i].PtInROI(aPt) then
           Result := i
       else
           i := i + 1;
   end;
end;


procedure   TROIList.Clear;
var	i: integer;
begin
	if Count > 0 then
		for i := 0 to Count - 1 do
			ROIs[i].Free;
	inherited Clear;
end;


destructor  TROIList.Destroy;
begin
	Clear;
	Inherited Destroy;
end;


function FindCommonRegion(start1, end1, start2, end2: integer;
	var commonStart, commonEnd: integer): boolean;
begin
	{make sure that end1 <= start1}
	if start1 > end1 then
   begin
   	commonStart := start1;
       start1 := end1;
       end1 := commonStart;
   end;

	if start2 > end2 then
   begin
   	commonStart := start2;
       start2 := end2;
       end2 := commonStart;
   end;

   if start1 > start2 then
   begin
   	commonStart := start1;
       commonEnd := end1;
//       start1 := start2; no need
       end1 := end2;
       start2 := commonStart;
       end2 := commonEnd;
   end;

   if start2 > end1 then
   	Result := False
   else
   begin
   	commonStart := start2;
       if end2 <= end1 then
       	commonEnd := end2
       else
       	commonEnd := end1;
       Result := True;
   end;
end;


procedure NormalizeRect(var rect: TRect);
var	i: integer;
begin
	if rect.Left > rect.Right then
   begin
   	i := rect.Left;
       rect.Left := rect.Right;
       rect.Right := i;
   end;
	if rect.Top > rect.Bottom then
   begin
   	i := rect.Top;
       rect.Top := rect.Bottom;
       rect.Bottom := i;
   end;
end;


function StgErrToFileErr(hr: HResult): TFileErr;
begin
	{if integer(hr) = STG_E_INVALIDFUNCTION then
  	else} if integer(hr) = STG_E_FILENOTFOUND then
   	Result := feCannotFindFile
  	else if integer(hr) = STG_E_PATHNOTFOUND then
   	Result := fePathNotFound
  	else if integer(hr) = STG_E_TOOMANYOPENFILES then
   	Result := feTooManyFilesOpened
  	else if integer(hr) = STG_E_ACCESSDENIED then
   	Result := feAccessDenied
{  	else if integer(hr) = STG_E_INVALIDHANDLE then
   	Result := ;}
  	else if integer(hr) = STG_E_INSUFFICIENTMEMORY then
   	Result := feOutOfMemory
  	else if integer(hr) = STG_E_INVALIDPOINTER then
   	Result := feBadData
{  	else if integer(hr) = STG_E_NOMOREFILES then
   	Result := ;}
  	else if integer(hr) = STG_E_DISKISWRITEPROTECTED then
   	Result := feCannotReadFile
  	else if integer(hr) = STG_E_SEEKERROR then
   	Result := feCannotReadFile
  	else if integer(hr) = STG_E_WRITEFAULT then
   	Result := feUnexpectedEOF
  	else if integer(hr) = STG_E_READFAULT then
   	Result := feCannotReadFile
  	else if integer(hr) = STG_E_SHAREVIOLATION then
   	Result := feShareViolation
{  	else if integer(hr) = STG_E_LOCKVIOLATION then
   	Result := ;
  	else if integer(hr) = STG_E_FILEALREADYEXISTS then
   	Result := ;
  	else if integer(hr) = STG_E_INVALIDPARAMETER then
   	Result := ;}
  	else if integer(hr) = STG_E_MEDIUMFULL then
   	Result := feDiskFull
{  	else if integer(hr) = STG_E_ABNORMALAPIEXIT then
   	Result := ;
  	else if integer(hr) = STG_E_INVALIDHEADER then
   	Result := ;}
  	else if integer(hr) = STG_E_INVALIDNAME then
   	Result := feInvalidName
  	else if integer(hr) = STG_E_UNKNOWN then
   	Result := feUnknownError
{  	else if integer(hr) = STG_E_UNIMPLEMENTEDFUNCTION then
   	Result := ;
  	else if integer(hr) = STG_E_INVALIDFLAG then
   	Result := ;}
   else
   	Result := feUnknownError;
end;


function PointStrictlyInRect(const aPt: TPoint; aRect: TRect): boolean;
begin
   NormalizeRect(aRect);
   if (aPt.X >= aRect.Left) and (aPt.X <= aRect.Right) and
       (aPt.Y >= aRect.Top) and (aPt.Y <= aRect.Bottom) then
       Result := True
   else
       Result := False;
end;


 {==================
 Get Long File Path
 ==================}
function GetLongFilePath(const FileName : String):string;
var   fncGetLongPathName: function (lpszShortPath: LPCTSTR;
      lpszLongPath: LPTSTR; cchBuffer: DWORD): DWORD stdcall;
      szBuffer: array[0..MAX_PATH] of Char;
begin
  Result := '';
  // try to use the function "GetLongPathNameA" (Win98/2000 and up)
  @fncGetLongPathName := GetProcAddress(
  GetModuleHandle('Kernel32.dll'), 'GetLongPathNameA');
  if (Assigned(fncGetLongPathName)) then
		if fncGetLongPathName(PChar(FileName), szBuffer, SizeOf(szBuffer)) <> 0
			then Result := szBuffer;
end;


end.
