unit oldmpfileu;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Forms, Registry, StdCtrls, ActiveX,
       ComObj, mpviewu, vfw;


type
   TMPFile = class;


   TFrame = class
   private
       chIndex, frameIndex: integer;
       mpFile: TMPFile;
       procedure   LoadFromFile(frameIndex: integer); virtual; abstract;
   public
       data: array of int16;
       constructor Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
   end;


   TVideoFrame = class(TFrame)
   private
       function    GetHeight: integer;
       function    GetPixels(x, y: integer): int16;
       function    GetWidth: integer;
       procedure   LoadFromFile(frameIndex: integer); override;
       procedure   SaveToFile;
   public
       procedure   CopyData(dest: TVideoFrame);
       procedure   CopyToDoubleArray(var dblArray: array of double);
       procedure   GammaCorrection(blackLevel, whiteLevel: integer);
       procedure   GetProfile(binSize: integer; var histogram: array of integer);

		constructor Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);

       property    Height: integer read GetHeight;
       property    Pixels[x, y: integer]: int16 read GetPixels;
       property    Width: integer read GetWidth;
   end;


   TAnalogFrame = class(TFrame)
   {x: index of analog sample in frame data}
   private
       function    GetIAnalogValue(x: integer): integer;
       function    GetfAnalogValue(x: integer): double;
       procedure   LoadFromFile(frameIndex: integer); override;
   public
       function    SampleToAbsoluteTime(x: integer): double; {in seconds}

		constructor Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);

       property    iAnalogValue[x: integer]: integer read GetIAnalogValue;
       property    fAnalogValue[x: integer]: double read GetfAnalogValue;
   end;


   TFrameCommentList = class(TStringList)
   private
       mpFile: TMPFile;
       procedure   SaveFrameComments;
       procedure   LoadFrameComments;
       function    GetFrameComment(frameIndex: integer): string;
       procedure   SetFrameComment(frameIndex: integer; const frameComment: string);
   public
       constructor Create(theMPFile: TMPFile);
       property    FrameComment[frameIndex: integer]: string read GetFrameComment write SetFrameComment;
   end;


   TFrameGroup = class
   private
       fWidth, fHeight, fFrameIndex: integer;
       mpFile: TMPFile;
       procedure   LoadFromFile(frameIndex: integer);
     {  procedure   SaveToFile;}
       procedure   SetFrameWidth(newWidth: integer);
       procedure   SetFrameHeight(newHeight: integer);
       procedure   SetFrameIndex(newIndex: integer);
   public
       comments: string;
       channels: array[0..MAX_CH - 1] of TFrame; {0..1: TVideoFrame; 2..3: TAnalogFrame}

       property    FrameIndex: integer read fFrameIndex write SetFrameIndex;
       property    FrameWidth: integer read fWidth write SetFrameWidth;
       property    FrameHeight: integer read fHeight write SetFrameHeight;
       constructor Create(mp_file: TMPFile);
       constructor CreateTempFrameGroup(mp_file: TMPFile);
       destructor  Destroy; override;
   end;


   //-------------------------------- FILE OBJECT ------------------------------
   TMPFile = class
   private
       fActiveFrameIndex: integer;
       fAngle: integer;
       {fbLineForXScan, OBSOLETE}
       fbIsMemoryFile, fbClosing: boolean;
       fChConvFactors: array[0..MAX_CH -1] of double;
       fChDataPtsPerFrames: array[0..MAX_CH -1] of integer;
       fChEnabled: array[0..MAX_CH - 1] of boolean;
       fChInputRanges: array[0..MAX_CH - 1] of TFullScaleVal;
       fChNames: array[0..MAX_CH - 1] of string;
       fChOffsets: array[0..MAX_CH -1] of double;
       fChPrefixes: array[0..MAX_CH -1] of TPrefix;
       fChUnits: array[0..MAX_CH - 1] of string;
       fFilename: string;
       fFrameCount: integer;
       fFrameHeight,
       fFrameWidth: integer;
       fLineRepeatCount: integer;
       fMagnification: double;
       fPixelClock {in increments of 50 ns (= 20 MHz)}: integer;
       fResolution: TFrameResolution; {RESOLUTION_8_BITS, 12, 16 bits}
       fRotation: integer;
       fScanMode: TScanMode;
       fStackAveragingCount,
       fStackFrameCount: integer;
       fStackInterval: double;
       fStackRepeatCount,
       fXFrameOffset,
       fXStagePosition,
       fYFrameOffset,
       fYStagePosition: integer;
       fZStagePosition: double;
       sComments: string;

		RegionPts: array of TPoint; {for region scans}

       chStreams: array[0..MAX_CH -1] of IStream;
       propertyStorage: IPropertyStorage;
       propertySetStorage: IPropertySetStorage;
       rootStorage: IStorage;

       fCurrentFrameGroup: TFrameGroup;
       frameCommentList: TFrameCommentList;
       frameGroupList: TList;
       bDirty: boolean;

     {  function    GetActiveFrameGroup: TFrameGroup;}
       procedure   AdjustFrameCount;
       function    GetAnalogChCount: integer;
       function    GetAnalogWndCount: integer;
       function    GetAnalogWnds(wndIndex: integer): TForm;
       function    GetChAcquisitionRate(chIndex: integer): double;
       function    GetChConvFactor(chIndex: integer): double;
       function    GetChDataPtsPerFrame(chIndex: integer): integer;
       function    GetChEnabled(chIndex: integer): boolean;
       function    GetChInputRange(chIndex: integer): TFullScaleVal;
       function    GetChNames(chIndex: integer): string;
       function    GetChUnit(chIndex: integer): string;
       function    GetChPrefix(chIndex: integer): TPrefix;
       function    GetChOffset(chIndex: integer): double;
      { function    GetCurrentFrame(chIndex: integer): TFrame;}
       function    GetFrameComment(frameIndex: integer): string;
       function    GetFrameCount: integer;
       function    GetFrameRate: double; {in frames per seconds}
       function	GetFrameSize: integer;
       function    GetFrames(frameIndex: integer): TFrameGroup;
       function	GetMaxPixelValue: integer;
       procedure	GetRegions;
       function	GetResolution: string;
       function    GetVideoChCount: integer;
       function    GetViewerCount: integer;
       function    GetViewers(viewerIndex: integer): TForm;
       procedure   LoadFrameGroup(newIndex: integer);
       procedure   ReadPropertySet;
       procedure   SetActiveFrameIndex(newIndex: integer);
       procedure   SetFrameComment(frameIndex: integer; sComments: string);
       procedure   WritePropertySet;

   public
       viewerList, analogWndList: TList;

       procedure   AddFrameGroup(newFrameGroup: TFrameGroup);
       procedure   AverageFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
       procedure   Close(Sender: TObject);
       procedure   CopyAnalogData(chIndex, fromFrame, toFrame: integer; var memHandle: THandle);
       procedure   CopyFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile); {from one file to a memory file, rect to rect}
       procedure	CloneROIs(ROIList: TROIList; fromIndex: integer); {duplicate each ROI but with opposite channel}
       procedure   DetectROIs(ROIList: TROIList; chIndex, fromFrame, toFrame, threshold, minArea: integer);
       function    DigitalToAnalog(chIndex, iValue: integer): double;
       procedure	DoBackgroundCorrection;
       function	GetAverage(frameIndex, chIndex: integer; rc: TRect): integer;
       function    GetROIAverageValue(ROIList: TROIList; roiIndex, frameIndex: integer): integer;
       function    GetPixelValue(frameIndex, chIndex, xData, yData: integer): int16;
       function	GetMax(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
       function	GetMin(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
       function	GetPropertyValue(propName: string): string;
       function    IsOperationOK(chIndex, fromFrame, toFrame: integer; dest: TMPFile): boolean;
       procedure   MakeAVIMovie(const avifilename: string; chIndex, fromFrame, toFrame, newframeRate: integer; bFalseColors: boolean);
       function    MakeTIFF(const avifilename: string; chIndex, fromFrame, toFrame: integer): boolean;
       procedure   NewAnalogWnd;
       procedure   NewViewer;
       procedure   OnNewFrames;
       procedure   OnWndClose(wnd: TForm); {when a viewer or analog wnd closes}
		function	SizeOfFrameCompatible(width, height: integer): boolean;
       procedure   StackX(chIndex, fromFrame, toFrame, fromY, toY: integer; dest: TMPFile);
       procedure   StackY(chIndex, fromFrame, toFrame, fromX, toX: integer; dest: TMPFile);
       procedure   StackZ(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
       function    SaveAs(const sFilename: string): TFileErr;
       procedure   SubtractFrame(chIndex, plusFrameIndex, minusFrameIndex: integer; dest: TMPFile);

       constructor CreateFromTemplate(const sFilename: string; templateFile: TMPFile);
       constructor CreateFromFile(const sFilename: string);
       destructor  Destroy; override;

       property    ActiveFrameIndex: integer read fActiveFrameIndex write SetActiveFrameIndex;
       property    AnalogChCount: integer read GetAnalogChCount;
       property    AnalogWndCount: integer read GetAnalogWndCount;
       property    AnalogWnds[wndIndex: integer]: TForm read GetAnalogWnds;
       property    Angle: integer read fAngle;
       property    ChAcquisitionRate[chIndex: integer]: double read GetChAcquisitionRate;
       property    ChConvFactor[chIndex: integer]: double read GetChConvFactor;
       property    ChDataPtsPerFrame[chIndex: integer]: integer read GetChDataPtsPerFrame;
       property    ChEnabled[chIndex: integer]: boolean read GetChEnabled;
       property    ChInputRange[chIndex: integer]: TFullScaleVal read GetChInputRange;
       property    ChNames[chIndex: integer]: string read GetChNames;
       property    ChUnit[chIndex: integer]: string read GetChUnit;
       property    ChPrefix[chIndex: integer]: TPrefix read GetChPrefix;
       property    ChOffset[chIndex: integer]: double read GetChOffset;
       property    Closing: boolean read fbClosing; {when user closes file from menu}
       property    Comments: string read sComments write sComments;
       property    Filename: string read ffilename;
       property    FrameComment[frameIndex: integer]: string read GetFrameComment
                       write SetFrameComment;
       property    FrameCount: integer read GetFrameCount;
       property    FrameHeight: integer read fFrameHeight;
       property    FrameRate: double read GetFrameRate; {in seconds}
       property	FrameSize: integer read GetFrameSize; {in samples}
       property    Frames[frameIndex: integer]: TFrameGroup read GetFrames;
       property    FrameWidth: integer read fFrameWidth;
       property    IsMemoryFile: boolean read fbIsMemoryFile;
       property    LineRepeatCount: integer read fLineRepeatCount;
     {  property    LineForXScan: boolean read fbLineForXScan;}
       property    Magnification: double read fMagnification;
       property	MaxPixelValue: integer read GetMaxPixelValue; {255, 2047 or 32767}
       property    PixelClock: integer read fPixelClock; {in ns}
       property	Resolution: string read GetResolution;
       property    Rotation: integer read fRotation;
       property    VideoChCount: integer read GetVideoChCount;
       property    ViewerCount: integer read GetViewerCount;
       property    Viewers[viewerIndex: integer]: TForm read GetViewers;
       property    ScanMode: TScanMode read fScanMode;
       property    StackAveragingCount: integer read fStackAveragingCount;
       property    StackFrameCount: integer read fStackFrameCount;
       property    StackInterval: double read fStackInterval;
       property    StackRepeatCount: integer read fStackRepeatCount;
       property    XFrameOffset: integer read fXFrameOffset;
       property    XStagePosition: integer read fXStagePosition;
       property    YFrameOffset: integer read fYFrameOffset;
       property    YStagePosition: integer read fYStagePosition;
       property    ZStagePosition: double read fZStagePosition;
   end;


   //------------------------------- FILE LIST ---------------------------------

   {holds all the files}

   TFileList = class(TStringList)
   private
       function    IsMPFile(const sFilename: string): boolean;
       function    GetTemporaryFileName: string;
       function    Load(const sFilename: string): TFileErr;
   public
       dataDirectory: string;
       procedure   FillComboBoxWithWorkspaces(aComboBox: TComboBox);
       procedure   NotifyFileClosing(Sender: TMPFile);
       procedure   NewFile(templateFile: TMPFile);
       procedure   Open(const sFilename: string);
       procedure   SaveFileAs(mpFile: TMPFile; const newName: string);
       function    WorkspaceCount: integer;

       destructor  Destroy; override;
   end;


function BoolToString(b: Boolean): string;
{function LineForXScanToString(bLine: boolean): string; OBSOLETE}
{function StringToLineForXScan(s: string): boolean;}
function ScanModeToString(smode: TScanMode): string;
function StringToScanMode(s: string): TScanMode;
function StringToBool(s: string): Boolean;
function StringToInputRange(s: string): TFullScaleVal;
function StringToPrefix(s: string): TPrefix;

{******************************************************************************}
{*}                                                                          {*}
{*}                            IMPLEMENTATION                                {*}
{*}                                                                          {*}
{******************************************************************************}


uses mainfrm, Math, analogu, vieweru, dialogs;

const
	TIFF_TAG_COUNT = 22; {Includes ImageDescription tag}
   TIFF_ASCII = 2;
   TIFF_SHORT = 3;
   TIFF_LONG = 4;
   TIFF_RATIONAL = 5;

   TIFF_ImageDescription = 270;
   FILE_PROP_COUNT = 40; {40 properties per file}

type
	TTIFFHeader = packed record
		order,
   	signature: Word;
       IFDOffset: integer;
   end;


   TTagRecord = packed record
   	tagID, tagType: Word;
       count, dataOffset: integer;
   end;

   TIFD = packed record
   	count: Word;
       tagRecords: array[1..TIFF_TAG_COUNT] of TTagRecord;
       nextIDFoffset: integer;
       XNum, XDenom,
       YNum, YDenom: integer;
   end;
   {followed by image description ASCII string}

const
	TiffHeader: TTIFFHeader = (order: $4949; signature: 42; IFDOffset: 8);

   OFLAGS = STGM_DIRECT or STGM_READWRITE or STGM_CREATE or STGM_SHARE_EXCLUSIVE;
   RFLAGS = STGM_DIRECT or STGM_READ or STGM_SHARE_EXCLUSIVE;

   FMTID_User_Defined_Properties: TGUID = '{D5CDD505-2E9C-101B-9397-08002B2CF9AE}';

	sTIFFPropNames: string =
   	'Channel: %d' + #13#10 +
   	'Resolution: %s' + #13#10 +
       'Scan Mode: %s' + #13#10 +
       'X Stage Position (microns): %d' + #13#10 +
       'Y Stage Position (microns): %d' + #13#10 +
       'Z Stage Position (microns): %f' + #13#10 +
       'Section Count: %d' + #13#10 +
       'z- Interval (microns): %f' + #13#10 +
       'Frames Per Section: %d' + #13#10 +
       'Stack Repeat Count: %d' + #13#10 +
       'Magnification: x%f' + #13#10 +
       'Rotation (degrees): %d' + #13#10 +
       'X Frame Offset (pixels): %d' + #13#10 +
       'Y Frame Offset (pixels): %d' + #13#10 +
       'Frame Rate (fps): %f';

   sPropNames: array[0..FILE_PROP_COUNT - 1] of string = (
   	'Resolution',
       'Scan Mode',
       'Frame Count',
       'X Position',
       'Y Position',
       'Z Position',
       'Stack Count',
       'z- Interval',
       'Averaging Count',
       'Repeat Count',
       'Magnification',
       'Rotation',
       'X Frame Offset',
       'Y Frame Offset',
       'Frame Width',
       'Frame Height',
       'Pixel Clock',
       'Channel name (1)', 'Enabled (1)', 'Input Range (1)',
       'Channel name (2)', 'Enabled (2)', 'Input Range (2)',
       'Channel name (3)', 'Enabled (3)', 'Input Range (3)', 'Channel Unit (3)',
       'Channel Prefix (3)', 'Conversion factor (3)', 'Offset (3)', 'Data Points Per Frame (3)',
       'Channel name (4)', 'Enabled (4)', 'Input Range (4)', 'Channel Unit (4)',
       'Channel Prefix (4)', 'Conversion factor (4)', 'Offset (4)', 'Data Points Per Frame (4)',
       'Comments');
   sAlreadyLoaded = 'File %s is already opened in MPView.';


procedure MakeIFD(var IFD: TIFD);
begin
	with IFD do
   begin
   	count := TIFF_TAG_COUNT;

       tagRecords[1].tagID := 254; {NewSubtitleType}
       tagRecords[1].tagType := TIFF_LONG;
       tagRecords[1].count := 1;
       tagRecords[1].dataOffset := 0;

       tagRecords[2].tagID := 256; {ImageWidth}
       tagRecords[2].tagType := TIFF_LONG;
       tagRecords[2].count := 1;

       tagRecords[3].tagID := 257; {ImageLength}
       tagRecords[3].tagType := TIFF_LONG;
       tagRecords[3].count := 1;

       tagRecords[4].tagID := 258; {BitsPerSample}
       tagRecords[4].tagType := TIFF_SHORT;
       tagRecords[4].count := 1;
       tagRecords[4].dataOffset := 16;

       tagRecords[5].tagID := 259; {Compression}
       tagRecords[5].tagType := TIFF_SHORT;
       tagRecords[5].count := 1;
       tagRecords[5].dataOffset := 1; {No compression}

       tagRecords[6].tagID := 262; {PhotometricInterpretation}
       tagRecords[6].tagType := TIFF_SHORT;
       tagRecords[6].count := 1;
       tagRecords[6].dataOffset := 1; {BlackIsZero}

       tagRecords[7].tagID := 263; {Tresholding}
       tagRecords[7].tagType := TIFF_SHORT;
       tagRecords[7].count := 1;
       tagRecords[7].dataOffset := 1;

       tagRecords[8].tagID := 266; {FillOrder}
       tagRecords[8].tagType := TIFF_SHORT;
       tagRecords[8].count := 1;
       tagRecords[8].dataOffset := 1;

       tagRecords[9].tagID := 273; {StripOffsets}
       tagRecords[9].tagType := TIFF_LONG;
       tagRecords[9].count := 1;

       tagRecords[10].tagID := 274; {Orientation}
       tagRecords[10].tagType := TIFF_SHORT;
       tagRecords[10].count := 1;
       tagRecords[10].dataOffset := 1;

       tagRecords[11].tagID := 277; {SamplesPerPixel}
       tagRecords[11].tagType := TIFF_SHORT;
       tagRecords[11].count := 1;
       tagRecords[11].dataOffset := 1;

       tagRecords[12].tagID := 278; {RowsPerStrip}
       tagRecords[12].tagType := TIFF_LONG;
       tagRecords[12].count := 1;

       tagRecords[13].tagID := 279; {StripByteCounts}
       tagRecords[13].tagType := TIFF_LONG;
       tagRecords[13].count := 1;

       tagRecords[14].tagID := 280; {MinSampleValue}
       tagRecords[14].tagType := TIFF_SHORT;
       tagRecords[14].count := 1;
       tagRecords[14].dataOffset := 0;

       tagRecords[15].tagID := 281; {MaxSampleValue}
       tagRecords[15].tagType := TIFF_SHORT;
       tagRecords[15].count := 1;
       tagRecords[15].dataOffset := 65535;

       tagRecords[16].tagID := 282; {XResolution}
       tagRecords[16].tagType := TIFF_RATIONAL;
       tagRecords[16].count := 1;

       tagRecords[17].tagID := 283; {YResolution}
       tagRecords[17].tagType := TIFF_RATIONAL;
       tagRecords[17].count := 1;

       tagRecords[18].tagID := 284; {PlanarConfiguration}
       tagRecords[18].tagType := TIFF_SHORT;
       tagRecords[18].count := 1;
       tagRecords[18].dataOffset := 1; {Chunky}

       tagRecords[19].tagID := 290; {GrayResponseUnit}
       tagRecords[19].tagType := TIFF_SHORT;
       tagRecords[19].count := 1;
       tagRecords[19].dataOffset := 1;

       tagRecords[20].tagID := 296; {ResolutionUnit}
       tagRecords[20].tagType := TIFF_SHORT;
       tagRecords[20].count := 1;
       tagRecords[20].dataOffset := 2; {Inch}

       tagRecords[21].tagID := 339; {SampleFormat}
       tagRecords[21].tagType := TIFF_SHORT;
       tagRecords[21].count := 1;
       tagRecords[21].dataOffset := 1;

       tagRecords[22].tagID := TIFF_ImageDescription;
       tagRecords[22].tagType := TIFF_ASCII;

       XNum := 72;
       XDenom := 1;
       YNum := 72;
       YDenom := 1;
   end;
end;

{
OBSOLETE FUNCTIONS
function LineForXScanToString(bLine: boolean): string;
begin
   if bLine then Result := 'Line' else Result := 'Sine';
end;


function StringToLineForXScan(s: string): boolean;
begin
   Result := (s = 'Line');
end;}


function ScanModeToString(smode: TScanMode): string;
begin
   case smode of
       SM_MOVIE: Result := 'Movie';
       SM_STACK: Result := 'Image Stack';
     {  SM_STACKMOVIE: Result := 'Image Stack Movie';}
       SM_LINESCAN: Result := 'Line Scan';
       SM_REPEAT_LINESCAN: Result := 'Repeat Line Scan';
       SM_REGIONSCAN: Result := 'Region Scan';
   else Result := '';
   end
end;


function StringToScanMode(s: string): TScanMode;
begin
   if s = 'Movie' then Result := SM_MOVIE else
   if s = 'Image Stack' then Result := SM_STACK else
 {  if s = 'Image Stack Movie' then Result := SM_STACKMOVIE else}
   if s = 'Line Scan' then Result := SM_LINESCAN else
   if s = 'Repeat Line Scan' then Result := SM_REPEAT_LINESCAN else
   if s = 'Region Scan' then Result := SM_REGIONSCAN else
       Result := SM_MOVIE;
end;


function BoolToString(b: Boolean): string;
begin
   if b then Result := 'True' else Result := 'False';
end;


function StringToBool(s: string): Boolean;
begin
   Result := (s = 'True');
end;


function StringToInputRange(s: string): TFullScaleVal;
begin
   if s = Chr(177) + '42V' then Result := pm_42V else
   if s = Chr(177) + '20V' then Result := pm_20V else
   if s = Chr(177) + '10V' then Result := pm_10V else
   if s = Chr(177) + '5V' then Result := pm_5V else
   if s = Chr(177) + '2V' then Result := pm_2V else
   if s = Chr(177) + '1V' then Result := pm_1V else
   if s = Chr(177) + '0.5V' then Result := pm_0_5V else
   if s = Chr(177) + '0.2V' then Result := pm_0_2V else
       Result := pm_10V;
end;


function StringToPrefix(s: string): TPrefix;
begin
   if s = 'x' then Result := tpXENNO else
   if s = 'y' then Result := tpYOCTO else
   if s = 'z' then Result := tpZEPTO else
   if s = 'a' then Result := tpATTO else
   if s = 'f' then Result := tpFEMTO else
   if s = 'p' then Result := tpPICO else
   if s = 'n' then Result := tpNANO else
   if s = #181 then Result := tpMICRO else
   if s = 'm' then Result := tpMILLI else
   if s = '' then Result := tpUNITY else
   if s = 'k' then Result := tpKILO else
   if s = 'M' then Result := tpMEGA else
   if s = 'G' then Result := tpGIGA else
   if s = 'T' then Result := tpTERA else
   if s = 'P' then Result := tpPETA  else
   if s = 'E' then Result := tpECTA else
   if s = 'Z' then Result := tpZETTA else
   if s = 'Y' then Result := tpYOTTA else
   if s = 'X' then Result := tpXENNA else
       Result := tpUNITY;
end;

//-------------------------------- FRAME  --------------------------------------

constructor TFrame.Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
var	i: integer;
begin
   mpFile := mp_File;
   frameIndex := iFrameIndex;
   chIndex := iChIndex;
   SetLength(data, mpFile.FrameWidth * mpFile.FrameHeight);
   if mpFile.ScanMode = SM_REGIONSCAN then
   	for i := 0 to Length(data) - 1 do data[i] := 0;
end;

//------------------------------ VIDEO FRAME  ----------------------------------

function    TVideoFrame.GetHeight: integer;
begin
   Result := mpFile.FrameHeight;
end;


function    TVideoFrame.GetPixels(x, y: integer): int16;
begin
   Result := data[x + y * Width];
end;


function    TVideoFrame.GetWidth: integer;
begin
   Result := mpFile.FrameWidth;
end;


procedure   TVideoFrame.LoadFromFile(frameIndex: integer);
var    libNewPosition: Largeint;
       i: integer;
       pt: TPoint;
       tempData: array of int16;
begin
   OleCheck(mpFile.chStreams[chIndex].Seek(frameIndex * mpFile.FrameSize * SizeOf(int16)
   	, STREAM_SEEK_SET, libNewPosition));
   if mpFile.ScanMode <> SM_REGIONSCAN then
   	OleCheck(mpFile.chStreams[chIndex].Read(@data[0], mpFile.FrameSize * SizeOf(int16), nil))
   else
   begin
   	{Region scan: the frame is 500 x 500, we fill with data indexed by RegionPts}
   	SetLength(tempData, mpFile.FrameSize);
   	OleCheck(mpFile.chStreams[chIndex].Read(@tempData[0], mpFile.FrameSize * SizeOf(int16), nil));
       for i := 0 to mpFile.FrameSize - 1 do
       begin
       	pt := mpFile.RegionPts[i];
           data[pt.x + pt.y * mpFile.FrameWidth] := tempData[i];
       end;
   end;
end;


procedure   TVideoFrame.SaveToFile;
begin
   mpFile.chStreams[chIndex].Write(@data[0], Width * Height * SizeOf(int16), nil);
end;


procedure   TVideoFrame.CopyData(dest: TVideoFrame);
var    i: integer;
begin
   for i := 0 to Width * Height - 1 do
       dest.data[i] := data[i];
end;


procedure   TVideoFrame.GammaCorrection(blackLevel, whiteLevel: integer);
var    i: integer;
       pixelVal: int16;
begin
   for i := 0 to Width * Height - 1 do
   begin
       pixelVal := data[i];
       if pixelVal > whiteLevel then
           pixelVal := 2047
       else if pixelVal < blackLevel then
           pixelVal := 0
       else
           pixelVal := int16(Muldiv(integer(pixelVal) - blackLevel, 2047,
               whiteLevel - blackLevel));
       data[i] := pixelVal;
   end;
end;


procedure    TVideoFrame.GetProfile(binSize: integer; var histogram: array of integer);
var    binCount, i, j: integer;
begin
   if (binSize <= 0) or (binSize > 1024) then Exit;
   binCount := 2048 div binSize + 1;
  { SetLength(histogram, binCount);}
   for i := 0 to binCount - 1 do
       histogram[i] := 0;

   for i := 0 to Width * Height - 1 do
       for j := 0 to binCount - 1 do
       begin
           if j = 0 then
           begin
               if data[i] < 0 then histogram[0] := histogram[0] + 1;
           end
           else
               if (data[i] >= (j - 1) * binSize) and (data[i] < (j * binSize)) then
                   histogram[j] := histogram[j] + 1;
       end;
end;


procedure   TVideoFrame.CopyToDoubleArray(var dblArray: array of double);
var    i: integer;
begin
   for i := 0 to Width * Height - 1 do
       dblArray[i] := data[i];
end;


constructor TVideoFrame.Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
begin
	inherited Create(mp_File, ichIndex, iFrameIndex);
   SetLength(data, mpFile.FrameWidth * mpFile.FrameHeight);
end;

///------------------------------ ANALOG FRAME  ---------------------------------

function    TAnalogFrame.SampleToAbsoluteTime(x: integer): double; {in seconds}
begin
   Result := (frameIndex * mpFile.ChDataPtsPerFrame[chIndex] + x) * mpFile.FrameRate /
       mpFile.ChDataPtsPerFrame[chIndex];
end;


function    TAnalogFrame.GetIAnalogValue(x: integer): integer;
begin
   Result := data[x];
end;


function    TAnalogFrame.GetfAnalogValue(x: integer): double;
begin
   Result := mpFile.DigitalToAnalog(chIndex, data[x]);
end;


procedure   TAnalogFrame.LoadFromFile(frameIndex: integer);
var    libNewPosition: Largeint;
begin
   mpFile.chStreams[chIndex].Seek(STREAM_SEEK_SET,
       frameIndex * mpFile.ChDataPtsPerFrame[chIndex] * SizeOf(int16), libNewPosition);
   mpFile.chStreams[chIndex].Read(@data[0], mpFile.ChDataPtsPerFrame[chIndex] * SizeOf(int16), nil);
end;


constructor TAnalogFrame.Create(mp_File: TMPFile; iChIndex, iFrameIndex: integer);
begin
	inherited Create(mp_File, ichIndex, iFrameIndex);
   SetLength(data, mpFile.fChDataPtsPerFrames[iChIndex]);
end;


//--------------------------- FRAME COMMENT LIST  ------------------------------

procedure   TFrameCommentList.SaveFrameComments;
var    commentStream: IStream;
       i, j: integer;
       s: string;
begin
   if Count > 0 then
       if Succeeded(mpFile.rootStorage.CreateStream('Frame comments', OFLAGS, 0, 0, commentStream)) then
       begin
           j := Count - 1;
           commentStream.Write(@j, SizeOf(j), nil);
           for i := 0 to Count - 1 do
           begin
               j := Integer(Objects[i]);
               commentStream.Write(@j, SizeOf(j), nil);
               j := Length(Strings[i]);
               commentStream.Write(@j, SizeOf(j), nil);
               s := Strings[i];
               commentStream.Write(@s[1], j, nil);
           end;
           commentStream := nil;
       end;
end;


procedure   TFrameCommentList.LoadFrameComments;
var    commentStream: IStream;
       i, frameIndex, stringLength, frameCount: integer;
       s: string;
begin
   if Succeeded(mpFile.rootStorage.OpenStream('Frame comments', nil, RFLAGS, 0, commentStream)) then
   begin
       commentStream.Read(@frameCount, SizeOf(frameCount), nil);
       if frameCount >= 0 then
       for i := 0 to frameCount do
       begin
           commentStream.Read(@frameIndex, SizeOf(frameIndex), nil);
           commentStream.Read(@stringLength, SizeOf(stringLength), nil);
           if stringLength > 0 then
           begin
               SetLength(s, stringLength);
               commentStream.Read(@s[1], stringLength, nil);
               AddObject(s, TObject(frameIndex));
           end;
       end;
       commentStream := nil;
   end;
end;


function    TFrameCommentList.GetFrameComment(frameIndex: integer): string;
begin
   Result := '';
   if (Count > 0) and (IndexOfObject(TObject(frameIndex)) >= 0) then
      Result := Strings[IndexOfObject(TObject(frameIndex))];
end;


procedure   TFrameCommentList.SetFrameComment(frameIndex: integer; const frameComment: string);
begin
   if IndexOfObject(TObject(frameIndex)) >= 0 then
       Strings[IndexOfObject(TObject(frameIndex))] := frameComment
   else
       AddObject(frameComment, TObject(frameIndex));
end;


constructor TFrameCommentList.Create(theMPFile: TMPFile);
begin
   mpFile := theMPFile;
end;


//------------------------------ FRAME GROUP  ----------------------------------

procedure   TFrameGroup.LoadFromFile(frameIndex: integer);
var    i: integer;
begin
   for i := 0 to MAX_VIDEO_CH_COUNT - 1 do
       if channels[i] <> nil then
           channels[i].LoadFromFile(frameIndex);
end;


procedure   TFrameGroup.SetFrameIndex(newIndex: integer);
var    i: integer;
begin
   fFrameIndex := newIndex;
   for i := 0 to MAX_VIDEO_CH_COUNT - 1 do
       if channels[i] <> nil then channels[i].frameIndex := newIndex;
end;


procedure   TFrameGroup.SetFrameWidth(newWidth: integer);
begin
   fWidth := newWidth;
end;


procedure   TFrameGroup.SetFrameHeight(newHeight: integer);
begin
   fHeight := newHeight;
   if channels[0] <> nil then SetLength(channels[0].data, fWidth * fHeight);
   if channels[1] <> nil then SetLength(channels[1].data, fWidth * fHeight);
end;


constructor TFrameGroup.Create(mp_file: TMPFile);
var    i: integer;
begin
   mpFile := mp_file;
   for i := 0 to 1 do
       if mpFile.ChEnabled[i] then channels[i] := TVideoFrame.Create(mpFile, i, fFrameIndex);
   for i := 2 to 3 do
       if mpFile.ChEnabled[i] then channels[i] := TAnalogFrame.Create(mpFile, i, fFrameIndex);
end;


constructor TFrameGroup.CreateTempFrameGroup(mp_file: TMPFile);
begin
   mpFile := mp_file;
   channels[0] := TVideoFrame.Create(mpFile, 0, 0);
 {  SetLength(channels[0].data, width * height);}
end;


destructor  TFrameGroup.Destroy;
var    i: integer;
begin
   for i := 0 to MAX_VIDEO_CH_COUNT - 1 do
       channels[i].Free;
   inherited Destroy;
end;


//-------------------------------- FILE OBJECT --------------------------------

//--------------------------- FILE OBJECT: PRIVATE -----------------------------

procedure   TMPFile.AdjustFrameCount;
var    statstg: TStatStg;
       validCh: integer;
begin
   if chEnabled[0] then validCh := 0 else validCh := 1;
   chStreams[validCh].Stat(statstg, STATFLAG_NONAME);
   fFrameCount := statstg.cbSize div (FrameWidth * FrameHeight * SizeOf(int16));
end;


function    TMPFile.GetAnalogChCount: integer;
begin
   Result := 0;
   if fChEnabled[2] then Result := 1;
   if fChEnabled[3] then Result := Result + 1;
end;


function    TMPFile.GetAnalogWndCount: integer;
begin
   if analogWndList <> nil then
       Result := analogWndList.Count
   else
       Result := 0;
end;


function    TMPFile.GetAnalogWnds(wndIndex: integer): TForm;
begin
   Result := nil;
   if (wndIndex >= 0) and (wndIndex < analogWndList.Count) then
       Result := TForm(analogWndList.Items[wndIndex]);
end;


function    TMPFile.GetChAcquisitionRate(chIndex: integer): double;
begin
   Result := FrameRate * fChDataPtsPerFrames[chIndex];
end;


function    TMPFile.GetChConvFactor(chIndex: integer): double;
begin
   Result := fChConvFactors[chIndex];
end;


function    TMPFile.GetChDataPtsPerFrame(chIndex: integer): integer;
begin
   Result := fChDataPtsPerFrames[chIndex];
end;


function    TMPFile.GetChEnabled(chIndex: integer): boolean;
begin
   Result := fChEnabled[chIndex];
end;


function    TMPFile.GetChInputRange(chIndex: integer): TFullScaleVal;
begin
   Result := fChInputRanges[chIndex];
end;


function    TMPFile.GetChNames(chIndex: integer): string;
begin
   Result := fChNames[chIndex];
end;


function    TMPFile.GetChUnit(chIndex: integer): string;
begin
   Result := fChUnits[chIndex];
end;


function    TMPFile.GetChPrefix(chIndex: integer): TPrefix;
begin
   Result := fChPrefixes[chIndex];
end;


function    TMPFile.GetChOffset(chIndex: integer): double;
begin
   Result := fChOffsets[chIndex];
end;


function    TMPFile.GetFrameComment(frameIndex: integer): string;
begin
   Result := frameCommentList.FrameComment[frameIndex];
end;


function    TMPFile.GetFrameCount: integer;
begin
   if IsMemoryFile then
       Result := frameGroupList.Count
   else
       Result := fFrameCount;
end;


function    TMPFile.GetFrameRate: double; {in frames per seconds}
begin
   {50 ns = 5e-8 s per pixel clock unit}
   if ScanMode <> SM_REGIONSCAN then
   	Result := 1 / (Muldiv(5, fFrameWidth, 4) * fFrameHeight * fPixelClock * BASE_CLOCK)
   else
   	Result := 1 / (FrameSize * fPixelClock * BASE_CLOCK);
end;


function	 TMPFile.GetFrameSize: integer;
begin
	if ScanMode <> SM_REGIONSCAN then
   	Result := FrameWidth * FrameHeight
   else
   	Result := Length(RegionPts);
end;


function    TMPFile.GetFrames(frameIndex: integer): TFrameGroup;
begin
   Result := nil;
   if (frameIndex >= 0) and (frameIndex < FrameCount) then
       if IsMemoryFile then
           Result := TFrameGroup(frameGroupList.Items[frameIndex])
       else
       begin
           ActiveFrameIndex := frameIndex;
           Result := fCurrentFrameGroup;
       end;
end;


function	TMPFile.GetMaxPixelValue: integer;
begin
	case fResolution of
   	RESOLUTION_8_BITS: Result := 255;
       RESOLUTION_12_BITS: Result := 2047;
       else Result := 32767;
   end;
end;


procedure	TMPFile.GetRegions;
var	regionStream: IStream;
		statstg: TStatStg;
       regionSize: integer;
	    libNewPosition: Largeint;
begin
	{opens the stream 'Regions'}
   OleCheck(rootStorage.OpenStream('Regions', nil, RFLAGS, 0, regionStream));
   {gets the size of the stream}
   regionStream.Stat(statstg, STATFLAG_NONAME);
   regionSize := statStg.cbSize;
   OleCheck(regionStream.Seek(0, STREAM_SEEK_SET, libNewPosition));
   SetLength(RegionPts, regionSize div SizeOf(TPoint));
   OleCheck(regionStream.Read(@RegionPts[0], regionSize, nil));
end;


function	 TMPFile.GetResolution: string;
begin
	case fResolution of
   	RESOLUTION_8_BITS: Result := '8-bit';
       RESOLUTION_12_BITS: Result := '12-bit';
       else Result := '16-bit';
   end;
end;


function    TMPFile.GetVideoChCount: integer;
begin
   Result := 0;
   if fChEnabled[0] then Result := 1;
   if fChEnabled[1] then Result := Result + 1;
end;


function    TMPFile.GetViewerCount: integer;
begin
   Result := viewerList.Count;
end;


function    TMPFile.GetViewers(viewerIndex: integer): TForm;
begin
   if (viewerIndex >= 0) and (viewerIndex < viewerList.Count) then
       Result := TForm(viewerList.Items[viewerIndex])
   else
       Result := nil;
end;


procedure   TMPFile.LoadFrameGroup(newIndex: integer);
begin
   fCurrentFrameGroup.LoadFromFile(newIndex);
end;

type
   TPropSpecArray = array[0..1000] of TPropSpec;
   TPropVariantArray = array[0..1000] of TPropVariant;
   TpPropSpecArray = ^TPropSpecArray;
   TpPropVariantArray = ^TPropVariantArray;


procedure   TMPFile.ReadPropertySet;
var    ps: TpPropSpecArray;
       pv: TpPropVariantArray;
       i: integer;
       s: string;
begin
   ps := nil; pv := nil;
   try
       GetMem(ps, FILE_PROP_COUNT * SizeOf(TPropSpec));
       GetMem(pv, FILE_PROP_COUNT * SizeOf(TPropVariant));
       for i := 0 to FILE_PROP_COUNT - 1 do
       begin
           ps^[i].ulKind := PRSPEC_LPWSTR;
           ps^[i].lpwstr := StringToOLEStr(sPropNames[i]);
       end;
       OleCheck(propertyStorage.ReadMultiple(FILE_PROP_COUNT, @ps[0], @pv[0]));

       if (pv^[0].pwszVal <> nil) and (WideCharToString(pv^[0].pwszVal) = '16-bit') then
			fResolution := RESOLUTION_16_BITS
       else fResolution := RESOLUTION_12_BITS;

       s := WideCharToString(pv^[1].pwszVal); {Scan Mode}
       fScanMode := StringToScanMode(s);
{       s := WideCharToString(pv^[2].pwszVal);} {Line Scan Mode}
{       fbLineForXScan := StringToLineForXScan(s);}

       s := WideCharToString(pv^[2].pwszVal); {Frame Count'}
       if s = '' then s := '0'; fFrameCount := StrToInt(s);
       s := WideCharToString(pv^[3].pwszVal); {X Position'}
       if s = '' then s := '0'; fXStagePosition := StrToInt(s);
       s := WideCharToString(pv^[4].pwszVal); {Y Position'}
       if s = '' then s := '0'; fYStagePosition := StrToInt(s);
       s := WideCharToString(pv^[5].pwszVal); {Z Position'}
       if s = '' then s := '0'; fZStagePosition := StrToInt(s);
       s := WideCharToString(pv^[6].pwszVal); {Stack Count'}
       if s = '' then s := '0'; fStackFrameCount := StrToInt(s);
       s := WideCharToString(pv^[7].pwszVal); {z- Interval'}
       if s = '' then s := '0'; fStackInterval := StrToFloat(s);
       s := WideCharToString(pv^[8].pwszVal); {Averaging Count'}
       if s = '' then s := '0'; fStackAveragingCount := StrToInt(s);
       s := WideCharToString(pv^[9].pwszVal); {Repeat Count'}
       if s = '' then s := '0'; fStackRepeatCount := StrToInt(s);
       s := WideCharToString(pv^[10].pwszVal); {Magnification - remove 'x'}
       s := Copy(s, 2, Length(s) - 1); if s = '' then s := '1'; fMagnification := StrToFloat(s);
       s := WideCharToString(pv^[11].pwszVal); {Rotation'}
       if s = '' then s := '0'; fRotation := StrToInt(s);
       s := WideCharToString(pv^[12].pwszVal); {X Frame Offset'}
       if s = '' then s := '0'; fXFrameOffset := StrToInt(s);
       s := WideCharToString(pv^[13].pwszVal); {Y Frame Offset'}
       if s = '' then s := '0'; fYFrameOffset := StrToInt(s);
       s := WideCharToString(pv^[14].pwszVal); {Frame Width'}
       if s = '' then s := '0'; fFrameWidth := StrToInt(s);
       s := WideCharToString(pv^[15].pwszVal); {Frame Height}
       if s = '' then s := '0'; fFrameHeight := StrToInt(s);
       s := WideCharToString(pv^[16].pwszVal); {Pixel Clock}
       if s = '' then s := '0'; fPixelClock := StrToInt(s);
       fChNames[0] := WideCharToString(pv^[17].pwszVal); {Channel name (1)}
       fChEnabled[0] := StringToBool(WideCharToString(pv^[18].pwszVal)); {Enabled (1)}
       fChInputRanges[0] := StringToInputRange(WideCharToString(pv^[19].pwszVal)); {Input Range  (1)}
       fChNames[1] := WideCharToString(pv^[20].pwszVal); {Channel name (2)}
       fChEnabled[1] := StringToBool(WideCharToString(pv^[21].pwszVal)); {Enabled (2)}
       fChInputRanges[1] := StringToInputRange(WideCharToString(pv^[22].pwszVal)); {Input Range (2)}
       fChNames[2] := WideCharToString(pv^[23].pwszVal); {Channel name (3)}
       fChEnabled[2] := StringToBool(WideCharToString(pv^[24].pwszVal)); {Enabled (3)}
       fChInputRanges[2] := StringToInputRange(WideCharToString(pv^[25].pwszVal)); {Input Range (3)}
       fChUnits[2] := WideCharToString(pv^[26].pwszVal); {Channel Unit (3)}
       fChPrefixes[2] := StringToPrefix(WideCharToString(pv^[27].pwszVal)); {Channel Prefix (3)}
       s := WideCharToString(pv^[28].pwszVal); {Conversion factor (3)}
       if s = '' then s := '0'; fChConvFactors[2] := StrToFloat(s);
       s := WideCharToString(pv^[29].pwszVal); {Offset (3)}
       if s = '' then s := '0'; fChOffsets[2] := StrToFloat(s);
       s := WideCharToString(pv^[30].pwszVal); {Data Points Per Frame (3)}
       if s = '' then s := '0'; fChDataPtsPerFrames[2] := StrToInt(s);
       fChNames[3] := WideCharToString(pv^[31].pwszVal); {Channel name (4)}
       fChEnabled[3] := StringToBool(WideCharToString(pv^[32].pwszVal)); {Enabled (4)}
       fChInputRanges[3] := StringToInputRange(WideCharToString(pv^[33].pwszVal)); {Input Range (4)}
       fChUnits[3] := WideCharToString(pv^[34].pwszVal); {Channel Unit (4)}
       fChPrefixes[3] := StringToPrefix(WideCharToString(pv^[35].pwszVal)); {Channel Prefix (4)}
       s := WideCharToString(pv^[36].pwszVal); {Conversion factor (4)}
       if s = '' then s := '0'; fChConvFactors[3] := StrToFloat(s);
       s := WideCharToString(pv^[37].pwszVal); {Offset (4)}
       if s = '' then s := '0'; fChOffsets[3] := StrToFloat(s);
       s := WideCharToString(pv^[38].pwszVal); {Data Points Per Frame (4)}
       if s = '' then s := '0'; fChDataPtsPerFrames[3] := StrToInt(s);
       sComments := WideCharToString(pv^[39].pwszVal); {Comments}
   finally
       if ps <> nil then Freemem(ps);
       if pv <> nil then Freemem(pv);
   end;
end;


procedure   TMPFile.SetActiveFrameIndex(newIndex: integer);
begin
   if (newIndex >= 0) and (newIndex < FrameCount) then
       if newIndex <> fActiveFrameIndex then
       begin
           if not IsMemoryFile then LoadFrameGroup(newIndex);
           fActiveFrameIndex := newIndex;
       end;
end;


procedure    TMPFile.SetFrameComment(frameIndex: integer; sComments: string);
begin
   frameCommentList.FrameComment[frameIndex] := sComments;
end;


procedure   TMPFile.WritePropertySet;
var    ps: TpPropSpecArray;
       pv: TpPropVariantArray;
       i: integer;
       sPropValues: array[0..FILE_PROP_COUNT - 1] of string;
begin
	sPropValues[0] := Resolution;
   sPropValues[1] := ScanModeToString(ScanMode); {Scan Mode}
   sPropValues[2] := IntToStr(FrameCount); {Frame Count'}
   sPropValues[3] := IntToStr(XStagePosition); {X Position'}
   sPropValues[4] := IntToStr(YStagePosition); {Y Position'}
   sPropValues[5] := FloatToStr(ZStagePosition); {Z Position'}
   sPropValues[6] := IntToStr(StackFrameCount); {Stack Count'}
   sPropValues[7] := FloatToStr(StackInterval); {z- Interval'}
   sPropValues[8] := IntToStr(StackAveragingCount); {Averaging Count'}
   sPropValues[9] := IntToStr(StackRepeatCount);  {Repeat Count'}
   sPropValues[10] := 'x' + FloatToStr(Magnification); {Magnification'}
   sPropValues[11] := IntToStr(Rotation);  {Rotation'}
   sPropValues[12] := IntToStr(XFrameOffset);  {X Frame Offset'}
   sPropValues[13] := IntToStr(YFrameOffset);  {Y Frame Offset'}
   sPropValues[14] := IntToStr(FrameWidth);  {Frame Width'}
   sPropValues[15] := IntToStr(FrameHeight);  {Frame Height}
   sPropValues[16] := IntToStr(PixelClock);  {Pixel Clock}
   sPropValues[17] := ChNames[0]; {Channel name (1)}
   sPropValues[18] := BoolToString(ChEnabled[0]); {Enabled (1)}
   sPropValues[19] := InputRangeToString(ChInputRange[0]); {Input Range  (1)}
   sPropValues[20] := ChNames[1]; {Channel name (2)}
   sPropValues[21] := BoolToString(ChEnabled[1]); {Enabled (2)}
   sPropValues[22] := InputRangeToString(ChInputRange[1]); {Input Range (2)}
   sPropValues[23] := ChNames[2]; {Channel name (3)}
   sPropValues[24] := BoolToString(ChEnabled[2]); {Enabled (3)}
   sPropValues[25] := InputRangeToString(ChInputRange[2]); {Input Range (3)}
   sPropValues[26] := ChUnit[2]; {Channel Unit (3)}
   sPropValues[27] := PrefixToString(ChPrefix[2]); {Channel Prefix (3)}
   sPropValues[28] := FloatToStr(ChConvFactor[2]); {Conversion factor (3)}
   sPropValues[29] := FloatToStr(ChOffset[2]); {Offset (3)}
   sPropValues[30] := IntToStr(ChDataPtsPerFrame[2]); {Data Points Per Frame (3)}
   sPropValues[31] := ChNames[3]; {Channel name (4)}
   sPropValues[32] := BoolToString(ChEnabled[3]); {Enabled (4)}
   sPropValues[33] := InputRangeToString(ChInputRange[3]); {Input Range (4)}
   sPropValues[34] := ChUnit[3]; {Channel Unit (4)}
   sPropValues[35] := PrefixToString(ChPrefix[3]); {Channel Prefix (4)}
   sPropValues[36] := FloatToStr(ChConvFactor[3]); {Conversion factor (4)}
   sPropValues[37] := FloatToStr(ChOffset[3]);{Offset (4)}
   sPropValues[38] := IntToStr(ChDataPtsPerFrame[3]); {Data Points Per Frame (4)}
   sPropValues[39] := sComments; {Comments}

   ps := nil; pv := nil;
   try
       GetMem(ps, FILE_PROP_COUNT * SizeOf(TPropSpec));
       GetMem(pv, FILE_PROP_COUNT * SizeOf(TPropVariant));
       for i := 0 to FILE_PROP_COUNT - 1 do
       begin
           ps^[i].ulKind := PRSPEC_LPWSTR;
           ps^[i].lpwstr := StringToOLEStr(sPropNames[i]);
           pv^[i].vt := VT_LPWSTR;
           pv^[i].pwszVal := StringToOLEStr(sPropValues[i]);
       end;
       OleCheck(propertyStorage.WriteMultiple(FILE_PROP_COUNT, @ps[0], @pv[0], 2));
   finally
       if ps <> nil then Freemem(ps);
       if pv <> nil then Freemem(pv);
   end;
end;


//--------------------------- FILE OBJECT: PUBLIC ------------------------------

procedure   TMPFile.AddFrameGroup(newFrameGroup: TFrameGroup);
begin
   if FrameCount = 0 then
   begin
       {allows overriding frame width and frame height when no frame in file}
       fFrameHeight := newFrameGroup.frameHeight;
       fFrameWidth := newFrameGroup.frameWidth; {sets the size of the data array}
   end;
   if (fFrameHeight = newFrameGroup.frameHeight) and (fFrameWidth = newFrameGroup.frameWidth) then
   begin
       newFrameGroup.FrameIndex := FrameCount;
       frameGroupList.Add(newFrameGroup);
   end;
end;


procedure   TMPFile.AverageFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       avgCount: integer;
       avgArray: array of integer;
       i, j: integer;
begin
   avgCount := toFrame - fromFrame + 1;

   SetLength(avgArray, FrameWidth * FrameHeight);
   for i := 0 to FrameWidth * FrameHeight - 1 do avgArray[i] := 0;

   newFrameGroup := TFrameGroup.Create(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := FrameHeight;

   for j := fromFrame to toFrame do
       for i := 0 to FrameWidth * FrameHeight - 1 do
           avgArray[i] := avgArray[i] + Frames[j].channels[chIndex].data[i];

   for i := 0 to FrameWidth * FrameHeight - 1 do
       newFrameGroup.channels[0].data[i] := avgArray[i] div avgCount;

   dest.AddFrameGroup(newFrameGroup);
end;


procedure   TMPFile.Close(Sender: TObject);
var    i: integer;
begin
   fbClosing := True;
   {closes all the windows}
   if ViewerCount > 0 then
       for i := 0 to ViewerCount - 1 do
           Viewers[i].Close;
   if AnalogWndCount > 0 then
       for i := 0 to AnalogWndCount - 1 do
           AnalogWnds[i].Close;
   Free;
end;


procedure   TMPFile.CopyAnalogData(chIndex, fromFrame, toFrame: integer; var memHandle: THandle);
begin
   if (fromFrame < 0) or (toFrame < 0) or (fromFrame >= FrameCount) or (toFrame
       >= FrameCount) or (toFrame < fromFrame) then Exit;
end;


procedure   TMPFile.CopyFrames(chIndex, fromFrame, toFrame: integer; dest: TMPFile); {from one file to a memory file, rect to rect}
var    newFrameGroup: TFrameGroup;
       i, j: integer;
begin
   for j := fromFrame to toFrame do
   begin
       newFrameGroup := TFrameGroup.Create(dest);
       newFrameGroup.FrameWidth := FrameWidth;
       newFrameGroup.FrameHeight := FrameHeight;
       for i := 0 to FrameWidth * FrameHeight - 1 do
           newFrameGroup.channels[0].data[i] := Frames[j].channels[chIndex].data[i];
       dest.AddFrameGroup(newFrameGroup);
   end;
end;


procedure	TMPFile.CloneROIs(ROIList: TROIList; fromIndex: integer); {duplicate each ROI but with opposite channel}
var 	i, j, oldCount, chIndex: integer;
		oldROI, newROI: TSimpleROI;
begin
	with ROIList do
   	if fromIndex <= Count - 1 then
       begin
       	oldCount := Count;
       	for i := fromIndex to oldCount - 1 do
           begin
           	oldROI := TSimpleROI(Items[i]);
               if oldROI.Channel = 0 then chIndex := 1 else chIndex := 0;
               if oldROI is TRectangularROI then
               begin
               	newROI := TRectangularROI.Create;
                   (newROI as TRectangularROI).roiRect := (oldROI as TRectangularROI).roiRect;
               end
             {  else if oldROI is TEllipticalROI then
               begin
               	newROI := TEllipticalROI.Create;
               end }
               else
               begin
               	newROI := TObjectROI.Create;
                   if oldROI.PixelCount > 0 then
                   	for j := 0 to oldROI.PixelCount - 1 do
							(newROI as TObjectROI).AddPt(oldROI.Pixels[j]);
               end;
               newROI.fChannel := chIndex;
               newROI.fROIIndex := ROIList.Count + 1;
               Add(newROI);
           end;
       end;
end;


const
   ALREADY_SCANNED = -32768;


procedure  TMPFile.DetectROIs(ROIList: TROIList; chIndex, fromFrame, toFrame, threshold, minArea: integer);
var	tempFrameGroup: TFrameGroup;
	    i, j: integer;
	    objectROI: TObjectROI;
       value, oldvalue: int16;

   {nested recursive procedure to find ROI in all adjacent points}
   procedure FindROIObject(i, j: integer; var aROIObject: TObjectROI);
   var	pixelValue: int16;
   begin
	    if (i < 0) or (i >= FrameWidth) or (j < 0) or (j >= FrameHeight) then Exit;
	    pixelValue := tempFrameGroup.channels[0].data[j * FrameWidth + i];
       if pixelValue >= threshold then
	    begin
		    tempFrameGroup.channels[0].data[j * FrameWidth + i] := ALREADY_SCANNED;
		    if aROIObject = nil then
           begin
               aROIObject := TObjectROI.Create;
               aROIObject.fChannel := chIndex;
               aROIObject.fROIIndex := ROIList.Count + 1;
           end;
           aROIObject.AddPt(Point(i, j));
	        FindROIObject(i - 1, j - 1, aROIObject);
	        FindROIObject(i    , j - 1, aROIObject);
	        FindROIObject(i + 1, j - 1, aROIObject);
	        FindROIObject(i - 1, j    , aROIObject);
	        FindROIObject(i + 1, j    , aROIObject);
	        FindROIObject(i - 1, j + 1, aROIObject);
	        FindROIObject(i    , j + 1, aROIObject);
	        FindROIObject(i + 1, j + 1, aROIObject);
	    end;
   end;

begin
	// create a temporary frame group
	tempFrameGroup := TFrameGroup.CreateTempFrameGroup(self);

	// maximally project frames from fromFrame to toFrame onto tempFrameGroup

	for i := 0 to FrameWidth * FrameHeight - 1 do tempFrameGroup.channels[0].data[i] := 0;

	for i := fromFrame to toFrame do
		for j := 0 to FrameWidth * FrameHeight - 1 do
		begin
			value := Frames[i].channels[chIndex].data[j];
			if value < 0 then value := 0;
			if value > 2047 then value := 2047;
			oldValue := tempFrameGroup.channels[0].data[j];
			if value > oldValue then
				tempFrameGroup.channels[0].data[j] := value;
		end;


	// search for objects in tempFrameGroup recursively
	for j := 0 to FrameHeight - 1 do
		for i := 0 to FrameWidth - 1 do
       begin
			objectROI := nil;
			FindROIObject(i, j, objectROI);
           {we found a new object}
			if objectROI <> nil then
               ROIList.Add(objectROI);
		end;

	// clean up; we eliminate all ObjectROIs that are not large enough

	if ROIList.Count > 0 then
		for i := ROIList.Count - 1 downto 0 do
			if ROIList.ROIs[i] is TObjectROI then
				if (ROIList.ROIs[i] as TObjectROI).PixelCount < minArea then
					ROIList.Delete(i);

	tempFrameGroup.Free;
end;


function    TMPFile.DigitalToAnalog(chIndex, iValue: integer): double;
begin
   Result := ChOffset[chIndex] + ChConvFactor[chIndex] * FullScaleToVal(ChInputRange[chIndex]) *
       (iValue / 2048.0);
end;


procedure	TMPFile.DoBackgroundCorrection;
var	ch, i, j, pixelIndex: integer;
		pixelVal, minValue: int16;
begin
	if FrameCount = 0 then Exit;
   for ch := 0 to 1 do
   	if ChEnabled[ch] then
       begin
       	minValue := 32767;
           for i := 0 to FrameWidth - 1 do
           	for j := 0 to FrameHeight - 1 do
               begin
               	pixelVal := Frames[ActiveFrameIndex].channels[ch].data[i + j * FrameWidth];
                   if pixelVal < minValue then minValue := pixelVal;
               end;
           if minValue < 0 then
           	for i := 0 to FrameWidth - 1 do
               	for j := 0 to FrameHeight - 1 do
                   begin
                   	pixelIndex := i + j * FrameWidth;
                   	pixelVal := Frames[ActiveFrameIndex].channels[ch].data[pixelIndex];
                       pixelVal := pixelVal - minValue;
                       Frames[ActiveFrameIndex].channels[ch].data[pixelIndex] := pixelVal;
                   end;
       end;
end;


function	TMPFile.GetAverage(frameIndex, chIndex: integer; rc: TRect): integer;
var	i, j: integer;
begin
	Result := 0;
   for i := rc.Left to rc.Right do
   	for j := rc.Top to rc.Bottom do
       	Result := Result + Frames[frameIndex].channels[chIndex].data[i + j * FrameWidth];
   Result := Result div ((rc.Right - rc.Left + 1) * (rc.Bottom - rc.Top + 1));
end;


function	TMPFile.GetMax(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
var	pixelVal, i, j: integer;
begin
	Result := -32768;
   for i := rc.Left to rc.Right do
   	for j := rc.Top to rc.Bottom do
       begin
       	pixelVal := Frames[frameIndex].channels[chIndex].data[i + j * FrameWidth];
           if pixelVal > Result then
           begin
           	Result := pixelVal;
           	x := i;
               y := j;
           end;
       end;
end;


function	TMPFile.GetMin(frameIndex, chIndex: integer; rc: TRect; var x, y: integer): integer;
var	pixelVal, i, j: integer;
begin
	Result := 32768;
   for i := rc.Left to rc.Right do
   	for j := rc.Top to rc.Bottom do
       begin
       	pixelVal := Frames[frameIndex].channels[chIndex].data[i + j * FrameWidth];
           if pixelVal < Result then
           begin
           	Result := pixelVal;
           	x := i;
               y := j;
           end;
       end;
end;


function   TMPFile.GetROIAverageValue(ROIList: TROIList; roiIndex, frameIndex: integer): integer;
var	i, j: integer;
	theROI: TSimpleROI;
	roiPixel: TPoint;
begin
	Result := 0;
	theROI := ROIList.ROIs[roiIndex];
	for i := 0 to theROI.PixelCount - 1 do
	begin
		roiPixel := theROI.Pixels[i];
		j := roiPixel.x + roiPixel.y * FrameWidth;
		Result := Result + Frames[frameIndex].channels[theROI.Channel].data[j];
	end;
	Result := Result div theROI.PixelCount;
end;


function    TMPFile.GetPixelValue(frameIndex, chIndex, xData, yData: integer): int16;
begin
   Result := 0;
   if (frameIndex >= 0) and (frameIndex < FrameCount) and (chIndex >= 0) and
       (chIndex <= 1) and (xData >= 0) and (xData < FrameWidth) and (yData >= 0)
       and (yData < FrameHeight) then
           Result := Frames[frameIndex].channels[chIndex].data[xData + yData * FrameWidth];
end;


function	TMPFile.GetPropertyValue(propName: string): string;
var    ps: TpPropSpecArray;
       pv: TpPropVariantArray;
begin
   ps := nil; pv := nil;
   try
       GetMem(ps, SizeOf(TPropSpec));
       GetMem(pv, SizeOf(TPropVariant));
       ps^[0].ulKind := PRSPEC_LPWSTR;
       ps^[0].lpwstr := StringToOLEStr(propName);
       try
       	OleCheck(propertyStorage.ReadMultiple(1, @ps[0], @pv[0]));
			Result := WideCharToString(pv^[0].pwszVal);
       except
           Result := 'No match';
       end;
   finally
       if ps <> nil then Freemem(ps);
       if pv <> nil then Freemem(pv);
   end;
end;


function    TMPFile.IsOperationOK(chIndex, fromFrame, toFrame: integer; dest: TMPFile): boolean;
begin
   Result := ChEnabled[chIndex] and (fromFrame >= 0) and (fromFrame < FrameCount) and (toFrame >= 0) and
       (toFrame < FrameCount) and (fromFrame <= toFrame) and dest.IsMemoryFile;
end;


type   EAVIError = class(Exception);
       TRGBTripleArray = array[0..Maxint div 8] of TRGBTriple;
       TpRGBTripleArray = ^TRGBTripleArray;


procedure   TMPFile.MakeAVIMovie(const avifilename: string; chIndex, fromFrame, toFrame, newframeRate: integer; bFalseColors: boolean);
var    aviFile: IAVIFile;
       aviStream: IAVIStream;
       bitmapInfo: TBITMAPINFO;
       pAviBitmap: TpRGBTripleArray;
       spCharFileName: array[0..255] of WideChar;
       spCharStreamName: array[0..63] of WideChar;
       streamInfo: TAVIStreamInfo;
       i, j, k: integer;
       hr: HResult;
const  sStreamName = '2-photon laser confocal microscope video stream';
begin
   if AVIFileOpen(aviFile,
       StringToWideChar(aviFilename, @spCharFileName, 255),
       OF_CREATE or OF_SHARE_EXCLUSIVE or OF_WRITE, nil) <> 0 then
           raise EAVIError.Create('Error in creating AVI file');
   FillChar(streamInfo, SizeOf(streamInfo), 0);
   StringToWideChar(sStreamName, @spCharStreamName, 63);
   with streamInfo do
   begin
       fccType := streamtypeVIDEO;
       fccHandler := 0;
       dwScale := 1;
       dwRate := Round(newframeRate);
       if dwRate > 24 then dwRate := 24; // bound to 24 fps
       dwSuggestedBufferSize := FrameWidth * FrameHeight;
       rcFrame := Rect(0, 0, FrameWidth - 1, FrameHeight - 1);
       for i := 0 to 63 do
       szName[i] := spCharStreamName[i];
   end;

   if AVIFileCreateStream(aviFile, aviStream, streamInfo) <> 0 then
       raise EAVIError.Create('Cannot create AVI stream');

   FillChar(bitmapInfo, SizeOf(TBITMAPINFO), 0);
   with bitmapInfo.bmiHeader do
   begin
       biSize := SizeOf(TBITMAPINFOHEADER);
       biWidth := FrameWidth;
       biHeight := FrameHeight; // we invert bitmap in the loop
       biPlanes := 1;
       biBitCount := 24;
       biCompression := BI_RGB;
   end;

   if AVIStreamSetFormat(aviStream, 0, @bitmapInfo, SizeOf(TBITMAPINFOHEADER)) <> 0 then
       raise EAVIError.Create('Cannot set format of video stream');

   GetMem(pAviBitmap, FrameHeight * FrameWidth * SizeOf(TRGBTriple));

   i := fromFrame;

   repeat
       for j := 0 to FrameHeight * FrameWidth - 1 do
       begin
           k := Frames[i].Channels[chIndex].data[  j];
           if k < 0 then k := 0;
           if k > MAX_FALSE_COLORS -  1 then k := MAX_FALSE_COLORS - 1;
           if bFalseColors then
               pAviBitmap^[j] := Mainform.falseColorTable[k]
           else
               pAviBitmap^[j] := Mainform.grayScaleTable[k];
       end;

       hr := AVIStreamWrite(aviStream, i - fromFrame, 1, pAviBitmap,
           FrameWidth * FrameHeight * SizeOf(TRGBTriple), AVIIF_KEYFRAME, nil, nil);
       i := i + 1;
   until not Succeeded(hr) or (i > toFrame);

   Freemem(pAviBitmap, FrameHeight * FrameWidth * SizeOf(TRGBTriple) );

   {instead of using the AVI API functions to release the stream and file,
   we do it the Delphi way by making them nil}
   aviStream := nil;
   aviFile := nil;

   if not Succeeded(hr) then
       raise EAVIError.Create('Error writing frame to file');
end;


function  TMPFile.MakeTIFF(const avifilename: string; chIndex, fromFrame, toFrame: integer): boolean;
var	Stream     : TFileStream;
       i, j, k, l, sLength: integer;
       pixelValue: int16;
       wPixelValues: array of Word;
       videoFrame: TFrame;
       ifd: TIFD;
       sImageDescription: string;
begin
	Result := True;
	Stream := TFileStream.Create(avifilename, fmCreate Or fmShareExclusive);
   try
   	sImageDescription := Format(sTIFFPropNames, [
   		chIndex + 1,
   		Resolution,
       	ScanModeToString(ScanMode),
       	XStagePosition,
       	YStagePosition,
       	ZStagePosition,
       	StackFrameCount,
       	StackInterval,
       	StackAveragingCount,
       	StackRepeatCount,
       	Magnification,
       	Rotation,
       	XFrameOffset,
       	YFrameOffset,
       	FrameRate ]);


   	{Write TIFF file header}
       Stream.Write(tiffHeader, SizeOf(tiffHeader));

       {Write the IFDs}
       MakeIFD(ifd);
       sLength := Length(sImageDescription) + 1; {includes the NULL finishing the string}

       with ifd do
       begin
       	tagRecords[22].Count := sLength;
       	tagRecords[2].dataOffset := FrameWidth;
       	tagRecords[3].dataOffset := FrameHeight;
       	tagRecords[12].dataOffset := FrameHeight; {RowsPerStrip}
       	tagRecords[13].dataOffset := FrameHeight * FrameWidth * SizeOf(Word); {StripByteCounts}
       end;

   	for i := fromFrame to toFrame do
       	with ifd do
       	begin
            	{XResolution}
       		ifd.tagRecords[16].dataOffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
               	- 4 * SizeOf(integer) - sLength;
               {YResolution}
       		ifd.tagRecords[17].dataOffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
               	- 2 * SizeOf(integer) - sLength;

               {ImageDescription}
				ifd.tagRecords[22].dataOffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
               	- sLength;

               {StripOffsets}
       		ifd.tagRecords[9].dataOffset := SizeOf(tiffHeader) + (toFrame - fromFrame + 1) * (SizeOf(ifd) + sLength) +
               	FrameWidth * FrameHeight * SizeOf(Word) * (i - fromFrame);

		       	if i <> toFrame then
               	nextIDFoffset := SizeOf(tiffHeader) + (i - fromFrame + 1) * (SizeOf(ifd) + sLength)
       	    else
           		nextIDFoffset := 0;

           	Stream.Write(ifd, SizeOf(ifd));

               {Write the description string}
               Stream.Write(sImageDescription[1], sLength);
           end;

		{Write frame data for each frame}
		SetLength(wPixelValues, FrameHeight * FrameWidth);
   	for i := fromFrame to toFrame do
       begin
           videoFrame := Frames[i].Channels[chIndex];
       	for j := 0 to FrameHeight - 1 do
           begin
               for k := 0 to FrameWidth - 1 do
               begin
               	l := j * FrameWidth + k;
               	pixelValue := videoFrame.data[l];

                   {truncate value and normalize to 0..65535}
                   if pixelValue < 0 then
                   begin
                   	wPixelValues[l] := 0;
                       Result := False;
                   end
                   else if fResolution = RESOLUTION_12_BITS then
                   	wPixelValues[l] := Muldiv(pixelValue, 65535, 2047)
                   else
                   	wPixelValues[l] := Muldiv(pixelValue, 65535, 32767);
               end;
           end;
			Stream.Write(wPixelValues[0], FrameHeight * FrameWidth * SizeOf(Word));
        end;
   finally
   	Stream.Free;
   end;
end;


procedure   TMPFile.NewAnalogWnd;
var    analogFrm: TAnalogFrm;
begin
   analogFrm := TAnalogFrm.Create(Mainform);
   analogFrm.Initialize(self);
   analogFrm.Show;
   analogWndList.Add(analogFrm);
end;


procedure   TMPFile.NewViewer;
var    viewerFrm: TViewerFrm;
begin
   viewerFrm := TViewerFrm.Create(Mainform);
   viewerFrm.Initialize(self, viewerList.Add(viewerFrm) + 1);
   viewerFrm.CurrentFrameIndex := 0;
   viewerFrm.Show;
end;


procedure   TMPFile.OnNewFrames;
var    i: integer;
begin
   if FrameCount <= 0 then Exit;
   if viewerList.Count > 0 then
       for i := 0 to viewerList.Count - 1 do
           TViewerFrm(viewerList.Items[i]).CurrentFrameIndex := FrameCount - 1;
end;


procedure   TMPFile.OnWndClose(wnd: TForm); {when a viewer or analog wnd closes}
var    i: integer;
begin
   if not Closing and not Mainform.bAppClosing then
       if AnalogWndCount + ViewerCount = 1 then
           {the last window standing for this file: if we close, file object is free}
           Free
       else
       begin
           if analogWndList <> nil then
           begin
               i := analogWndList.IndexOf(wnd);
               if i >= 0 then analogWndList.Remove(wnd);
           end;
           i := viewerList.IndexOf(wnd);
           if i >= 0 then viewerList.Remove(wnd);
       end;
end;


function	TMPFile.SizeOfFrameCompatible(width, height: integer): boolean;
begin
	if FrameCount = 0 then
   	Result := True
   else
   	if (width = FrameWidth) and (height = FrameHeight) then
       	Result := True
       else
       begin
       	MessageDlg( 'Operation not possible.' + CRLF +
       				'Destination workspace has different frame size.',
                       mtError, [mbOK], 0);
       	Result := False;
       end;
end;


procedure   TMPFile.StackX(chIndex, fromFrame, toFrame, fromY, toY: integer; dest: TMPFile);{creates new frame}
var    newFrameGroup: TFrameGroup;
       i, j, k: integer;
       value: int16;
       pDataElt: ^int16;
begin
   newFrameGroup := TFrameGroup.Create(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := toFrame - fromFrame + 1;

   dest.AddFrameGroup(newFrameGroup);
   for i := fromFrame to toFrame do
   begin
       ActiveFrameIndex := i;
       for j := 0 to FrameWidth - 1 do
       begin
           pDataElt := @(newFrameGroup.Channels[0].data[(i - fromFrame) * FrameWidth + j]);
           pDataElt^ := 0;
           for k := fromY to toY do
           begin
               value := Frames[ActiveFrameIndex].channels[chIndex].data[k * FrameWidth + j];
               if value > 2047 then value := 2047;
               if value > pDataElt^ then pDataElt^ := value;
           end;
       end;
       dest.OnNewFrames; {update each time we project - looks nicer!}
   end;
end;


procedure   TMPFile.StackY(chIndex, fromFrame, toFrame, fromX, toX: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       i, j, k: integer;
       value: int16;
       pDataElt: ^int16;
begin
   newFrameGroup := TFrameGroup.Create(dest);
   newFrameGroup.FrameWidth := FrameHeight;
   newFrameGroup.FrameHeight := toFrame - fromFrame + 1;

   dest.AddFrameGroup(newFrameGroup);
   for i := fromFrame to toFrame do
   begin
       ActiveFrameIndex := i;
       for j := 0 to FrameHeight - 1 do
       begin
           pDataElt := @newFrameGroup.Channels[0].data[(i - fromFrame) * FrameHeight + j];
           pDataElt^ := 0;
           for k := fromX to toX do
           begin
               value := Frames[ActiveFrameIndex].channels[chIndex].data[j * FrameWidth + k];
               if value > 2047 then value := 2047;
               if value > pDataElt^ then pDataElt^ := value;
           end;
       end;
       dest.OnNewFrames; {update each time we project - looks nicer!}
   end;
end;


procedure   TMPFile.StackZ(chIndex, fromFrame, toFrame: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       i, j, k: integer;
       value: int16;
       pDataElt: ^int16;
begin
   newFrameGroup := TFrameGroup.Create(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := FrameHeight;

   for i := 0 to FrameWidth * FrameHeight - 1 do
       newFrameGroup.Channels[0].data[i] := 0;

   dest.AddFrameGroup(newFrameGroup);
   for i := fromFrame to toFrame do
   begin
       ActiveFrameIndex := i;
       for j := 0 to FrameHeight - 1 do
           for k := 0 to FrameWidth - 1 do
           begin
               pDataElt := @newFrameGroup.Channels[0].data[j * FrameWidth + k];
               value := Frames[ActiveFrameIndex].channels[chIndex].data[j * FrameWidth + k];
               if value > 2047 then value := 2047;
               if value > pDataElt^ then pDataElt^ := value;
           end;
       dest.OnNewFrames; {update each time we project - looks nicer!}
   end;
end;


function   TMPFile.SaveAs(const sFilename: string): TFileErr;
var    hr: HResult;
       i: integer;
begin
   // create the file on disk
   hr := StgCreateDocFile(StringToOLEStr(sFilename), OFLAGS, 0, rootStorage);
   if Succeeded(hr) then
   begin
       propertySetStorage := rootStorage as IPropertySetStorage;
       propertySetStorage.Create(FMTID_User_Defined_Properties,
           FMTID_User_Defined_Properties, PROPSETFLAG_DEFAULT, OFLAGS, propertyStorage);
       rootStorage.SetClass(GUID_MPD);

       // save user information
       WritePropertySet;
       frameCommentList.SaveFrameComments;

       // change status of file and filename
       fFilename := sFilename;
       fbIsMemoryFile := False;

       fFrameCount := frameGroupList.Count;

       // opens the channels
       OLECheck(rootStorage.CreateStream('Ch0', OFLAGS, 0, 0, chStreams[0]));

       // save the frame list
       for i := 0 to frameGroupList.Count - 1 do
           (TFrameGroup(frameGroupList.Items[i]).channels[0] as TVideoFrame).SaveToFile;

       // create the current frame group (at frame Index 0)
       fActiveFrameIndex := 0;
       fCurrentFrameGroup := TFrameGroup(frameGroupList.Items[0]);

       // destroys the frame list (except first frame at index 0)
       if frameGroupList.Count> 1 then
           for i := 1 to frameGroupList.Count - 1 do
               TFrameGroup(frameGroupList.Items[i]).Free;
       frameGroupList.Free;
       frameGroupList := nil;

       Result := feOK;

       // notify all the viewers as the name change
       if viewerList.Count > 0 then
           for i := 0 to viewerList.Count - 1 do
               TForm(viewerList.Items[i]).Caption := 'Video Channels - ' +
               ExtractFileName(fFilename);
   end
   else
       Result := StgErrToFileErr(hr);
end;


procedure   TMPFile.SubtractFrame(chIndex, plusFrameIndex, minusFrameIndex: integer; dest: TMPFile);
var    newFrameGroup: TFrameGroup;
       i: integer;
       value, minValue: int16;
begin
   newFrameGroup := TFrameGroup.Create(dest);
   newFrameGroup.FrameWidth := FrameWidth;
   newFrameGroup.FrameHeight := FrameHeight;

   for i := 0 to FrameWidth * FrameHeight - 1 do
       newFrameGroup.channels[0].data[i] := Frames[plusFrameIndex].channels[chIndex].data[i];

   minValue := 2047;
   for i := 0 to FrameWidth * FrameHeight - 1 do
   begin
       value := newFrameGroup.channels[0].data[i] - Frames[minusFrameIndex].channels[chIndex].data[i];
       if value < minValue then minValue := value;
       if value > 2047 then value := 2047;
       newFrameGroup.channels[0].data[i] := value;
   end;

   {bring minimal pixel level to 0}
   if minValue < 0 then
   for i := 0 to FrameWidth * FrameHeight - 1 do
   begin
       value := newFrameGroup.channels[0].data[i] - minValue;
       if value > 2047 then value := 2047;
       newFrameGroup.channels[0].data[i] := value;
   end;

   dest.AddFrameGroup(newFrameGroup);
end;


constructor TMPFile.CreateFromTemplate(const sFilename: string; templateFile: TMPFile);
var    i: integer;
begin
   {a worskpace file has always one video channel and no analog channel}
   fActiveFrameIndex := -1;
   fAngle := templateFile.Angle;
   fbIsMemoryFile := True;
   fChEnabled[0] := True;
   fChEnabled[1] := False;
   fChEnabled[2] := False;
   fChEnabled[3] := False;
   for i := 0 to 3 do
   begin
       fChConvFactors[i] := templateFile.ChConvFactor[i];
       fChDataPtsPerFrames[i] := templateFile.ChDataPtsPerFrame[i];
       fChInputRanges[i] := templateFile.ChInputRange[i];
       fChNames[i] := templateFile.ChNames[i];
       fChPrefixes[i] := templateFile.ChPrefix[i];
       fChUnits[i] := templateFile.ChUnit[i];
       fChOffsets[i] := templateFile.ChOffset[i];
   end;
   fFilename := sFilename;
   fFrameCount := 0;
   fFrameHeight := templateFile.FrameHeight;
   fFrameWidth := templateFile.FrameWidth;
   fLineRepeatCount := templateFile.LineRepeatCount;
   fMagnification := templateFile.Magnification;
   fPixelClock := templateFile.PixelClock;
   fRotation := templateFile.Rotation;
   fScanMode := SM_MOVIE; {a workspace is always a movie}
   fStackAveragingCount := templateFile.LineRepeatCount;
   fStackFrameCount := templateFile.StackFrameCount;
   fStackInterval := templateFile.StackInterval;
   fStackRepeatCount := templateFile.StackRepeatCount;
   fXFrameOffset := templateFile.XFrameOffset;
   fXStagePosition := templateFile.XStagePosition;
   fYFrameOffset := templateFile.YFrameOffset;
   fYStagePosition := templateFile.YStagePosition;
   fZStagePosition := templateFile.ZStagePosition;

   frameCommentList := TFrameCommentList.Create(self);
   frameGroupList := TList.Create;
   bDirty := False;
   viewerList := TList.Create;
   NewViewer;
end;


constructor TMPFile.CreateFromFile(const sFilename: string);
var    i: integer;
       hr: HResult;
begin
   {opens the storage}
   hr := StgOpenStorage(StringToOLEStr(sFilename), nil, RFLAGS, nil, 0, rootStorage);
   if Succeeded(hr) then
   begin
       propertySetStorage := rootStorage as IPropertySetStorage;
       if not Succeeded(propertySetStorage.Open(FMTID_User_Defined_Properties,
        RFLAGS, propertyStorage)) then MessageDlg('Cannot open property set', mtError, [mbOK], 0);

       ReadPropertySet;
       for i := 0 to MAX_CH - 1 do
           if chEnabled[i] then
               rootStorage.OpenStream(StringToOleStr('Ch' + IntToStr(i)), nil, RFLAGS,
                   0, chStreams[i]);

       AdjustFrameCount;
       fFilename := sFilename;

       if ScanMode = SM_REGIONSCAN then GetRegions;

       fCurrentFrameGroup := TFrameGroup.Create(self);
       fActiveFrameIndex := -1; ActiveFrameIndex := 0; {forces loading}

       frameCommentList := TFrameCommentList.Create(self);
       frameCommentList.LoadFrameComments;

       viewerList := TList.Create;
       NewViewer;

       if AnalogChCount > 0 then
       begin
           analogWndList := TList.Create;
           NewAnalogWnd;
       end;
   end
   else
       MessageDlg(FileErrToStr(StgErrToFileErr(hr)), mtError, [mbOK], 0);
end;


destructor  TMPFile.Destroy;
var    i: integer;
begin
   fCurrentFrameGroup.Free;
   frameGroupList.Free;

	if analogWndList <> nil then
   begin
   	if analogWndList.Count > 0 then
       	for i := 0 to analogWndList.Count - 1 do
       	begin
           	TAnalogFrm(analogWndList.Items[i]).mpFile := nil;
           	TAnalogFrm(analogWndList.Items[i]).Close;
           	if Mainform.bAppClosing then Application.ProcessMessages;
       	end;
   	analogWndList.Free;
   end;
   {a lot of Windows voodoo here - why it seems to work beats me}
   if viewerList.Count > 0 then
       for i := 0 to viewerList.Count - 1 do
       begin
           TViewerFrm(viewerList.Items[i]).mpFile := nil;
           TViewerFrm(viewerList.Items[i]).Close;
           if Mainform.bAppClosing then Application.ProcessMessages;
       end;
   viewerList.Free;

   frameCommentList.Free;
  { for i := 0 to MAX_CH - 1 do channels[i] := nil;}

   propertyStorage := nil;
   propertySetStorage := nil;
   rootStorage := nil;
   Mainform.FileList.NotifyFileClosing(self);
end;


{********************************* TFileList **********************************}

{------------------------------- TFileList: private ---------------------------}

function    TFileList.IsMPFile(const sFilename: string): boolean;
{var clsid: TCLSID;}
begin
  { Result := False;}
   {if Succeeded(StgIsStorageFile(StringToOLEStr(sFilename)) then}
  { if Succeeded(GetClassFile(StringToOLEStr(sFilename), clsid)) then
       Result := IsEqualGUID(clsid, GUID_MPD);}
   Result := True;
end;


function    TFileList.GetTemporaryFileName: string;
var    i: integer;
begin
   i := 1;
   Result := 'Workspace ' + IntToStr(i);
   while IndexOf(Result) >= 0 do
   begin
       i := i + 1;
       Result := 'Workspace ' + IntToStr(i);
   end;
end;


function    TFileList.Load(const sFilename: string): TFileErr;
begin
   if not IsMPFile(sFilename) then
       Result := feNotAMPFile
   else if IndexOf(sFilename) >= 0 then
      Result := feFileAlreadyOpen
   else
   begin
       AddObject(sFilename, TMPFile.CreateFromFile(sFilename));
       Result := feOK;
   end;
end;


{------------------------------- TFileList: public ---------------------------}

procedure   TFileList.FillComboBoxWithWorkspaces(aComboBox: TComboBox);
var    i: integer;
begin
   aComboBox.Clear;
   if Count > 0 then
   begin
       for i := 0 to Count - 1 do
           if TMPFile(Objects[i]).IsMemoryFile then
               aComboBox.Items.AddObject(TMPFile(Objects[i]).Filename, Objects[i]);
       if aComboBox.Items.Count > 0 then aComboBox.ItemIndex := 0;
   end;
end;


procedure   TFileList.NewFile(templateFile: TMPFile);
var    tempFileName: string;
begin
   tempFileName := GetTemporaryFileName;
   AddObject(tempFileName, TMPFile.CreateFromTemplate(tempFileName, templateFile));
end;


procedure   TFileList.NotifyFileClosing(Sender: TMPFile);
var    i: integer;
begin
   if not Mainform.bAppClosing then
   begin
       i := IndexOfObject(Sender);
       if i >= 0 then Delete(i);
   end;
end;


function    TFileList.WorkspaceCount: integer;
var    i: integer;
begin
   Result := 0;
   if Count > 0 then
       for i := 0 to Count - 1 do
           if TMPFile(Objects[i]).IsMemoryFile then Result := Result + 1;
end;


procedure   TFileList.SaveFileAs(mpFile: TMPFile; const newName: string);
var    i: integer;
       fe: TFileErr;
begin
   i := IndexOfObject(mpFile);
   if i >= 0 then
       if TMPFile(Objects[i]).IsMemoryFile then
       begin
           fe := TMPFile(Objects[i]).SaveAs(newName);
           if fe = feOK then
               {swaps names}
               Strings[i] := newName
           else
               MessageDlg('Cannot save ' + Strings[i] + ' as' + CRLF +
                       newName + CRLF +
                       FileErrToStr(fe), mtError, [mbOK], 0);
       end;
end;


procedure   TFileList.Open(const sFilename: string);
var    fe: TFileErr;
begin
   if IndexOf(sFilename) >= 0 then
       MessageDlg(Format(sAlreadyLoaded, [sFilename]), mtInformation, [mbOK], 0)
   else
   begin
       fe := Load(sFilename);
       if fe <> feOK then
           MessageDlg(FileErrToStr(fe), mtError, [mbOK], 0)
       else
           dataDirectory := ExtractFilePath(sFilename);
   end;
end;


destructor  TFileList.Destroy;
var    i: integer;
begin
   if Count > 0 then
       for i := 0 to Count - 1 do
           TMPFile(Objects[i]).Free;
   inherited Destroy;
end;




end.
